/*****************************************************************************
 *
 * FILE:	wdbm.c
 * DESCRIPTION:	wyvern database manager
 * DATE:	Sat, Jun 30 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2001-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: wdbm.c,v 1.6.6.1 2003/03/18 16:05:20 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wdbm.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	EOT		'\004'	/* End Of Text */
#define	EOL		'\0'
#define	CR		'\015'
#define	LF		'\012'
#define	WDBM_SUFFIX	".db"

#if	TEXT_DB
#define	DATA_REPLACE	0
#define	DATA_DELETE	1
#endif	/* TEXT_DB */


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static char *	getDBName(const char *);
#if	TEXT_DB
static int	transform(WDBM_FILE, wdatum, wdatum, int);
static char *	chomp(char *);
#ifndef	HAVE_MKSTEMP
static int	mkstemp(char *);
#endif	/* HAVE_MKSTEMP */
#endif	/* TEXT_DB */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	wdbm_open
 * Description:		open database
 * Arguments:		database name, database flag, open mode
 * Return Value:	Successful: WDBM_FILE pointer, otherwise: NULL
 */
WDBM_FILE
wdbm_open(name, flags, mode)
	const char *	name;
	int		flags;
	int		mode;
{
#if	USE_DB
  WDBM_FILE	dp	= NULL;
  int		db_flags;
  char *	db_name = getDBName(name);

  if (db_name == NULL) {
    return NULL;
  }

  switch (flags) {
    case WDBM_WRITER:
      db_flags	= O_RDWR;
      break;
    case WDBM_WRCREAT:
      db_flags	= O_RDWR | O_CREAT;
      break;
    case WDBM_NEWDB:
      db_flags	= O_RDWR | O_CREAT | O_TRUNC;
      break;
    case WDBM_READER:
    default:
      db_flags	= O_RDONLY;
      break;
  }

  dp = dbopen(db_name, db_flags, mode, DB_HASH, NULL);
  free((char *)db_name);
  return dp;
#elif	USE_DB2
  WDBM_FILE	dp	= NULL;
  int		db_flags;
  int		ret;
  char *	db_name = getDBName(name);

  if (db_name == NULL) {
    return NULL;
  }

  dp = (WDBM_FILE)calloc(1, sizeof(_wdbm_file));
  if (dp == NULL) {
    free((char *)db_name);
    return NULL;
  }
  dp->dbp  = NULL;
  dp->dbcp = NULL;

  switch (flags) {
    case WDBM_WRITER:
      db_flags	= 0;
      break;
    case WDBM_WRCREAT:
      db_flags	= DB_CREATE;
      break;
    case WDBM_NEWDB:
      db_flags	= DB_CREATE | DB_TRUNCATE;
      break;
    case WDBM_READER:
    default:
      db_flags	= DB_RDONLY;
      break;
  }

  ret = db_open(db_name, DB_HASH, db_flags, mode, NULL, NULL, &dp->dbp);
  free((char *)db_name);
  return ret == 0 ? dp : NULL;
#elif	USE_NDBM
  int	db_flags;

  switch (flags) {
    case WDBM_WRITER:
      db_flags	= O_RDWR;
      break;
    case WDBM_WRCREAT:
      db_flags	= O_RDWR | O_CREAT;
      break;
    case WDBM_NEWDB:
      db_flags	= O_RDWR | O_CREAT | O_TRUNC;
      break;
    case WDBM_READER:
    default:
      db_flags	= O_RDONLY;
      break;
  }

  return dbm_open(name, db_flags, mode);
#elif	USE_GDBM
  GDBM_FILE	dp;
  int		db_flags;
  char *	db_name = getDBName(name);

  if (db_name == NULL) {
    return NULL;
  }

  switch (flags) {
    case WDBM_WRITER:
      db_flags	= GDBM_WRITER;
      break;
    case WDBM_WRCREAT:
      db_flags	= GDBM_WRCREAT;
      break;
    case WDBM_NEWDB:
      db_flags	= GDBM_NEWDB;
      break;
    case WDBM_READER:
    default:
      db_flags	= GDBM_READER;
      break;
  }

  dp = gdbm_open(db_name, 512, db_flags, mode, NULL);
  free((char *)db_name);

  return dp;
#else
  WDBM_FILE	dp = NULL;
  int		fd = -1;
  int		db_flags;
  char *	db_mode = "r";

  dp = (WDBM_FILE)calloc(1, sizeof(_wdbm_file));
  if (dp == NULL) {
    return NULL;
  }

  dp->name = getDBName(name);
  if (dp->name == NULL) {
    free((WDBM_FILE)dp);
    return NULL;
  }

  switch (flags) {
    case WDBM_WRITER:
      db_flags	= O_RDWR | O_EXLOCK;
      db_mode	= "a+";
      break;
    case WDBM_WRCREAT:
      db_flags	= O_RDWR | O_CREAT | O_EXLOCK;
      db_mode	= "a+";
      break;
    case WDBM_NEWDB:
      db_flags	= O_RDWR | O_CREAT | O_EXLOCK | O_TRUNC;
      db_mode	= "w+";
      break;
    case WDBM_READER:
    default:
      db_flags	= O_RDONLY;
      db_mode	= "r";
      break;
  }

  fd = open(dp->name, db_flags, mode);
  if (fd == -1) {
    free((char *)dp->name);
    free((WDBM_FILE)dp);
    return NULL;
  }
  dp->fp = fdopen(fd, db_mode);
  if (dp->fp == NULL) {
    free((char *)dp->name);
    free((WDBM_FILE)dp);
    return NULL;
  }

  return dp;
#endif
}

/*
 * Function Name:	wdbm_close
 * Description:		close database
 * Arguments:		WDBM_FILE pointer
 * Return Value:	none
 */
void
wdbm_close(dp)
	WDBM_FILE	dp;
{
#if	USE_DB
  dp->close(dp);
#elif	USE_DB2
  dp->dbp->close(dp->dbp, 0);
  if (dp->dbcp) {
    dp->dbcp->c_close(dp->dbcp);
  }
  free((WDBM_FILE)dp);
#elif	USE_NDBM
  dbm_close(dp);
#elif	USE_GDBM
  gdbm_close(dp);
#else
  fclose(dp->fp);
  free((char *)dp->name);
  free((WDBM_FILE)dp);
#endif
}


/*
 * Function Name:	wdbm_store
 * Description:		store data into database
 * Arguments:		WDBM_FILE pointer, key, content, flag
 * Return Value:	Successful: 0, otherwise: -1
 */
int
wdbm_store(dp, key, content, flag)
	WDBM_FILE	dp;
	wdatum		key;
	wdatum		content;
	int		flag;
{
#if	USE_DB
  u_int	db_flag;
  DBT	db_key;
  DBT	db_content;

  db_key.data	  = key.dptr;
  db_key.size	  = key.dsize;
  db_content.data = content.dptr;
  db_content.size = content.dsize;

  switch (flag) {
    case WDBM_REPLACE:
      db_flag = 0;
      break;
    case WDBM_INSERT:
    default:
      db_flag = R_NOOVERWRITE;
      break;
  }

  return dp->put(dp, &db_key, &db_content, db_flag);
#elif	USE_DB2
  int	db_flag;
  DBT	db_key;
  DBT	db_content;

  switch (flag) {
    case WDBM_REPLACE:
      db_flag = 0;
      break;
    case WDBM_INSERT:
    default:
      db_flag = DB_NOOVERWRITE;
      break;
  }
  db_key.data	  = key.dptr;
  db_key.size	  = key.dsize;
  db_content.data = content.dptr;
  db_content.size = content.dsize;
  return dp->dbp->put(dp->dbp, NULL, &db_key, &db_content, db_flag) == 0 ? 0
									 : -1;
#elif	USE_NDBM
  int	db_flag;

  switch (flag) {
    case WDBM_REPLACE:
      db_flag = DBM_REPLACE;
      break;
    case WDBM_INSERT:
    default:
      db_flag = DBM_INSERT;
      break;
  }
  return dbm_store(dp, key, content, db_flag) == -1 ? -1 : 0;
#elif	USE_GDBM
  int	db_flag;

  switch (flag) {
    case WDBM_REPLACE:
      db_flag = GDBM_REPLACE;
      break;
    case WDBM_INSERT:
    default:
      db_flag = GDBM_INSERT;
      break;
  }
  return gdbm_store(dp, key, content, db_flag) == -1 ? -1 : 0;
#else
  if (flag == WDBM_INSERT) {
    if (fseek(dp->fp, 0L, SEEK_END) == -1) {
      return -1;
    }

    if (flag == WDBM_INSERT && wdbm_exists(dp, key) == 0) {
      return -1;
    }
    fprintf(dp->fp, "%s%c%s\n", key.dptr, EOT, content.dptr);
    fflush(dp->fp);

    return 0;
  }
  else {
    return transform(dp, key, content, DATA_REPLACE);
  }
#endif
}


/*
 * Function Name:	wdbm_fetch
 * Description:		fetch data from database
 * Arguments:		WDBM_FILE pointer, key
 * Return Value:	data
 */
wdatum
wdbm_fetch(dp, key)
	WDBM_FILE	dp;
	wdatum		key;
{
#if	USE_DB
  DBT		db_key;
  DBT		db_data;
  wdatum	data;
  int		ret;

  db_key.data = key.dptr;
  db_key.size = key.dsize;
  ret = dp->get(dp, &db_key, &db_data, 0);
  if (ret != 0) {
    db_data.data = NULL;
    db_data.size = 0;
  }
  data.dptr  = db_data.data;
  data.dsize = db_data.size;
  return data;
#elif	USE_DB2
  DBT		db_key;
  DBT		db_data;
  wdatum	data;
  int		ret;

  db_key.data = key.dptr;
  db_key.size = key.dsize;
  ret = dp->dbp->get(dp->dbp, NULL, &db_key, &db_data, 0);
  if (ret != 0) {
    db_data.data = NULL;
    db_data.size = 0;
  }
  data.dptr  = db_data.data;
  data.dsize = db_data.size;
  return data;
#elif	USE_NDBM
  return dbm_fetch(dp, key);
#elif	USE_GDBM
  return gdbm_fetch(dp, key);
#else
  wdatum	data;
  long		curp;
  char		buf[BUFSIZ];

  data.dptr  = NULL;
  data.dsize = 0;

  curp = ftell(dp->fp);	/* get current file position */
  (void)fseek(dp->fp, 0L, SEEK_SET);
  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), dp->fp) != NULL) {
    if (strncmp(key.dptr, buf, key.dsize) == 0) {
      if (buf[key.dsize] == EOT) {
	char *	p = buf + key.dsize;
	*p++ = EOL;
	chomp(p);
	data.dptr  = strdup(p);
	data.dsize = strlen(p);
	break;
      }
    }
  }
  (void)fseek(dp->fp, curp, SEEK_SET);

  return data;
#endif
}

/*
 * Function Name:	wdbm_delete
 * Description:		delete data in database
 * Arguments:		WDBM_FILE pointer, key
 * Return Value:	Successful: 0, otherwise: -1
 */
int
wdbm_delete(dp, key)
	WDBM_FILE	dp;
	wdatum		key;
{
#if	USE_DB
  DBT	db_key;

  db_key.data = key.dptr;
  db_key.size = key.dsize;
  return dp->del(dp, &db_key, 0);
#elif	USE_DB2
  DBT	db_key;

  db_key.data = key.dptr;
  db_key.size = key.dsize;
  return dp->dbp->del(dp->dbp, NULL, &db_key, 0) == 0 ? 0 : -1;
#elif	USE_NDBM
  return dbm_delete(dp, key) == -1 ? -1 : 0;
#elif	USE_GDBM
  return gdbm_delete(dp, key);
#else
  /*
   * NO COMPLETED CODE
   */
  wdatum	data;

  data.dptr  = NULL;
  data.dsize = 0;
  return transform(dp, key, data, DATA_DELETE);
#endif
}

/*
 * Function Name:	wdbm_firstkey
 * Description:		get firstkey from database
 * Arguments:		WDBM_FILE pointer
 * Return Value:	data
 */
wdatum
wdbm_firstkey(dp)
	WDBM_FILE	dp;
{
#if	USE_DB
  DBT		db_key;
  DBT		db_data;
  wdatum	firstkey;
  int		ret;

  firstkey.dptr	 = NULL;
  firstkey.dsize = 0;

  ret = dp->seq(dp, &db_key, &db_data, R_FIRST);
  if (ret == 0) {
    firstkey.dptr  = db_key.data;
    firstkey.dsize = db_key.size;
  }
  return firstkey;
#elif	USE_DB2
  DBT		db_key;
  DBT		db_data;
  wdatum	firstkey;
  int		ret;

  firstkey.dptr	 = NULL;
  firstkey.dsize = 0;

  ret = dp->dbp->cursor(dp->dbp, NULL, &dp->dbcp, 0);
  if (ret == 0) {
    ret = dp->dbcp->c_get(dp->dbcp, &db_key, &db_data, DB_FIRST);
    if (ret != DB_NOTFOUND) {
      firstkey.dptr  = db_key.data;
      firstkey.dsize = db_key.size;
    }
  }
  return firstkey;
#elif	USE_NDBM
  return dbm_firstkey(dp);
#elif	USE_GDBM
  return gdbm_firstkey(dp);
#else
  wdatum	firstkey;
  char		buf[BUFSIZ];

  firstkey.dptr	 = NULL;
  firstkey.dsize = 0;

  (void)fseek(dp->fp, 0L, SEEK_SET);
  memset(buf, 0, sizeof(buf));
  if (fgets(buf, sizeof(buf), dp->fp) != NULL) {
    char *	p = strchr(buf, EOT);
    if (p != NULL) {
      *p = EOL;
    }
    firstkey.dptr  = strdup(buf);
    firstkey.dsize = strlen(buf);
  }
  return firstkey;
#endif
}

/*
 * Function Name:	wdbm_nextkey
 * Description:		get nextkey from database
 * Arguments:		WDBM_FILE pointer, current key
 * Return Value:	data
 */
wdatum
wdbm_nextkey(dp, key)
	WDBM_FILE	dp;
	wdatum		key;
{
#if	USE_DB
  DBT		db_key;
  DBT		db_data;
  wdatum	nextkey;
  int		ret = 0;

  nextkey.dptr  = NULL;
  nextkey.dsize = 0;

  db_key.data = key.dptr;
  db_key.size = key.dsize;
  ret = dp->seq(dp, &db_key, &db_data, R_NEXT);
  if (ret == 0) {
    nextkey.dptr  = db_key.data;
    nextkey.dsize = db_key.size;
  }
  return nextkey;
#elif	USE_DB2
  DBT		db_key;
  DBT		db_data;
  wdatum	nextkey;
  int		ret = 0;

  nextkey.dptr	= NULL;
  nextkey.dsize = 0;

  if (dp->dbcp == NULL) {
    ret = dp->dbp->cursor(dp->dbp, NULL, &dp->dbcp, 0);
  }
  if (ret == 0) {
    db_key.data = key.dptr;
    db_key.size = key.dsize;
    ret = dp->dbcp->c_get(dp->dbcp, &db_key, &db_data, DB_NEXT);
    if (ret != DB_NOTFOUND) {
      nextkey.dptr  = db_key.data;
      nextkey.dsize = db_key.size;
    }
  }
  return nextkey;
#elif	USE_NDBM
  return dbm_nextkey(dp);
#elif	USE_GDBM
  return gdbm_nextkey(dp, key);
#else
  wdatum	nextkey;
  char		buf[BUFSIZ];

  nextkey.dptr	= NULL;
  nextkey.dsize = 0;

  memset(buf, 0, sizeof(buf));
  if (fgets(buf, sizeof(buf), dp->fp) != NULL) {
    char *	p = strchr(buf, EOT);
    if (p != NULL) {
      *p = EOL;
    }
    nextkey.dptr  = strdup(buf);
    nextkey.dsize = strlen(buf);
  }
  return nextkey;
#endif
}

/*
 * Function Name:	wdbm_exists
 * Description:		does key exist in database?
 * Arguments:		WDBM_FILE pointer, key
 * Return Value:	Successful: 0, otherwise: -1
 */
int
wdbm_exists(dp, key)
	WDBM_FILE	dp;
	wdatum		key;
{
#if	USE_DB
  DBT	db_key;
  DBT	db_data;

  db_key.data = key.dptr;
  db_key.size = key.dsize;
  return dp->get(dp, &db_key, &db_data, 0);
#elif	USE_DB2
  DBT	db_key;
  DBT	db_data;

  db_key.data = key.dptr;
  db_key.size = key.dsize;
  return dp->dbp->get(dp->dbp, NULL, &db_key, &db_data, 0) == 0 ? 0 : -1;
#elif	USE_NDBM
  wdatum	ret = dbm_fetch(dp, key);
  return ret.dptr == NULL ? -1 : 0;
#elif	USE_GDBM
  return gdbm_exists(dp, key) ? 0 : -1;
#else
  long	curp;
  char	buf[BUFSIZ];

  curp = ftell(dp->fp);	/* get current file position */
  (void)fseek(dp->fp, 0L, SEEK_SET);
  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), dp->fp) != NULL) {
    if (strncmp(key.dptr, buf, key.dsize) == 0 &&
	buf[key.dsize] == EOT) {
      (void)fseek(dp->fp, curp, SEEK_SET);
      return 0;
    }
    memset(buf, 0, sizeof(buf));
  }
  (void)fseek(dp->fp, curp, SEEK_SET);

  return -1;
#endif
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	getDBName
 * Description:		make a regular database name
 * Arguments:		name
 * Return Value:	database name
 */
static char *
getDBName(name)
	const char *	name;
{
  char *	db_name = strdup(name);
  char *	p	= NULL;

  if (name == NULL) {
    return NULL;
  }

  if ((p = strrchr(db_name, '.')) != NULL && strcmp(p, WDBM_SUFFIX) == 0) {
    return db_name;
  }

#if	HAVE_ASPRINTF
  /* set db_name to be a NULL if error happened */
  (void)asprintf(&db_name, "%s%s", name, WDBM_SUFFIX);
#else
  db_name = (char *)calloc(strlen(name) + sizeof(WDBM_SUFFIX), sizeof(char));
  if (db_name == NULL) {
    return NULL;
  }
  (void)sprintf(db_name, "%s%s", name, WDBM_SUFFIX);
#endif	/* HAVE_ASPRINTF */

  return db_name;
}

#if	TEXT_DB
/*
 * Function Name:	transform
 * Description:	
 * Arguments:
 * Return Value:
 */
static int
transform(dp, key, data, flag)
	WDBM_FILE	dp;
	wdatum		key;
	wdatum		data;
	int		flag;
{
  FILE *	fp = NULL;
  char		temp[] = "/tmp/wdbm.XXXXXX";
  char		buf[BUFSIZ];

  /* open temporary file */
  fp = fdopen(mkstemp(temp), "w+");
  if (fp == NULL) {
    return -1;
  }
  (void)unlink(temp);

  /* copy */
  (void)fseek(dp->fp, 0L, SEEK_SET);
  while (!feof(dp->fp)) {
    size_t	len = fread(buf, sizeof(char), sizeof(buf), dp->fp);
    if (len > 0) {
      (void)fwrite(buf, sizeof(char), len, fp);
    }
  }

  if (ftruncate(fileno(dp->fp), (off_t)0) == -1) {
    fclose(fp);
    return -1;
  }
  (void)fseek(fp, 0L, SEEK_SET);
  (void)fseek(dp->fp, 0L, SEEK_SET);


  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fp) != NULL) {
    char *	p = strchr(buf, EOT);
    if (p != NULL) {
      *p++ = EOL;	/* p indicates the start address of data */
    }
    if (strcmp(key.dptr, buf) == 0 && flag == DATA_REPLACE) {
      fprintf(dp->fp, "%s%c%s\n", key.dptr, EOT, data.dptr);
    }
    else {
      fprintf(dp->fp, "%s%c%s", buf, EOT, p);
    }
    memset(buf, 0, sizeof(buf));
  }
  fflush(dp->fp);
  fclose(fp);

  return 0;
}

/*
 * Function Name:	chomp
 * Description:		remove  CRLF, CR or LF from string end
 * Arguments:		string
 * Return Value:	string
 */
static char *
chomp(s)
	char *	s;
{
  char *	p;
  size_t	len;

  if (s == NULL) {
    return NULL;
  }

  len = strlen(s);
  p   = &s[len - 1];
  if (*p == LF) {
    *p-- = EOL;
  }
  if (*p == CR) {
    *p-- = EOL;
  }

  return s;
}

#ifndef	HAVE_MKSTEMP
/*
 * Function Name:	mkstemp
 * Description:		create temporary file and return its file descriptor
 * Arguments:		template path
 * Return Value:	file descriptor
 */
static int
mkstemp(path)
	char *	path;
{
  char *	p = mktemp(path);

  if (p == NULL) {
    return -1;
  }

  return open(path, O_CREAT|O_EXCL|O_RDWR, 0600);
}
#endif	/* HAVE_MKSTEMP */
#endif	/* TEXT_DB */


/**/
#if	DEBUG_WDBM
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  WDBM_FILE	dp;
  wdatum	key;
  wdatum	data;
  char *	fname = "/tmp/test";

  if (argc > 1) {
    fname = *++argv;
  }


  /**/
  dp = wdbm_open(fname, WDBM_WRCREAT, 0600);
  if (dp == NULL) {
    fputs("wdbm_open: \n", stderr);
    exit(1);
  }

  key.dptr = "name"; key.dsize = 4;
  data.dptr = "kouichi"; data.dsize = 7;
  if (wdbm_store(dp, key, data, WDBM_INSERT) == -1) {
    fputs("wdbm_store: \n", stderr);
  }

  key.dptr = "software"; key.dsize = strlen(key.dptr);
  data.dptr = "Wyvern 2.0-PR05"; data.dsize = strlen(data.dptr);
  if (wdbm_store(dp, key, data, WDBM_INSERT) == -1) {
    fputs("wdbm_store: \n", stderr);
  }

  key.dptr = "soft"; key.dsize = strlen(key.dptr);
  data.dptr = "AnyDBM"; data.dsize = strlen(data.dptr);
  if (wdbm_store(dp, key, data, WDBM_INSERT) == -1) {
    fputs("wdbm_store: \n", stderr);
  }

  key.dptr = "today"; key.dsize = 5;
  data.dptr = "Sun, Jul 1 2001"; data.dsize = strlen(data.dptr);
  if (wdbm_store(dp, key, data, WDBM_INSERT) == -1) {
    fputs("wdbm_store: \n", stderr);
  }

  key.dptr = "today"; key.dsize = 5;
  data.dptr = "Sun, Jul 8 2001"; data.dsize = strlen(data.dptr);
  if (wdbm_store(dp, key, data, WDBM_REPLACE) == -1) {
    fputs("wdbm_store: not replaced\n", stderr);
  }
  wdbm_close(dp);
  fputs("store ... ok\n", stderr);

  /**/
  dp = wdbm_open(fname, WDBM_READER, 0400);
  if (dp == NULL) {
    fputs("wdbm_open: \n", stderr);
    exit(1);
  }
  for (key = wdbm_firstkey(dp); key.dptr; key = wdbm_nextkey(dp, key)) {
    char *	kstr = NULL;
    char *	dstr = NULL;
    if (key.dptr) {
      data = wdbm_fetch(dp, key);
      if (data.dptr) {
	kstr = (char *)calloc((size_t)(key.dsize + 1), sizeof(char));
	dstr = (char *)calloc((size_t)(data.dsize + 1), sizeof(char));
	strncpy(kstr, key.dptr, (size_t)key.dsize);
	strncpy(dstr, data.dptr, (size_t)data.dsize);
	fprintf(stdout, "%s[%d] := %s[%d]\n",
		kstr, key.dsize, dstr, data.dsize);
	free((char *)kstr);
	free((char *)dstr);
      }
    }
  }
  wdbm_close(dp);
  fputs("fetch ... ok\n", stderr);

  return 0;
}
#endif	/* DEBUG_WDBM */

