/*****************************************************************************
 *
 * FILE:	cond.c
 * DESCRIPTION:	addressing the conditional request
 * DATE:	Wed, Dec 15 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cond.c,v 1.17.2.15 2003/03/18 16:06:05 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <time.h>
#include <sys/stat.h>
#include "wyvern.h"
#include "functions.h"



/*****************************************************************************
 *
 *	declaration of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	ifModifiedSince(Message *);
static int	ifUnmodifiedSince(Message *);
static int	ifMatch(Message *);
static int	ifNoneMatch(Message *);
static int	ifRange(Message *);
static int	getRequestByteRanges(Message *);
static int	setContentRange(Message *);
static int	setContentLocation(Message *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static String	if_modified_since	= NULL;
static String	if_unmodified_since	= NULL;
static String	if_match		= NULL;
static String	if_none_match		= NULL;
static String	if_range		= NULL;
static String	range			= NULL;
static String	etag			= NULL;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	conditionalRequestChecker
 * Description:		perform the conditional request
 * Arguments:		message
 * Return Value:	successful: RET_OK, otherwise: RET_NG
 */
int
conditionalRequestChecker(msg)
	register Message *	msg;
{
#if	DEBUG
  fputs("<conditionalRequestChecker>\n", stderr);
#endif	/* DEBUG */
  if_modified_since = headerLookup(msg->request.headers, "IF_MODIFIED_SINCE");
  if_unmodified_since
	= headerLookup(msg->request.headers, "IF_UNMODIFIED_SINCE");
  if_match	= headerLookup(msg->request.headers, "IF_MATCH");
  if_none_match = headerLookup(msg->request.headers, "IF_NONE_MATCH");
  if_range	= headerLookup(msg->request.headers, "IF_RANGE");
  range		= headerLookup(msg->request.headers, "RANGE");
  etag		= headerLookup(msg->response.headers, "ETag");

  if (responseHeaderEntry(&msg->response.headers,
			  "Accept-Range", RANGE_UNIT) == NULL) {
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    return RET_NG;
  }

  if (if_range != NULL) {
    return ifRange(msg);
  }
  else if (if_modified_since != NULL) {
    return ifModifiedSince(msg);
  }
  else if (if_unmodified_since != NULL) {
    return ifUnmodifiedSince(msg);
  }
  else if (if_match != NULL) {
    return ifMatch(msg);
  }
  else if (if_none_match != NULL) {
    return ifNoneMatch(msg);
  }

  return RET_CONT;
}

/*****************************************************************************/

/*
 * Function Name:	ifModifiedSince
 * Description:		perform the conditional request with If-Modified-Since
 * Arguments:		message
 * Return Value:	successful/continue: RET_CONT, otherwise: RET_NG
 */
static int
ifModifiedSince(msg)
	register Message *	msg;
{
  time_t	ims;

  ims = convertDate(if_modified_since);
  if (ims == (time_t)0) {
    msg->response.status = STATUS_FORBIDDEN;
  }
  else if (msg->file.status.st_mtime <= ims) {
    msg->response.status = STATUS_NOT_MODIFIED;
    return setContentLocation(msg);
  }
  else {
    return RET_CONT;
  }

  return RET_NG;
}

/*
 * Function Name:	ifUnmodifiedSince
 * Description:		perform the conditional request with If-Unmodified-Since
 * Arguments:		message
 * Return Value:	successful/continue: RET_CONT, otherwise: RET_NG
 */
static int
ifUnmodifiedSince(msg)
	register Message *	msg;
{
  time_t	ius;

  ius = convertDate(if_unmodified_since);
  if (ius == (time_t)0) {
    msg->response.status = STATUS_FORBIDDEN;
  }
  else if (msg->file.status.st_mtime > ius) {
    msg->response.status = STATUS_NOT_MODIFIED;
    return setContentLocation(msg);
  }
  else {
    return RET_CONT;
  }

  return RET_NG;
}

/*
 * Function Name:	ifMatch
 * Description:		perform the conditional request with If-Match
 * Arguments:		message
 * Return Value:	successful: RET_CONT, otherwise: RET_NG
 */
static int
ifMatch(msg)
	register Message *	msg;
{
  if (wvStrcmp(if_match, etag) != 0 &&
      wvStrcmp(if_match, "*") != 0) {
    msg->response.status = STATUS_NOT_MODIFIED;
    return setContentLocation(msg);
  }
  return RET_CONT;
}

/*
 * Function Name:	ifNoneMatch
 * Description:		check the conditional request with If-None-Match
 * Arguments:		message
 * Return Value:	successful: RET_CONT, otherwise: RET_NG
 */
static int
ifNoneMatch(msg)
	register Message *	msg;
{
  if (wvStrcmp(if_none_match, etag) == 0 ||
      wvStrcmp(if_none_match, "*") == 0) {
    msg->response.status = STATUS_NOT_MODIFIED;
    return setContentLocation(msg);
  }
  return RET_CONT;
}

/*
 * Function Name:	ifRange
 * Description:		perform the conditional request with If-Range
 * Arguments:		message
 * Return Value:	successful: RET_OK, otherwise: RET_NG
 */
static int
ifRange(msg)
	register Message *	msg;
{
  time_t	t = convertDate(if_range);

  if (wvStrcmp(if_range, etag) != 0 &&		/* check the entity-tag */
      msg->file.status.st_mtime != t) {	/* check the HTTP-date */
    msg->response.status = STATUS_REQUESTED_RANGE_NOT_SATISFIABLE;
    return RET_NG;
  }
  return getRequestByteRanges(msg);
}

/*
 * Function Name:	getRequestByteRanges
 * Description:		get the byte ranges
 * Arguments:		message
 * Return Value:	successful: RET_OK, otherwise: RET_NG
 */
static int
getRequestByteRanges(msg)
	register Message *	msg;
{
  register String	p;
  register String	first;
  register String	last;

  if (range == NULL) {
    msg->response.status = STATUS_REQUESTED_RANGE_NOT_SATISFIABLE;
    return RET_NG;
  }

  p = strstr(range, "bytes=");
  if (p == NULL) {
    return RET_NG;	/* invalid range specification */
  }
  first = p + 6;

  p = strchr(first, '-');
  if (p == NULL) {
    return RET_NG;	/* invalid range specification */
  }
  *p++ = EOL;
  last = p;

  if (*first == '-') {		/* ex. bytes=-500 */
    msg->file.range.first = 0;
    msg->file.range.last  = atoi(last);
  }
  else if (*first == EOL && *last == EOL) {
    return RET_NG;	/* invalid range specification */
  }
  else if (*last == EOL) {	/* ex. bytes=9500- */
    msg->file.range.first = atoi(first);
    msg->file.range.last  = msg->file.status.st_size;
  }
  else {			/* ex. bytes=500-900 */
    msg->file.range.first = atoi(first);
    msg->file.range.last  = atoi(last);
  }

  /* set new position in buffer */
  msg->response.content_length
	= msg->file.range.last - msg->file.range.first + 1;
  msg->response.status = STATUS_PARTIAL_CONTENT;
  return setContentRange(msg);
}

/*
 * Function Name:	setContentRange
 * Description:		set Content-Range field
 * Arguments:		message
 * Return Value:	successful: RET_OK, otherwise: RET_NG
 */
static int
setContentRange(msg)
	register Message *	msg;
{
  String	content_range;
  Headers *	hp;

  if (asprintf(&content_range, "%s %d-%d/%ld",
		RANGE_UNIT,
		msg->file.range.first, msg->file.range.last,
		(long)msg->file.status.st_size) == -1) {
    return wvNoMemoryError(msg);
  }

  hp = responseHeaderEntry(&msg->response.headers,
			  "Content-Range", content_range);
  free((String)content_range);
  if (hp != NULL) {
    return RET_OK;
  }
  return wvNoMemoryError(msg);
}

/*
 * Function Name:	setContentLocation
 * Description:		set Content-Location field
 * Arguments:		message
 * Return Value:	successful: RET_OK, otherwise: RET_NG
 */
static int
setContentLocation(msg)
	register Message *	msg;
{
  String	content_location;
  Headers *	hp;
  int		n;

#if	USE_SSL
  if (msg->request.http == HTTP_11) {
    if (msg->conn->sock.type == SOC_SECURE) {
      n = asprintf(&content_location, "https://%s%s",
 		   msg->request.host, msg->request.uri);
    }
    else {
      n = asprintf(&content_location, "http://%s%s",
 		   msg->request.host, msg->request.uri);
    }
  }
  else {
    if (msg->conn->sock.type == SOC_SECURE) {
      n = asprintf(&content_location, "https://%s:%d%s",
		   msg->vhost->serverName, msg->conn->sock.port,
		   msg->request.uri);
    }
    else {
      n = asprintf(&content_location, "http://%s:%d%s",
		   msg->vhost->serverName, msg->conn->sock.port,
		   msg->request.uri);
    }
  }
  if (n == -1) {
    return wvNoMemoryError(msg);
  }
#else
  if (msg->request.http == HTTP_11) {
    n = asprintf(&content_location, "http://%s%s",
		 msg->request.host, msg->request.uri);
  }
  else {
    n = asprintf(&content_location, "http://%s:%d%s",
		 msg->vhost->serverName, msg->conn->sock.port,
		 msg->request.uri);
  }
  if (n == -1) {
    return wvNoMemoryError(msg);
  }
#endif	/* USE_SSL */

  hp = responseHeaderEntry(&msg->response.headers,
			  "Content-Location", content_location);
  free((String)content_location);
  if (hp == NULL) {
    return wvNoMemoryError(msg);
  }

  return RET_OK;
}

