/*****************************************************************************
 *
 * FILE:	date.c
 * DESCRIPTION:	time/date functions
 * DATE:	Wed, Dec  1 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: date.c,v 1.11.6.7 2003/08/18 12:57:45 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <time.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	DATE_BUFSIZE	(36)


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	getTimeOfRFC822
 * Description:		return the time formatted in RFC 822
 * Arguments:		specified time_t, GMT flag
 * Return Value:	time
 */
String
getTimeOfRFC822(t, gmt)
	time_t	t;
	int	gmt;
{
#ifndef	HAVE_STRFTIME
  static const char	rfc822_days[][4] =
			{ "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
  static const char	rfc822_months[][4] =
			{ "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
#endif	/* HAVE_STRFTIME */
  struct tm *	tmp;
  static char	timebuf[DATE_BUFSIZE];

  if (t == 0) {	/* no specified time */
    time_t	now;
    time(&now);
    tmp = gmt ? gmtime(&now) : localtime(&now);
  }
  else {
    tmp = gmt ? gmtime(&t) : localtime(&t);
  }

  /*
   * time format per RFC 822/1123
   * Sun 06 Nov 1994 08:49:37 GMT
   */
#if	HAVE_STRFTIME
  strftime(timebuf, sizeof(timebuf), "%a, %d %b %Y %H:%M:%S %Z", tmp);
#else
  snprintf(timebuf, sizeof(timebuf),
	   "%.3s, %.02d %.3s %.4d %.02d:%.02d:%.02d %s",
	   rfc822_days[tmp->tm_wday],
	   tmp->tm_mday,
	   rfc822_months[tmp->tm_mon],
	   tmp->tm_year > 1900 ? tmp->tm_year : tmp->tm_year + 1900,
	   tmp->tm_hour,
	   tmp->tm_min,
	   tmp->tm_sec,
	   gmt ? "GMT" : "JST");
#endif	/* HAVE_STRFTIME */

  return timebuf;
}

/*
 * Function Name:	convertDate
 * Description:		convert RFC822/1123 date format to the time_t value
 * Arguments:		HTTP-date string
 * Return Value:	time_t value
 */
time_t
convertDate(date)
	String	date;
{
  struct tm	gmt;
  char		month[10];
  String	p;
  time_t	gmt_date;

  /*
   * Some browser send the strange If-Modified-Since request field such as
   *	Tue, 23 Jan 2001 12:28:29 GMT; length=969
   *
   * I read RFC-2616 and I couldn't understand the mean of "length"... 
   * So, I added the following code to solve the above problem.
   */
  p = strchr(date, ';');
  if (p != NULL) {
    *p = EOL;
  }

  if (strlen(date) > DATE_BUFSIZE) {
    return (time_t)0;
  }

  /*
   * Sun, 06 Nov 1994 08:49:37 GMT
   */
  if (sscanf(date, "%*s %02d %3s %4d %02d:%02d:%02d %*s",
		&(gmt.tm_mday), month, &(gmt.tm_year),
		&(gmt.tm_hour), &(gmt.tm_min), &(gmt.tm_sec)) == 6) {
    gmt.tm_year -= 1900;
  }
  /*
   * Sunday, 06-Nov-94 08:49:37 GMT
   */
  else if (sscanf(date, "%*s %02d-%3s-%2d %02d:%02d:%02d %*s",
		&(gmt.tm_mday), month, &(gmt.tm_year),
		&(gmt.tm_hour), &(gmt.tm_min), &(gmt.tm_sec)) == 6) {
  }
  /*
   * Sun Nov  6 08:49:37 1994 (ANSI C's asctime() format)
   */
  else if (sscanf(date, "%*s %3s %2d %02d:%02d:%02d %4d",
		month, &(gmt.tm_mday),
		&(gmt.tm_hour), &(gmt.tm_min), &(gmt.tm_sec),
		&(gmt.tm_year)) == 6) {
    gmt.tm_year -= 1900;
  }
  /*
   * Sun, Nov  6 1994 08:49:37 GMT
   */
  else if (sscanf(date, "%*s %3s %2d %4d %02d:%02d:%02d %*s",
		month, &(gmt.tm_mday), &(gmt.tm_year),
		&(gmt.tm_hour), &(gmt.tm_min), &(gmt.tm_sec)) == 6) {
    gmt.tm_year -= 1900;
  }
  /*
   * Unknown time format
   */
  else {
    return (time_t)0;
  }

  /*
   * convert string month to number
   */
  if	  (wvStrcasecmp(month, "Jan") == 0)	gmt.tm_mon = 0;
  else if (wvStrcasecmp(month, "Feb") == 0)	gmt.tm_mon = 1;
  else if (wvStrcasecmp(month, "Mar") == 0)	gmt.tm_mon = 2;
  else if (wvStrcasecmp(month, "Apr") == 0)	gmt.tm_mon = 3;
  else if (wvStrcasecmp(month, "May") == 0)	gmt.tm_mon = 4;
  else if (wvStrcasecmp(month, "Jun") == 0)	gmt.tm_mon = 5;
  else if (wvStrcasecmp(month, "Jul") == 0)	gmt.tm_mon = 6;
  else if (wvStrcasecmp(month, "Aug") == 0)	gmt.tm_mon = 7;
  else if (wvStrcasecmp(month, "Sep") == 0)	gmt.tm_mon = 8;
  else if (wvStrcasecmp(month, "Oct") == 0)	gmt.tm_mon = 9;
  else if (wvStrcasecmp(month, "Nov") == 0)	gmt.tm_mon = 10;
  else if (wvStrcasecmp(month, "Dec") == 0)	gmt.tm_mon = 11;
  gmt.tm_isdst = 0;

  gmt_date = mktime(&gmt);

#define	JST_OFFSET	(32400)
  return (gmt_date + JST_OFFSET);
}

