/*****************************************************************************
 *
 * FILE:	error.c
 * DESCRIPTION:	send an error message
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: error.c,v 1.39.2.15 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include "wyvern.h"
#include "mesg.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static errdoc_t		errorDocType(const String);
static ErrorDoc *	lookup(ErrorDoc *, const int);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	errorDocEntry
 * Description:		register response code and it's document file into hash
 * Arguments:		errorDocTable, response code, document location
 * Return Value:	none
 */
void
errorDocEntry(errorDocTable, code, text)
	ErrorDoc **	errorDocTable;
	const int	code;
	const String	text;
{
  ErrorDoc *	new;

  if (code <= STATUS_CODE_4XX) {
    return ;	/* invalid code */
  }

  if (lookup(*errorDocTable, code) != NULL) {
    return;	/* already registered */
  }

  new		 = (ErrorDoc *)wvMalloc(sizeof(ErrorDoc));
  new->code	 = code;
  new->text	 = wvStrdup(text);
  new->type	 = errorDocType(text);
  new->next	 = *errorDocTable;
  *errorDocTable = new;
}

/*
 * Function Name:	errorDocType
 * Description:		guess error document type
 * Arguments:		text of error document
 * Return Value:	errdoc_t
 */
static errdoc_t
errorDocType(text)
	const String	text;
{
  if (text[0] == '/') {
    return ED_FILE;
  }
  if (wvStrncasecmp(text, "http://", 7) == 0) {
    return ED_URL;
  }
  return ED_STRING;
}

/*
 * Function Name:	lookup
 * Description:		lookup the error document with response status code
 * Arguments:		errorDocTable, response status code
 * Return Value:	pointer of error document
 */
static ErrorDoc *
lookup(errorDocTable, code)
	ErrorDoc *	errorDocTable;
	const int	code;
{
  register ErrorDoc *	edp;

  for (edp = errorDocTable; edp != NULL; edp = edp->next) {
    if (edp->code == code) {
      return edp;
    }
  }

  return NULL;
}

/*
 * Function Name:	errorDocChecker
 * Description:		customizable error document performer
 * Arguments:		message
 * Return Value:	none
 */
void
errorDocChecker(msg)
	register Message *	msg;
{
  register ErrorDoc *	edp;

#if	DEBUG
  fprintf(stderr, "[errorDocChecker] status(%d)\n",
	  responseStatusTable[msg->response.status].code);
#endif	/* DEBUG */
  if (msg->response.status <= STATUS_CODE_4XX) {
    return;	/* no error */
  }

  /* close opened file */
  if (msg->file.fd > 0) {
    wvClose(msg->file.fd);
    msg->file.fd = -1;
  }

  edp = lookup(msg->vhost->errorDocTable,
	       responseStatusTable[msg->response.status].code);
  if (edp == NULL) {
    return;
  }

  switch (edp->type) {
    case ED_STRING:
      msg->response.error_mesg = edp->text;
      break;
    case ED_FILE: {
	String	path;

	if (asprintf(&path, "%s%s", msg->vhost->docRoot, edp->text) != -1) {
	  msg->file.fd = wvSafeOpen(path, O_RDONLY, &msg->file.status, FALSE);
	  free((String)path);
	  if (msg->file.fd != -1) {
	    msg->response.content_length = msg->file.status.st_size;
#if	DEBUG
	    fprintf(stderr, "[errorDoc] path = %s(%d)\n",
		    path, msg->file.fd);
#endif	/* DEBUG */
	  }
	}
      }
      break;
    case ED_URL:
      if (responseHeaderEntry(&msg->response.headers,
			      "Location", edp->text) != NULL) {
	msg->response.status = STATUS_MOVED_PERMANENTLY;
      }
      break;
    default:
      break;
  }
}

/*****************************************************************************/

/*
 * Function Name:	errorMessageSender
 * Description:		send an error message
 * Arguments:		thread
 * Return Value:	successful: 0, otherwise: -1
 */
int
errorMessageSender(thd)
	register Thread *	thd;
{
  /*
   * send header part
   */
  iosPrintf(thd,
	"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n"
	"<html lang=\"ja\">\n"
	"<head>\n"
	"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=EUC-JP\">\n"
	"<style type=\"text/css\">\n"
	"<!--\n"
	"body { background: white }\n"
	"h1 { text-align: center; color: red }\n"
	"-->\n"
	"</style>\n"
	"<title>%d %s</title>\n"
	"</head>\n"
	"<body lang=\"ja\">\n"
	"<h1>%d %s</h1>\n"
	"<p style=\"text-align: center\">\n",
	responseStatusTable[thd->msg->response.status].code,
	responseStatusTable[thd->msg->response.status].phrase,
	responseStatusTable[thd->msg->response.status].code,
	responseStatusTable[thd->msg->response.status].phrase);

  /*
   * send body part
   */
  if (thd->msg->response.error_mesg == NULL) {
    String	uri = sanitizeURI(thd->msg->request.uri);

    switch (thd->msg->response.status) {
      case STATUS_BAD_REQUEST:			/* 400 */
	iosPuts(thd, MSG_BAD_REQUEST);
	break;
      case STATUS_UNAUTHORIZED:			/* 401 */
	iosPrintf(thd, MSG_UNAUTHORIZED, uri);
	break;
      case STATUS_FORBIDDEN:			/* 403 */
	iosPrintf(thd, MSG_FORBIDDEN, uri);
	break;
      case STATUS_NOT_FOUND:			/* 404 */
	iosPrintf(thd, MSG_NOT_FOUND, uri);
	break;
      case STATUS_METHOD_NOT_ALLOWED:		/* 405 */
	iosPrintf(thd, MSG_METHOD_NOT_ALLOWED,
		  methodName[thd->msg->request.method]);
	break;
      case STATUS_NOT_ACCEPTABLE:			/* 406 */
	iosPrintf(thd, MSG_NOT_ACCEPTABLE, thd->msg->file.mime);
	break;
      case STATUS_REQUEST_TIMEOUT:			/* 408 */
	iosPuts(thd, MSG_REQUEST_TIMEOUT);
	break;
      case STATUS_PRECONDITION_FAILED:		/* 412 */
	iosPuts(thd, MSG_PRECONDITION_FAILED);
	break;
      case STATUS_REQUEST_URI_TOO_LONG:		/* 414 */
	iosPuts(thd, MSG_REQUEST_URI_TOO_LONG);
	break;
      case STATUS_INTERNAL_SERVER_ERROR:		/* 500 */
	iosPrintf(thd, MSG_INTERNAL_SERVER_ERROR, uri);
	break;
      case STATUS_NOT_IMPLEMENTED:			/* 501 */
	iosPrintf(thd, MSG_NOT_IMPLEMENTED, uri);
	break;
      case STATUS_HTTP_VERSION_NOT_SUPPORTED:	/* 505 */
	iosPuts(thd, MSG_HTTP_VERSION_NOT_SUPPORTED);
	break;
      default:
	break;
    }
    wvFree((String)uri);
  }
  else {
    iosPuts(thd, thd->msg->response.error_mesg);
  }

  /*
   * send footer part
   */
  iosPrintf(thd,
	"</p>\n"
	"<hr>\n"
	"<address>\n"
	"Powered by&nbsp;"
	"<a href=\"%s\">%s</a>\n"
	"&nbsp;&lt;<a href=\"mailto:%s\">%s</a>&gt;\n"
	"</address>\n"
	"</body>\n"
	"</html>\n",
	SERVER_ADDRESS, SERVER_VERSION,
	thd->msg->vhost->serverAdmin, thd->msg->vhost->serverAdmin);
  iosFlush(thd);

  return RET_OK;
}

