/*****************************************************************************
 *
 * FILE:	file.c
 * DESCRIPTION:	HEAD/GET method operation
 * DATE:	Fri, Aug 16 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2002-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: file.c,v 1.1.2.17 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static StrList *	newEntry(StrList **, const String);
static int		openDirectory(Message *);
static int		setETag(Message *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	indexesEntry
 * Description:		register index file of DirectoryIndexes directive
 * Arguments:		indexes string (CSV)
 * Return Value:	StrList pointer
 */
StrList *
indexesEntry(indexes)
	String	indexes;
{
  StrList *	new = NULL;
  String	s   = indexes;
  String	p;

  while ((p = strchr(s, ',')) != NULL) {
    *p++ = EOL;
    if (s != NULL && *s != EOL) {
      newEntry(&new, s);
    }
    s = p;
  }
  if (s != NULL && *s != EOL) {
    newEntry(&new, s);
  }

  return new;
}

/*
 * Function Name:	newEntry
 * Description:		register index file into string list
 * Arguments:		StrList, index filename
 * Return Value:	StrList pointer
 */
static StrList *
newEntry(list, name)
	StrList **	list;
	const String	name;
{
  StrList *	new;

  new	    = (StrList *)wvMalloc(sizeof(StrList));
  new->text = wvStrdup(name);
  new->next = *list;
  *list	    = new;

  return new;
}

/*****************************************************************************/

/*
 * Function Name:	translateDocRoot
 * Description:		translate a requests URI
 * Arguments:		message
 * Return Value:	successful: 0, no translated: 1, otherwise: -1
 */
int
translateDocRoot(msg)
	register Message *	msg;
{
  int	n;

#if	DEBUG
  fputs("<translateDocRoot>\n", stderr);
#endif	/* DEBUG */
  n = asprintf(&msg->file.path, "%s%s", msg->vhost->docRoot, msg->request.uri);
  if (n == -1) {
    return wvNoMemoryError(msg);
  }
#if	DEBUG
  fprintf(stderr, "[translateDocRoot] %s\n", msg->file.path);
#endif	/* DEBUG */

  return RET_OK;
}

/*
 * Function Name:	translatedPathChecker
 * Description:		verify translated path
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
translatedPathChecker(msg)
	register Message *	msg;
{
  int	fd;

#if	DEBUG
  fprintf(stderr, "[translatedPathChecker] status(%d)\n",
	  responseStatusTable[msg->response.status].code);
#endif	/* DEBUG */

again:
  fd = wvSafeOpen(msg->file.path, O_RDONLY, &msg->file.status,
		  httpd->secureLevel >= SECURE_LEVEL3 ? FALSE : TRUE);
  if (fd == -1) {
    switch (errno) {
      case ENOENT:
	msg->response.status = STATUS_NOT_FOUND;
	return RET_NG;
      case EMLINK:
      case EACCES:
	msg->response.status = STATUS_FORBIDDEN;
	return RET_NG;
      case EISDIR:
	return openDirectory(msg);
      case EINTR:
	goto again;
      default:
	wvSyslog(LOG_ERR, "failed to open file: %s\n", msg->file.path);
	msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
	return RET_NG;
    }
  }

  /* following symbolic link is restricted to secure region */
  if (msg->vhost->symlinks == FALSE &&
      safePathChecker(msg, msg->file.path) == RET_NG) {
    wvClose(fd);
    return RET_NG;
  }
  msg->file.fd = fd;
  msg->response.content_length = msg->file.status.st_size;
  return setETag(msg);
}

/*
 * Function Name:	openDirectory
 * Description:		walking directory to find Indexes
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
openDirectory(msg)
	register Message *	msg;
{
  /* redirect an illegal request URI */
  if (msg->request.uri[strlen(msg->request.uri)-1] != '/') {
    String	location;

    if (asprintf(&location, "%s/", msg->request.uri) != -1) {
      register Headers *	hp;

      hp = responseHeaderEntry(&msg->response.headers, "Location", location);
      free((String)location);
      if (hp != NULL) {
	msg->response.status = STATUS_MOVED_PERMANENTLY;
	return RET_OK;
      }
    }
  }
  /* DefaultIndexes */
  else {
    register StrList *	slp;
    String		path = wvStrdup(msg->file.path);	/* base */

    if (path != NULL) {
      for (slp = msg->vhost->dirIndexes; slp != NULL; slp = slp->next) {
	wvFree((String)msg->file.path);
	if (asprintf(&msg->file.path, "%s%s", path, slp->text) != -1) {
#if	DEBUG
	  fprintf(stderr, "[openDir] check %s\n", msg->file.path);
#endif	/* DEBUG */
	  if (translatedPathChecker(msg) == 0) {	/* recursive call */
	    msg->response.status = STATUS_NONE;	/* clear status */
	    free((String)path);
	    return RET_OK;
	  }
	}
      }
      /* directory! */
      msg->file.mime	   = DIR_MAGIC_TYPE;
      msg->response.status = STATUS_FORBIDDEN;	/* initialize */
      msg->file.path	   = path;	/* restore original */
      return RET_OK;
    }
  }

  return wvNoMemoryError(msg);
}

/*
 * Function Name:	fileMapping
 * Description:		map response content into memory
 * Arguments:		message
 * Return Value:	none
 */
void
fileMapping(msg)
	register Message *	msg;
{
  if (msg->file.fd == -1) {	/* error */
    return;
  }

  msg->file.data = mmap(NULL,
			(size_t)msg->file.status.st_size,
			PROT_READ,
			MAP_PRIVATE,
			msg->file.fd,
			(off_t)0);
  if (msg->file.data == MAP_FAILED) {
    wvSyslog(LOG_ERR, "failed to map file into memory\n");
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    msg->file.data	 = NULL;
    return;
  }

  if (msg->response.status == STATUS_PARTIAL_CONTENT) {
    msg->file.curp = msg->file.data + msg->file.range.first;
  }
  else {
    msg->file.curp = msg->file.data;
  }
  msg->file.rest = msg->response.content_length;
}

/*
 * Function Name:	setETag
 * Description:		set the entity tag
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
setETag(msg)
	register Message *	msg;
{
  String	etag;
  Headers *	hp;

  if (asprintf(&etag, "\"%lx-%lx-%lx\"",
		(long)msg->file.status.st_ino,
		(long)msg->file.status.st_size,
		msg->file.status.st_mtime) == -1) {
    return wvNoMemoryError(msg);
  }

  hp = responseHeaderEntry(&msg->response.headers, "ETag", etag);
  free((String)etag);
  if (hp == NULL) {
    return wvNoMemoryError(msg);
  }

  return RET_OK;
}

