/*****************************************************************************
 *
 * FILE:	headers.c
 * DESCRIPTION:	manipulate request/response headers
 * DATE:	Sun, Aug  4 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2002-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: headers.c,v 1.1.2.10 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static Headers *	lookup(Headers *, const String);
static String		toUpper(const String);
#if	0
static String		toLower(const String);
#endif


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	requestHeaderEntry
 * Description:		register request header into headerTable
 * Arguments:		headerTable, request header field
 * Return Value:	headerTable
 */
Headers *
requestHeaderEntry(headerTable, field)
	Headers **	headerTable;
	String		field;	/* destroyed original */
{
  Headers *	new;
  String	name;	/* header name */
  String	value;	/* header value */
  String	p;
  char		c;

  if (field == NULL) {
    return NULL;
  }

  p = strchr(field, ':');
  if (p == NULL) {
    return NULL;
  }
  *p++ = EOL;

  /* remove white space */
  while ((c = *p) && (c == SPC || c == TAB)) {
    p++;
  }
  value = wvChomp(p);
  name	= toUpper(field);

  new = lookup(*headerTable, name);
  if (new != NULL) {	/* already registered */
    String	merged;

    if (asprintf(&merged, "%s,%s", new->value, value) == -1) {	/* merged */
      return NULL;
    }
    wvFree((String)new->value);
    new->value = merged;
    return new;
  }

  new = (Headers *)malloc(sizeof(Headers));
  if (new == NULL) {
    return NULL;
  }
  new->name	= wvStrdup(name);
  new->value	= wvStrdup(value);
  new->next	= *headerTable;
  *headerTable	= new;

  return new;
}

/*
 * Function Name:	responseHeaderEntry
 * Description:		register response header into headerTable
 * Arguments:		headerTable, response header name/value pair
 * Return Value:	headerTable
 */
Headers *
responseHeaderEntry(headerTable, name, value)
	Headers **	headerTable;
	const String	name;
	const String	value;
{
  Headers *	new;

  if (name == NULL || value == NULL) {
    return NULL;
  }

  new = lookup(*headerTable, name);
  if (new != NULL) {
    wvFree((String)new->value);
    new->value = wvStrdup(value);
    return new;	/* replaced */
  }

  new = (Headers *)malloc(sizeof(Headers));
  if (new == NULL) {
    return NULL;
  }
  new->name	= wvStrdup(name);
  new->value	= wvStrdup(value);
  new->next	= *headerTable;
  *headerTable	= new;

  return new;
}

/*
 * Function Name:	lookup
 * Description:		lookup header from headerTable
 * Arguments:		headerTable, header field name
 * Return Value:	headerTable
 */
static Headers *
lookup(headerTable, name)
	Headers *	headerTable;
	const String	name;
{
  register Headers *	hp;

  for (hp = headerTable; hp != NULL; hp = hp->next) {
    if (wvStrcmp(hp->name, name) == 0) {
      return hp;
    }
  }
  return NULL;
}

/*
 * Function Name:	headerLookup
 * Description:		lookup header from headerTable
 * Arguments:		headerTable, header field name
 * Return Value:	header field value
 */
String
headerLookup(headerTable, name)
	Headers *	headerTable;
	const String	name;
{
  register Headers *	hp = lookup(headerTable, name);

  if (hp != NULL) {
    return hp->value;
  }
  return NULL;
}

/*
 * Function Name:	parseHeader
 * Description:		parse request headers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
parseHeader(msg)
	register Message *	msg;
{
  String	connection;

#if	DEBUG
  fputs("<parseHeader>\n", stderr);
#endif	/* DEBUG */
  /* Connection */
  connection = headerLookup(msg->request.headers, "CONNECTION");
  if (connection != NULL) {
    if (wvStrcasecmp(connection, "keep-alive") == 0) {
#if	BUGGY_KEEP_ALIVE
      if (msg->request.http == HTTP_10 && msg->conn->keepalive.max > 0) {
	msg->conn->keepalive.max--;
	msg->conn->keepalive.status = KA_CONTINUE;
      }
#endif	/* BUGGY_KEEP_ALIVE */
    }
    else if (wvStrcasecmp(connection, "close") == 0) {
      msg->conn->keepalive.status = KA_CLOSE;
    }
    else {
      msg->conn->keepalive.status = KA_DISABLED;
    }
  }

  /* Host */
  msg->request.host = headerLookup(msg->request.headers, "HOST");

  /* X-Forwarded-For */
  msg->request.x_forwarded_for
	= headerLookup(msg->request.headers, "X_FORWARDED_FOR");

  /* the conditional request is not overlap */
  if ((headerLookup(msg->request.headers, "IF_MATCH") != NULL &&
       headerLookup(msg->request.headers, "IF_NON_MATCH") != NULL) ||
      (headerLookup(msg->request.headers, "IF_MODIFIED_SINCE") != NULL &&
       headerLookup(msg->request.headers, "IF_UNMODIFIED_SINCE") != NULL)){
    msg->response.status = STATUS_PRECONDITION_FAILED;
  }
  /* the HTTP/1.1 request needs Host header field */
  else if (msg->request.http == HTTP_11 &&
	   msg->request.host == NULL) {
    msg->response.status = STATUS_BAD_REQUEST;
  }
  else {
    return RET_OK;
  }

  return RET_NG;
}

/*****************************************************************************/

/*
 * Function Name:	toUpper
 * Description:		upper case letter conversion in string
 * Arguments:		string
 * Return Value:	upper letters string
 * Note:		destroy original string
 */
static String
toUpper(s)
	String	s;
{
  String	ret = s;

  if (s == NULL) {
    return NULL;
  }

  for (; *s != EOL; s++) {
    if (*s >= 'a' && *s <= 'z') {
      *s -= 'a' - 'A';
    }
    else if (*s == '-') {
      *s = '_';
    }
  }

  return ret;
}

#if	0
/*
 * Function Name:	toLower
 * Description:		lower case letter conversion in string
 * Arguments:		string
 * Return Value:	lower letters string
 */
static String
toLower(str)
	const String	str;
{
  String	ret;
  String	wrk;

  if (str == NULL) {
    return NULL;
  }

  wrk = wvStrdup(str);
  if (wrk == NULL) {
    return str;
  }
  ret = wrk;

  for (; *wrk != EOL; wrk++) {
    if (*wrk >= 'A' && *wrk <= 'Z') {
      *wrk += 'a' - 'A';
    }
    else if (*wrk == '-') {
      *wrk = '_';
    }
  }

  return ret;
}
#endif

