/*****************************************************************************
 *
 * FILE:	hook.c
 * DESCRIPTION:	hook procedure manager for wyvern
 * DATE:	Tue, Aug 13 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2000-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: hook.c,v 1.1.2.22 2003/05/30 09:23:41 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
typedef struct _HookList {
  int			(*routine)(Message *);
  String		mime;		/* mime type for Handler */
  size_t		mlen;		/* mimte type valid length */
  hook_ordering_t	position;	/* hooked position */
  struct _HookList *	next;
} HookList;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	runFilter(HookList *, Message *, hook_ordering_t);
static int	runHookList0(HookList *, Message *);
static int	runHookList1(HookList *, Message *);
static int	runHookList2(HookList *, Message *);
static void	makeHookList(HookList **, int (*)(Message *),
			     hook_ordering_t, const String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static HookList *	headerParser	= NULL;
static HookList *	accessChecker	= NULL;
static HookList *	translateURI	= NULL;
static HookList *	checkUserID	= NULL;
static HookList *	authChecker	= NULL;
static HookList *	typeChecker	= NULL;
static HookList *	fixups		= NULL;
static HookList *	handler		= NULL;
static HookList *	filter		= NULL;
static HookList *	logger		= NULL;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	wvRunHeaderParser
 * Description:		run hooked header parsers
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunHeaderParser(msg)
	register Message *	msg;
{
  return runHookList0(headerParser, msg);
}

/*
 * Function Name:	wvRunAccessChecker
 * Description:		run hooked access checkers
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunAccessChecker(msg)
	register Message *	msg;
{
  return runHookList2(accessChecker, msg);
}

/*
 * Function Name:	wvRunTranslateURI
 * Description:		run hooked uri translators
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunTranslateURI(msg)
	register Message *	msg;
{
  register HookList *	hlp;

  for (hlp = translateURI; hlp != NULL; hlp = hlp->next) {
    switch (hlp->routine(msg)) {
      case RET_NG:
	return RET_NG;
      case RET_OK:
	if (hlp->position != HOOK_INITIAL) {
	  goto done;
	}
	break;
      case RET_EMB:
	return RET_OK;
      default:
	break;
    }
  }
done:
  return translatedPathChecker(msg);
}

/*
 * Function Name:	wvRunCheckUserID
 * Description:		run hooked uid checkers
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunCheckUserID(msg)
	register Message *	msg;
{
  return runHookList2(checkUserID, msg);
}

/*
 * Function Name:	wvRunAuthChecker
 * Description:		run hooked authentication checkers
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunAuthChecker(msg)
	register Message *	msg;
{
  return runHookList2(authChecker, msg);
}

/*
 * Function Name:	wvRunTypeChecker
 * Description:		run hooked mimetype checkers
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunTypeChecker(msg)
	register Message *	msg;
{
  return runHookList1(typeChecker, msg);
}

/*
 * Function Name:	wvRunFixups
 * Description:		run hooked fixups
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunFixups(msg, position)
	register Message *	msg;
	hook_ordering_t		position;
{
  register HookList *	hlp;

  for (hlp = fixups; hlp != NULL; hlp = hlp->next) {
    if (hlp->position == position) {
#if	DEBUG
	int	ret;
	ret = hlp->routine(msg);
	fprintf(stderr, "[fixups] return = %d\n", ret);
	if (ret == RET_NG) {
	  return ret;
	}
#else
      if (hlp->routine(msg) == RET_NG) {
	return RET_NG;
      }
#endif	/* DEBUG */
    }
  }
  return RET_OK;
}

/*
 * Function Name:	wvRunHandler
 * Description:		run hooked handlers
 * Arguments:		mesage
 * Return Value:	successful: RET_OK, cgi: RET_WAIT, otherwise: RET_NG
 */
int
wvRunHandler(msg)
	register Message *	msg;
{
  register HookList *	hlp;

  for (hlp = handler; hlp != NULL; hlp = hlp->next) {
    if (hlp->mime == NULL) {
      continue;
    }
    if (wvStrncmp(hlp->mime, msg->file.mime, hlp->mlen) == 0) {
      int	ret;

      ret = hlp->routine(msg);
#if	DEBUG
      fprintf(stderr, "[wvRunHandler] ret = %d\n", ret);
#endif	/* DEBUG */
      if (ret != RET_CONT) {
       	return ret;
      }
    }
  }

  return RET_OK;
}

/*
 * Function Name:	wvRunFilter
 * Description:		run hooked filters
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunFilter(msg)
	register Message *	msg;
{
  if (runFilter(filter, msg, HOOK_FIRST)  == RET_NG ||
      runFilter(filter, msg, HOOK_MIDDLE) == RET_NG ||
      runFilter(filter, msg, HOOK_LAST)	  == RET_NG ||
      runFilter(filter, msg, HOOK_FINAL)  == RET_NG) {
    return RET_NG;
  }
  return RET_OK;
}

/*
 * Function Name:	runFilter
 * Description:		run hooked filters
 * Arguments:		mesage, thread, position
 * Return Value:	successful: 0, otherwise: -1
 */
static int
runFilter(hookList, msg, position)
	register HookList *	hookList;
	register Message *	msg;
	hook_ordering_t		position;
{
  register HookList *	hlp;
  String		content_type = msg->file.mime;
  
  if (content_type == NULL) {
    content_type = headerLookup(msg->response.headers, "Content-Type");
  }
  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    if (hlp->position == position) {
      if (hlp->mime != NULL && content_type != NULL &&
	  wvStrncmp(hlp->mime, content_type, hlp->mlen) == 0) {
#if	DEBUG
	fprintf(stderr, "[filter] %s matched to %s\n",
		content_type, hlp->mime);
#endif	/* DEBUG */
       	return hlp->routine(msg);
      }
    }
  }
  return RET_OK;
}

/*
 * Function Name:	wvRunLogger
 * Description:		run hooked loggers
 * Arguments:		mesage
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvRunLogger(msg)
	register Message *	msg;
{
  return runHookList1(logger, msg);
}


/*
 * Function Name:	runHookList0
 * Description:		run hooked functions
 * Arguments:		hookList, mesage, position
 * Return Value:	successful: 0, otherwise: -1
 */
static int
runHookList0(hookList, msg)
	register HookList *	hookList;
	register Message *	msg;
{
  register HookList *	hlp;

  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    if (hlp->routine(msg) == RET_NG) {
      return RET_NG;
    }
  }
  return RET_OK;
}

/*
 * Function Name:	runHookList1
 * Description:		run hooked functions, return at first matching
 * Arguments:		hookList, mesage, position
 * Return Value:	successful: 0, otherwise: -1
 */
static int
runHookList1(hookList, msg)
	register HookList *	hookList;
	register Message *	msg;
{
  register HookList *	hlp;

  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    if (hlp->routine(msg) == RET_OK) {
      return RET_OK;
    }
  }
  return RET_NG;
}

/*
 * Function Name:	runHookList2
 * Description:		run hooked functions, return at first matching
 * Arguments:		hookList, mesage, position
 * Return Value:	successful: 0, otherwise: -1
 */
static int
runHookList2(hookList, msg)
	register HookList *	hookList;
	register Message *	msg;
{
  register HookList *	hlp;
  volatile int		ret = 0;

  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    ret = hlp->routine(msg);
    if (ret == 0) {
      return RET_OK;
    }
  }
  return ret;
}

/*****************************************************************************/

/*
 * Function Name:	wvHookHeaderParser
 * Description:		hook header parsers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookHeaderParser(routine, position)
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&headerParser, routine, position, NULL);
}

/*
 * Function Name:	wvHookAccessChecker
 * Description:		hook access checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookAccessChecker(routine, position)
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&accessChecker, routine, position, NULL);
}

/*
 * Function Name:	wvHookTranslateURI
 * Description:		hook uri translators
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookTranslateURI(routine, position)
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&translateURI, routine, position, NULL);
}

/*
 * Function Name:	wvHookCheckUserID
 * Description:		hook uid checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookCheckUserID(routine, position)
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&checkUserID, routine, position, NULL);
}

/*
 * Function Name:	wvHookAuthChecker
 * Description:		hook authentication checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookAuthChecker(routine, position)
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&authChecker, routine, position, NULL);
}

/*
 * Function Name:	wvHookTypeChecker
 * Description:		hook mimetype checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookTypeChecker(routine, position)
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&typeChecker, routine, position, NULL);
}

/*
 * Function Name:	wvHookFixups
 * Description:		hook fixups
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookFixups(routine, position)	
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&fixups, routine, position, NULL);
}

/*
 * Function Name:	wvHookHandler
 * Description:		hook handlers
 * Arguments:		routine, position, mimetype
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookHandler(routine, position, mime)
	int		(*routine)(Message *);
	hook_ordering_t	position;
	const String	mime;
{
  makeHookList(&handler, routine, position, mime);
}

/*
 * Function Name:	wvHookFilter
 * Description:		hook filters
 * Arguments:		routine, position, mimetype
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookFilter(routine, position, mime)	
	int		(*routine)(Message *);
	hook_ordering_t	position;
	const String	mime;
{
  makeHookList(&filter, routine, position, mime);
}

/*
 * Function Name:	wvHookLogger
 * Description:		hook loggers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wvHookLogger(routine, position)	
	int		(*routine)(Message *);
	hook_ordering_t	position;
{
  makeHookList(&logger, routine, position, NULL);
}

/*
 * Function Name:	makeHookList
 * Description:		make hook list
 * Arguments:		hookList, routine, position, mimetype
 * Return Value:	none
 */
static void
makeHookList(hookList, routine, position, mime)
	HookList **	hookList;
	int		(*routine)(Message *);
	hook_ordering_t	position;
	const String	mime;
{
  HookList *		new;
  register HookList *	hlp;
  register HookList *	prev_hlp;

  new		= (HookList *)wvMalloc(sizeof(HookList));
  new->routine	= routine;
  new->next	= NULL;
  if (mime != NULL) {
    size_t	len = strlen(mime);

    new->mime = wvStrdup(mime);
    new->mlen = mime[len - 1] == '*' ? len - 1 : len;
  }
  else {
    new->mime = NULL;
    new->mlen = 0;
  }
  new->position	= position;

  /* find position */
  for (prev_hlp = NULL, hlp = *hookList;
       hlp != NULL && hlp->position < position; hlp = hlp->next) {
    prev_hlp = hlp;
  }
  /* adjust position */
  for ( ; hlp != NULL && hlp->position == position; hlp = hlp->next) {
    prev_hlp = hlp;
  }

  if (hlp == NULL) {
    if (*hookList == NULL) {	/* at first */
      new->next = *hookList;
      *hookList	= new;
    }
    else if (prev_hlp != NULL) {
      prev_hlp->next = new;
    }
  }
  else {
    new->next = hlp;
    if (prev_hlp != NULL) {
      prev_hlp->next = new;
    }
    else {
      *hookList = new;
    }
  }
}

