/*****************************************************************************
 *
 * FILE:	module.c
 * DESCRIPTION:	module manager for wyvern
 * DATE:	Wed, Sep 13 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2000-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: module.c,v 1.36.2.14 2003/11/28 16:06:51 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_DLOPEN
#include <dlfcn.h>
#else
#if	HAVE_SHL_LOAD
#include <dl.h>
#else
#if	HAVE_LIBDLD
#include <dld.h>
#endif	/* HAVE_LIBDLD */
#endif	/* HAVE_SHL_LOAD */
#endif	/* HAVE_DLOPEN */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	MODULE_PATH	SERVER_ROOT "/modules"
#define	DLSYM_NAME	"getModuleInfo"

typedef struct _ModuleList {
  Module *		module;
  struct _ModuleList *	prev;
  struct _ModuleList *	next;
} ModuleList;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void *		load(const String);
static void *		fetch(void *);
static int		install(Module *);
static int		verify(Module *);

static int		unload(void *);
static int		uninstall(const String);
static ModuleList *	lookup(const String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static ModuleList *	moduleList = NULL;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	moduleLoad
 * Description:		load a module
 * Arguments:		module path
 * Return Value:	successful: 0, otherwise: -1
 */
int
moduleLoad(path)
	const String	path;
{
  void *	handle;
  Module *	module;
  int		ret;

  handle = load(path);
  if (handle == NULL) {
    goto error1;
  }

  module = (Module *)fetch(handle);
  if (module == NULL) {
    goto error;
  }

  ret = verify(module);
  if (ret == RET_NG) {
    goto error;
  }

  ret = install(module);
  if (ret == RET_NG) {
    goto error;
  }
  module->handle = handle;

  return RET_OK;

error:
  unload(handle);
error1:
  wvSyslog(LOG_ERR, "failed to load a module (%s)\n", path);

  return RET_NG;
}

#if	0
/*
 * Function Name:	moduleUnload
 * Description:		unload a module
 * Arguments:		module name
 * Return Value:	successful: 0, otherwise: -1
 */
int
moduleUnload(name)
	const String	name;
{
  return uninstall(name);
}
#endif

/*
 * Function Name:	moduleUnloadVulnerable
 * Description:		unload vulernable modules
 * Arguments:		none
 * Return Value:	none
 */
void
moduleUnloadVulnerable(void)
{
  register ModuleList *	mlp;
  register ModuleList *	mlp_next;

  for (mlp = moduleList; mlp != NULL; mlp = mlp_next) {
    int	ret;

    mlp_next = mlp->next;
    if (mlp->module->safety == SAFETY_VULNERABLE) {
      String	name = strdup(mlp->module->name);

      if (mlp->module->finish != NULL) {
	mlp->module->finish();
      }
      ret = unload(mlp->module->handle);
      if (ret == -1) {
	wvSyslog(LOG_ERR, "failed to unload the module: %s\n", name);
      }
#if	DEBUG
      fprintf(stderr, "DEBUG[module] the module: %s\n", name);
#endif	/* DEBUG */
      wvFree(name);

      if (mlp->prev == NULL) {
	mlp->next->prev	= mlp->prev;
	moduleList	= mlp->next;
      }
      else if (mlp->next == NULL) {
	mlp->prev->next	= mlp->next;
      }
      else {
	mlp->next->prev	= mlp->prev;
	mlp->prev->next	= mlp->next;
      }
      free((ModuleList *)mlp);	mlp = NULL;
    }
  }
}

/*
 * Function Name:	moduleInitializer
 * Description:		call initialize routine of modules
 * Arguments:		none
 * Return Value:	none
 */
void
moduleInitializer(void)
{
  /* drop privilege */
  if (dropPrivilege() == RET_OK) {
    register ModuleList *	mlp;

    for (mlp = moduleList; mlp != NULL; mlp = mlp->next) {
      if (mlp->module->init != NULL) {
	mlp->module->init();
      }
      mlp->module->register_hooks();
    }
    /* raise privilege */
    raisePrivilege();
  }
}

/*
 * Function Name:	moduleDelete
 * Description:		remove modules from wyvern
 * Arguments:		none
 * Return Value:	none
 */
void
moduleDelete(void)
{
  register ModuleList *	mlp;
  register ModuleList *	mlp_next;

  for (mlp = moduleList; mlp != NULL; mlp = mlp_next) {
#if	DEBUG
    fprintf(stderr, "[moduleDelete] %s\n", mlp->module->name);
#endif	/* DEBUG */
    mlp_next = mlp->next;
    if (mlp->module->finish != NULL) {
      mlp->module->finish();
    }
    unload(mlp->module->handle);
    free((ModuleList *)mlp);	mlp = NULL;
  }
  moduleList = NULL;
}

/*****************************************************************************/

/*
 * Function Name:	load
 * Description:		load a module
 * Arguments:		module path
 * Return Value:	successful: module handle, otherwise: NULL
 */
static void *
load(path)
	const String	path;
{
#if	HAVE_DLOPEN
  void *	handle;

#ifndef RTLD_NOW
#define RTLD_NOW	1
#endif	/* RTLD_NOW */
#ifndef RTLD_GLOBAL
#define RTLD_GLOBAL	0
#endif	/* RTLD_GLOBAL */
  handle = dlopen(path, RTLD_NOW|RTLD_GLOBAL);
#elif	HAVE_SHL_LOAD
  shl_t		handle;

  handle = shl_load(path, BIND_IMMEDIATE|BIND_VERBOSE|BIND_NOSTART, 0L);
#elif	HAVE_LIBDLD
#endif	/* HAVE_DLOPEN */
  if (handle == NULL) {
    wvSyslog(LOG_ERR, "failed to load a module(%s): %s\n", path,
#if	HAVE_DLOPEN
			(String)dlerror()
#elif	HAVE_SHL_LOAD
			(String)strerror(errno)
#elif	HAVE_LIBDLD
			"link failed"
#endif	/* HAVE_DLOPEN */
	);
  }

  return handle;
}

/*
 * Function Name:	fetch
 * Description:		fetch a symbol from loaded module
 * Arguments:		module handle
 * Return Value:	successful: symbol pointer, otherwise: NULL
 */
static void *
fetch(handle)
	void *	handle;
{
  void *	(*ret)(void);

#if	HAVE_DLOPEN
#if defined(USCORE) && !defined(DLSYM_ADDS_USCORE)
  String	usymb;

  if (asprintf(&usymb, "_%s", DLSYM_NAME) == -1) {
    wvSyslog(LOG_ERR, "module: %s\n", ERR_OUT_OF_MEMORY);
    return NULL;
  }
  ret = dlsym(handle, usymb);
  free((String)usymb);
#else
  ret = dlsym(handle, DLSYM_NAME);
#endif	/* USCORE */
  if (ret == NULL) {
    wvSyslog(LOG_ERR, "failed to get a symbol [%s]\n", DLSYM_NAME);
    return NULL;
  }
#elif	HAVE_SHL_LOAD
  int	status;

  errno	 = 0;
  status = shl_findsym((shl_t *)&handle, DLSYM_NAME, TYPE_PROCEDURE,
		       (int *)&ret);
  if (status == -1 && errno == 0) {
    status = shl_findsym((shl_t *)&handle, DLSYM_NAME, TYPE_DATA,
			 (int *)&ret);
  }
  if (status == -1) {
    wvSyslog(LOG_ERR, "failed to get a symbol [%s]\n", DLSYM_NAME);
    return NULL;
  }
#elif	HAVE_LIBDLD
#endif	/* HAVE_DLOPEN */

  return ret();
}

/*
 * Function Name:	install
 * Description:		install new module to module list
 * Arguments:		module info
 * Return Value:	successful: 0, otherwise: -1
 */
static int
install(module)
	Module *	module;
{
  ModuleList *	new;
  
  new = (ModuleList *)malloc(sizeof(ModuleList));
  if (new == NULL) {
    return RET_NG;
  }

  new->module = module;
  if (moduleList == NULL) {
    new->prev = new->next = NULL;
  }
  else {
    moduleList->prev = new;
    new->next	     = moduleList;
  }
  moduleList = new;

  return RET_OK;
}

/*
 * Function Name:	verify
 * Description:		verify loaded module
 * Arguments:		module
 * Return Value:	successful: 0, otherwise: -1
 */
static int
verify(module)
	Module *	module;
{
  if (module == NULL) {
    return RET_NG;
  }

  if (module->version != MODULE_MAGIC_NUMBER_MAJOR) {
    wvSyslog(LOG_ERR, "module: invalid version %d\n", module->version);
    return RET_NG;
  }

  if (module->name == NULL) {
    wvSyslog(LOG_ERR, "module: no module name\n");
    return RET_NG;
  }

  if (lookup(module->name) != NULL) {
    wvSyslog(LOG_ERR, "module: %s is already installed\n", module->name);
    return RET_NG;
  }

  if (module->info == NULL) {
    wvSyslog(LOG_ERR, "module: no module description\n");
    return RET_NG;
  }

  if (module->author == NULL) {
    wvSyslog(LOG_ERR, "module: no creator name\n");
    return RET_NG;
  }

  return RET_OK;
}

/*
 * Function Name:	unload
 * Description:		unload module
 * Arguments:		module handle
 * Return Value:	successful: 0, otherwise: -1
 */
static int
unload(handle)
	void *	handle;
{
  int	ret = 0;

  if (handle != NULL) {
#if	HAVE_DLOPEN
    ret = dlclose(handle);
#elif	HAVE_SHL_LOAD
    ret = shl_unload((shl_t)handle);
#elif	HAVE_LIBDLD
#endif	/* HAVE_DLOPEN */
  }

  return ret;
}

#if	0
/*
 * Function Name:	uninstall
 * Description:		delete a loaded module from module list
 * Arguments:		module name
 * Return Value:	successful: 0, otherwise: -1
 */
static int
uninstall(name)
	const String	name;
{
  register ModuleList *	mlp;
  int			ret;

  mlp = lookup(name);
  if (mlp == NULL) {
    return -1;
  }

  if (mlp->module->finish != NULL) {
    mlp->module->finish();
  }
  ret = unload(mlp->module->handle);
  if (ret == -1) {
    wvSyslog(LOG_ERR, "failed to unload the module: %s\n", mlp->module->name);
  }

  if (mlp->prev == NULL) {
    mlp->next->prev	= mlp->prev;
    moduleList		= mlp->next;
  }
  else if (mlp->next == NULL) {
    mlp->prev->next	= mlp->next;
  }
  else {
    mlp->next->prev	= mlp->prev;
    mlp->prev->next	= mlp->next;
  }
  free((ModuleList *)mlp);
  mlp = NULL;

  return 0;
}
#endif

/*
 * Function Name:	lookup
 * Description:		lookup module with type/subtype pair
 * Arguments:		module name
 * Return Value:	found: pointer, otherwise: NULL
 */
static ModuleList *
lookup(name)
	const String	name;
{
  register ModuleList *	mlp;

  for (mlp = moduleList; mlp != NULL; mlp = mlp->next) {
    if (wvStrcmp(mlp->module->name, name) == 0) {
#if	DEBUG
    fprintf(stderr, "\t[lookup] module = %s\n", mlp->module->name);
#endif	/* DEBUG */
      return mlp;
    }
  }

  return NULL;
}

