/*****************************************************************************
 *
 * FILE:	auth.c
 * DESCRIPTION:	authentication functions
 * DATE:	Mon, Sep 18 2000
 * UPDATE:	Sun, Dec 19 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(c) 2000-2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (c) 2000-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: auth.c,v 1.23.2.14 2004/12/19 13:55:28 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#if	!defined(USE_SSL)
#include "md5.h"
#endif	/* USE_SSL, USE_CONTENT_MD5 */
#include "wyvern.h"
#include "auth.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static FileMgr *	fileMgrTable = NULL;
static Auth *		authTable    = NULL;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static FileMgr *	fileMgrEntry(const String, const String, const String);
static void		fileMgrCleaner(void);
static void		authDirCleaner(AuthDir *);
static void		authListCleaner(AuthList *);
static AuthDir *	accessFileReader(File *);
static int		addressChecker(Message *, AuthList *);
static int		compareAddr(const String, const String);
static int		userChecker(Message *, AuthList *,
				int (*)(Message *, const String, const String));
static String		toMD5(const String);
static void		setBasicAuth(Message *);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

static FileMgr *
fileMgrEntry(name, access_path, passwd_path)
	const String	name;
	const String	access_path;
	const String	passwd_path;
{
  FileMgr *	new;
  struct stat	sbuf;
  time_t	access_mtime = (time_t)0;
  time_t	passwd_mtime = (time_t)0;

  if (lstat(access_path, &sbuf) == 0) {
    access_mtime = sbuf.st_mtime;
  }
  if (lstat(passwd_path, &sbuf) == 0) {
    passwd_mtime = sbuf.st_mtime;
  }

  new = (FileMgr *)malloc(sizeof(FileMgr));
  if (new == NULL) {
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
    return NULL;
  }
  new->username		= wvStrdup(name);
  new->waccess.path	= wvStrdup(access_path);
  new->waccess.mtime	= access_mtime;
  new->wpasswd.path	= wvStrdup(passwd_path);
  new->wpasswd.mtime	= passwd_mtime;
  new->next		= fileMgrTable;
  fileMgrTable		= new;

  return new;
}

static void
fileMgrCleaner(void)
{
  register FileMgr *	fmp;
  register FileMgr *	fmp_next;

  for (fmp = fileMgrTable; fmp != NULL; fmp = fmp_next) {
    fmp_next = fmp->next;
    wvFree((String)fmp->username);
    wvFree((String)fmp->waccess.path);
    wvFree((String)fmp->wpasswd.path);
    wvFree((FileMgr *)fmp);
  }
}

FileMgr *
fileMgrLookup(name)
	const String	name;
{
  register FileMgr *	fmp;

  if (name == NULL) { return NULL; }	/* NULL cheker */
  for (fmp = fileMgrTable; fmp != NULL; fmp = fmp->next) {
    if (wvStrncmp(fmp->username, name, strlen(name)) == 0) {
      return fmp;
    }
  }
  return NULL;
}

/*****************************************************************************/

int
authNew(name, access_path, passwd_path)
	const String	name;
	const String	access_path;
	const String	passwd_path;
{
  FileMgr *	fmp;
  Auth *	new;

  fmp = fileMgrLookup(name);
  if (fmp != NULL) {
    return RET_NG;
  }

  fmp = fileMgrEntry(name, access_path, passwd_path);
  if (fmp == NULL) {
    return RET_NG;
  }

  new = (Auth *)malloc(sizeof(Auth));
  if (new == NULL) {
    return RET_NG;
  }
  new->name	   = wvStrdup(name);
  new->file	   = fmp;
  new->authDir	   = accessFileReader(&fmp->waccess);
  new->next	   = authTable;
  authTable	   = new;

  return RET_OK;
}

void
authCleaner(void)
{
  register Auth *	ap;
  register Auth *	ap_next;

  fileMgrCleaner();
  for (ap = authTable; ap != NULL; ap = ap_next) {
    ap_next = ap->next;
    wvFree((String)ap->name);
    authDirCleaner(ap->authDir);
    wvFree((Auth *)ap);
  }
}

Auth *
authLookup(name)
	const String	name;
{
  register Auth *	ap;

  if (name == NULL) { return NULL; }	/* NULL cheker */
  for (ap = authTable; ap != NULL; ap = ap->next) {
    if (wvStrncmp(ap->name, name, strlen(name)) == 0) {
      return ap;
    }
  }
  return NULL;
}


int
authUpdate(ap)
	register Auth *	ap;
{
  struct stat	sbuf;

  if (lstat(ap->file->waccess.path, &sbuf) == -1) {
    return errno == ENOENT ? RET_CONT : RET_NG;
  }

  if (sbuf.st_mtime > ap->file->waccess.mtime) {
    authDirCleaner(ap->authDir);
    ap->authDir = accessFileReader(&ap->file->waccess);
  }

  return RET_OK;
}

/*
 * Function Name:	authDirCleaner
 * Description:		release memory area of AuthDir
 * Arguments:		pointer of AuthDir
 * Return Value:	none
 */
static void
authDirCleaner(authDir)
	register AuthDir *	authDir;
{
  register AuthDir *	adp;
  register AuthDir *	adp_next;

  for (adp = authDir; adp != NULL; adp = adp_next) {
    adp_next = adp->next;
    wvFree((String)adp->dir);
    authListCleaner(adp->user);
    authListCleaner(adp->addr);
    wvFree((AuthDir *)adp);
  }
}

/*
 * Function Name:	authListCleaner
 * Description:		release memory area of AuthList
 * Arguments:		pointer of AuthList
 * Return Value:	none
 */
static void
authListCleaner(authList)
	register AuthList *	authList;
{
  register AuthList *	alp;
  register AuthList *	alp_next;

  for (alp = authList; alp != NULL; alp = alp_next) {
    alp_next = alp->next;
    wvFree((String)alp->name);
    wvFree((AuthList *)alp);
  }
}

/*****************************************************************************/

/*
 * Function Name:	accessFileReader
 * Description:		read '.waccess' file to authenticate
 * Arguments:		waccess structure, file path for '.waccess'
 * Return Value:	successful: new pointer, otherwise: NULL
 */
static AuthDir *
accessFileReader(waccess)
	register File *	waccess;
{
#define	MAX_ARGS	(2)
#define	AC_KEY		(0)
#define	AC_VALUE	(1)
  register FILE *	fin;
  char			buf[BUFSIZ];
  AuthDir *		new = NULL;

  /* open '.waccess' file */
  fin = wvSafeFOpen(waccess->path, "r");
  if (fin == NULL) {
    return NULL;
  }

  /* parsing */
  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fin) != NULL) {
    String *	ap;
    String	args[MAX_ARGS];
    String	p = buf;
    int		n = 0;

    /* skip comment lines */
    if (wvSkipLine(p) == 0) {
      continue;
    }
    for (ap = args; (*ap = strsep(&p, " \t\r\n")) != NULL; ) {
      if (**ap != EOL) {
	n++;
	if (++ap >= &args[MAX_ARGS]) {
	  break;
	}
      }
    }

    if (wvStrcasecmp(args[AC_KEY], "AuthDir") == 0) {
      AuthDir *	adp = (AuthDir *)malloc(sizeof(AuthDir));
      if (adp == NULL) {
	break;	/* exit while loop */
      }
      adp->dir	= wvStrdup(args[AC_VALUE]);
      adp->user	= NULL;
      adp->addr	= NULL;
      adp->next	= new;
      new	= adp;
    }
    else if (wvStrcasecmp(args[AC_KEY], "AuthUser") == 0) {
      AuthList * alp = (AuthList *)malloc(sizeof(AuthList));
      if (alp == NULL) {
	break;	/* exit while loop */
      }
      alp->name	= wvStrdup(args[AC_VALUE]);
      alp->next	= new->user;
      new->user	= alp;
    }
    else if (wvStrcasecmp(args[AC_KEY], "AuthAddr") == 0) {
      AuthList * alp = (AuthList *)malloc(sizeof(AuthList));
      if (alp == NULL) {
	break;	/* exit while loop */
      }
      alp->name	= wvStrdup(args[AC_VALUE]);
      alp->next	= new->addr;
      new->addr	= alp;
    }
    else {	/* illegal type */
      wvSyslog(LOG_WARNING, "Illegal authentication directive: %s\n",
	       args[AC_KEY]);
    }
    memset(buf, 0, sizeof(buf));
  }
  fclose(fin);

  return new;
}

/*****************************************************************************/

int
authChecker(msg, adp, passwordChecker)
	register Message *	msg;
	register AuthDir *	adp;
	int			(*passwordChecker)(Message *, const String,
						   const String);
{
  /*
   * first, check IP address/Domain name,
   * next, try whether user can access the resource by username
   */
  if (addressChecker(msg, adp->addr) == RET_NG &&
      userChecker(msg, adp->user, passwordChecker) == RET_NG) {
#if	DEBUG
    fprintf(stderr, "[mod_auth]: %s authentication NG\n", msg->user.name);
#endif	/* DEBUG */
    msg->response.status = STATUS_UNAUTHORIZED;
    setBasicAuth(msg);
    return RET_NG;
  }

  return RET_OK;
}


/*
 * Function Name:	addressChecker
 * Description:		check IP address/domain name to access
 * Arguments:		message, AuthList
 * Return Value:	allow: 0, otherwise: -1
 */
static int
addressChecker(msg, addrList)
	register Message *	msg;
	register AuthList *	addrList;
{
  register AuthList *	alp;

  for (alp = addrList; alp != NULL; alp = alp->next) {
    if (compareAddr(msg->conn->client.name, alp->name) == RET_OK ||
	compareAddr(msg->conn->client.ip,   alp->name) == RET_OK) {
      return RET_OK;
    }
  }

  return RET_NG;
}

/*
 * Function Name:	compareAddr
 * Description:		check the hostname, IP addr and the access list
 * Arguments:		Hostname/IP addr, compared one
 * Return Value:	Matching: 0, Otherwise: -1
 */
static int
compareAddr(name, pat)
	const String	name;
	const String	pat;
{
  size_t	nlen;
  size_t	plen;

  if (name == NULL || pat == NULL) {
    return RET_NG;
  }

  nlen = strlen(name);
  plen = strlen(pat);
  if (nlen < plen) {
    return RET_NG;
  }

  /*
   * check to match the name with the pattern like as '.ad.jp'
   */
  if (pat[0] == '.') {
    if (nlen >= plen) {
      size_t	offset = nlen - plen;
      if (wvStrncmp(name + offset, pat, plen) == 0) {
	return RET_OK;
      }
    }
  }
  /*
   * check to match the name with the pattern like as '133.24.' or IP addr
   */
  else if (pat[plen - 1] == '.') {
    if (wvStrncmp(name, pat, plen) == 0) {
      return RET_OK;	/* match! */
    }
  }
  /*
   * check to match the name with the pattern of FQDN 
   */
  else {
    if (wvStrncmp(name, pat, nlen) == 0) {
      return RET_OK;	/* match! */
    }
  }

  return RET_NG;	/* no match */
}

/*
 * Function Name:	userChecker
 * Description:		check password of user to access
 * Arguments:		message, userList, passwordChecker
 * Return Value:	allow: 0, otherwise: -1
 */
static int
userChecker(msg, userList, passwordChecker)
	register Message *	msg;
	register AuthList *	userList;
	int			(*passwordChecker)(Message *, const String,
						   const String);
{
  String	authorization;
  String	scheme;
  String	param;
  int		ret = RET_NG;

  authorization = headerLookup(msg->request.headers, "AUTHORIZATION");
  if (authorization == NULL) {
    return RET_NG;
  }

  /* divide 'authorization' into 'scheme' and 'param' */
  scheme = authorization;
  param	 = (String)strchr(authorization, SPC);
  if (param == NULL) {
    return RET_NG;
  }
  *param++ = EOL;

  /* basic authentication scheme */
  if (wvStrcasecmp(scheme, "BASIC") == 0) {
#define	MAX_AUTH_ARGS	(2)
#define	PW_USERNAME	(0)
#define	PW_PASSWORD	(1)
    String *		ap;
    String		args[MAX_AUTH_ARGS];
    String		p = decodeBase64(param);
    register AuthList *	alp;

    /* set username and password */
    for (ap = args; (*ap = strsep(&p, ":")) != NULL; ) {
      if (**ap != EOL) {
	if (++ap >= &args[MAX_AUTH_ARGS]) {
	  break;
	}
      }
    }

    for (alp = userList; alp != NULL; alp = alp->next) {
      if (alp->name == NULL) {
	continue;
      }
      if (wvStrncmp(args[PW_USERNAME], alp->name, strlen(alp->name)) == 0) {
	String		md5password = toMD5(args[PW_PASSWORD]);
#if	DEBUG
	fprintf(stderr, "[userChecker] %s(%s) md5(%s)\n",
		args[PW_USERNAME], args[PW_PASSWORD], md5password);
#endif	/* DEBUG */
	ret = (*passwordChecker)(msg, args[PW_USERNAME], md5password);
	wvFree((String)md5password);
	break;
      }
    } /* end of for */
    wvFree((String)p);
  }
  /* unknown authentication scheme */
  else {
    msg->response.status = STATUS_UNAUTHORIZED;
  }

  return ret;
}

static String
toMD5(s)
	const String	s;
{
  extern void	MD5Init(MD5_CTX *);
  extern void	MD5Update(MD5_CTX *, unsigned char *, unsigned int);
  extern void	MD5Final(unsigned char [MD5_DIGEST_LENGTH], MD5_CTX *);

  MD5_CTX	md5_ctx;
  unsigned char	md5[MD5_DIGEST_LENGTH];
  unsigned char	md5_hex[MD5_DIGEST_LENGTH*2+1];
  int		i;

  if (s == NULL) { return NULL; }	/* NULL cheker */
  MD5Init(&md5_ctx);
  MD5Update(&md5_ctx, s, strlen(s));
  MD5Final(md5, &md5_ctx);

  memset(md5_hex, 0, sizeof(md5_hex));
  for (i = 0; i < MD5_DIGEST_LENGTH; i++) {
    (void)snprintf(md5_hex + i * 2, sizeof(md5_hex) - 1, "%02x", md5[i]);
  }

  return wvStrdup(md5_hex);
}

/****************************************************************************/

/*
 * Function Name:	setBasicAuth
 * Description:		set WWW-Authenticate field for Basic authentication
 * Arguments:		message
 * Return Value:	none
 */
static void
setBasicAuth(msg)
	register Message *	msg;
{
  String	www_authenticate;
  Headers *	hp;

  if (asprintf(&www_authenticate, "Basic realm=\"%s%%%s\"",
		msg->user.name == NULL ? "anonymous" : msg->user.name,
		msg->vhost->serverName) != -1) {
    hp = responseHeaderEntry(&msg->response.headers,
			      "WWW-Authenticate", www_authenticate);
    free((String)www_authenticate);
    if (hp != NULL) {
      return;
    }
  }

  wvSyslog(LOG_ERR, "[basic_auth] failed to allocate memory\n");
  msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
}

