/*****************************************************************************
 *
 * FILE:	mode_access.c
 * DESCRIPTION:	access control module for Wyvern
 * DATE:	Sat, Sep 16 2000
 * UPDATE:	Sun, Dec 19 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(c) 2000-2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (c) 2000-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_access.c,v 1.22.2.13 2004/12/19 13:55:29 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include "wyvern.h"
#include "wpath.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	SPECIAL_KEYWORD		"DEFAULT"

typedef enum {
  PERM_DENY  = -1,
  PERM_ALLOW = 0
} permission_t;

typedef struct _Restrict {
  String		host;
  permission_t		access;
  struct _Restrict *	next;
} Restrict;


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static Restrict *	accessHostTable = NULL;
static permission_t	defaultAccess;	/* default access permission */


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int		accessChecker(register Message *);
static void		accessInit(void);
static void		accessFinish(void);
static void		accessHostEntry(FILE *);
static Restrict *	lookup(const String);
static bool_t		compareAddr(const String, const String);
static void		accessRegisterHooks(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	accessInit
 * Description:		initialize the host access list
 * Arguments:		none
 * Return Value:	none
 */
static void
accessInit(void)
{
  FILE *	fin;

  /* register allowed access hosts */
  fin = wvSafeFOpen(WYVERN_HOSTS_ALLOW, "r");
  if (fin != NULL) {
    defaultAccess = PERM_ALLOW;	/* allow all hosts to access */
    accessHostEntry(fin);
    fclose(fin);
  }
}

/*
 * Function Name:	accessFinish
 * Description:		release the memory area of access lists
 * Arguments:		none
 * Return Value:	none
 */
void
accessFinish(void)
{
  register Restrict *	rp;
  register Restrict *	rp_next;

  for (rp = accessHostTable; rp != NULL; rp = rp_next) {
    rp_next = rp->next;
    wvFree((Restrict *)rp);
  }
  accessHostTable = NULL;
}


/*
 * Function Name:	accessChecker
 * Description:		permit the access to the host
 * Arguments:		message
 * Return Value:	allowed host: 0, denied host: -1
 */
static int
accessChecker(msg)
	register Message *	msg;
{
  struct sockaddr *	cliaddr;
  char			hbuf[NI_MAXHOST];
  register Restrict *	rp;

  /* no table entry: allow all hosts */
  if (accessHostTable == NULL) {
    return RET_OK;
  }
#if	DEBUG
  fputs("<accessChecker>\n", stderr);
#endif	/* DEBUG */

  /* get client hostname with DNS */
  cliaddr = (struct sockaddr *)&msg->conn->client.addr;
  if (getnameinfo(cliaddr, cliaddr->sa_len, hbuf, sizeof(hbuf), NULL, 0,
      NI_NAMEREQD) == 0) {
    wvFree((String)msg->conn->client.name);
    msg->conn->client.name = wvStrdup(hbuf);
#if	DEBUG
    fprintf(stderr, "[accessChecker] %s(%s)\n",
	    msg->conn->client.name, msg->conn->client.ip);
#endif	/* DEBUG */
  }

  /* check the client hostname */
  rp = lookup(msg->conn->client.name);
  if (rp != NULL && compareAddr(msg->conn->client.name, rp->host) == TRUE) {
    return (int)rp->access;
  }

  /* check the client IP address */
  rp = lookup(msg->conn->client.ip);
  if (rp != NULL && compareAddr(msg->conn->client.ip, rp->host) == TRUE) {
    return (int)rp->access;
  }

  return (int)defaultAccess;
}

/*****************************************************************************/

/*
 * Function Name:	accessHostEntry
 * Description:		register host to the host access list
 * Arguments:		FILE pointer
 * Return Value:	none
 */
static void
accessHostEntry(fin)
	register FILE *	fin;
{
#define	MAX_ARGS	(2)
#define	AC_NAME		(0)	/* hostname */
#define	AC_PERMISSION	(1)	/* permission */
  char	buf[MAX_HOSTNAME_LEN + 1];

  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fin) != NULL) {
    String *		ap;
    String		args[MAX_ARGS];
    String		p = buf;
    permission_t	perm;
    Restrict *		rp;

    /* skip comment lines */
    if (wvSkipLine(p) == 0) {
      continue;
    }
    for (ap = args; (*ap = strsep(&p, " \t\r\n")) != NULL; ) {
      if (**ap != EOL) {
	if (++ap >= &args[MAX_ARGS]) {
	  break;
	}
      }
    }
    if (args[AC_NAME] == NULL || args[AC_PERMISSION] == NULL) {
      continue;
    }

    if (wvStrcasecmp(args[AC_PERMISSION], "allow") == 0) {
      perm = PERM_ALLOW;
    }
    else if (wvStrcasecmp(args[AC_PERMISSION], "deny") == 0) {
      perm = PERM_DENY;
    }

    if (wvStrcasecmp(SPECIAL_KEYWORD, args[AC_NAME]) == 0) {
#if	DEBUG
      fprintf(stderr, "[mod_access]: default access = %s\n",
	      perm == PERM_ALLOW ? "allow" : "deny");
#endif	/* DEBUG */
      defaultAccess = perm;
    }
    else if (lookup(args[AC_NAME]) == NULL) {
      rp		= (Restrict *)wvMalloc(sizeof(Restrict));
      rp->host		= wvStrdup(args[AC_NAME]);
      rp->access	= perm;
      rp->next		= accessHostTable;
      accessHostTable	= rp;
    }
    memset(buf, 0, sizeof(buf));
  } /* while end */
}

/*
 * Function Name:	lookup
 * Description:		check host in the host access list
 * Arguments:		host
 * Return Value:	host or NULL
 */
static Restrict *
lookup(host)
	const String	host;
{
  register Restrict *	rp;

  for (rp = accessHostTable; rp != NULL; rp = rp->next) {
    if (rp->host == NULL) {
      continue;
    }
    if (wvStrncmp(rp->host, host, strlen(rp->host)) == 0) {
      return rp;
    }
  }

  return NULL;
}

/*
 * Function Name:	compareAddr
 * Description:		check the hostname, IP addr and the access list
 * Arguments:		Hostname/IP addr, compared one
 * Return Value:	Matching: TRUE, Otherwise: FALSE
 */
static bool_t
compareAddr(name, pat)
	const String	name;
	const String	pat;
{
  size_t	nlen;
  size_t	plen;

  if (name == NULL || pat == NULL) {
    return FALSE;
  }

  nlen = strlen(name);
  plen = strlen(pat);
  if (nlen < plen) {
    return FALSE;
  }

  /* check to match the name with the pattern like as '.ad.jp' */
  if (pat[0] == '.') {
    if (nlen >= plen) {
      int	offset = nlen - plen;
      if (wvStrncmp(name + offset, pat, plen) == 0) {
	return TRUE;
      }
    }
  }
  /* check to match the name with the pattern like as '133.24.' or IP addr */
  else if (pat[plen-1] == '.') {
    if (wvStrncmp(name, pat, plen) == 0) {
      return TRUE;	/* match! */
    }
  }
  /* check to match the name with the pattern of FQDN */
  else {
    if (wvStrncmp(name, pat, nlen) == 0) {
      return TRUE;	/* match! */
    }
  }

  return FALSE;	/* no match */
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
accessRegisterHooks(void)
{
  wvHookAccessChecker(accessChecker, HOOK_MIDDLE);
}

Module	mod_access = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_STANDARD,		/* safety */
  "mod_access",			/* name */
  "Access Control Module version 1.0.1",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  (void *)accessInit,		/* init */
  (void *)accessFinish,		/* finish */
  (void *)accessRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_access;
}
