/*****************************************************************************
 *
 * FILE:	mod_auth.c
 * DESCRIPTION:	authentication module for wyvern
 * DATE:	Sun, Sep 17 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2000-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_auth.c,v 1.30.2.3 2003/03/18 16:06:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#include "wyvern.h"
#include "auth.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of global functions
 *
 *****************************************************************************/
extern bool_t	allowUser(Thread *);


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	SYSTEM_AUTH	ALLOW_USERS_TABLE	/* system auth index */


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static AuthTable *	authTable[ALLOW_USERS_TABLE + 1];


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void		readSystemAuthFile(void);
static AuthTable *	readUserAuthFile(Thread *);
static AuthTable *	checkAuthTable(String);
static void		setBasicAuth(Thread *);
static void		freeAuthTable(AuthTable *);
static unsigned int	hash(String);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	_init
 * Description:		initialize server's authentication list
 *			(automatic loading)
 * Arguments:		none
 * Return Value:	none
 */
void
_init(void)
{
  int	i;

#if	DEBUG
  fputs("[mod_auth]: plugged-in\n", stderr);
#endif	/* DEBUG */

  /* initialize authentication cache */
  for (i = 0; i < ALLOW_USERS_TABLE + 1; i++) {
    authTable[i] = NULL;
  }

  readSystemAuthFile();
}

/*
 * Function Name:	_fini
 * Description:		free memory area of server's authentication list
 *			(automatic loading)
 * Arguments:		none
 * Return Value:	none
 */
void
_fini(void)
{
  int	i;

  for (i = 0; i < ALLOW_USERS_TABLE + 1; i++) {
    if (authTable[i] == NULL) {
      continue;
    }
    freeAuthTable(authTable[i]);
  }
}

/*
 * Function Name:	readSystemAuthFile
 * Description:		read system authentication file and set it to authTable
 * Arguments:		none
 * Return Value:	none
 */
static void
readSystemAuthFile(void)
{
  AuthTable *	auth;

  /* for server self */
  auth = (AuthTable *)xcalloc(1, sizeof(AuthTable));
  if (auth == NULL) {
    die(ERR_OUT_OF_MEMORY);
  }

  /* read 'ServerRoot/etc/.waccess' */
  if (xasprintf(&auth->waccess.path, "%s/%s",
		httpd->serverRoot, GLOBAL_HOSTS_ACCESS) == -1) {
    die(ERR_OUT_OF_MEMORY);
  }
  readWAccess(&auth->waccess);

  /* read 'ServerRoot/etc/.wpasswd' */
  if (xasprintf(&auth->wpasswd.path, "%s/%s",
		httpd->serverRoot, GLOBAL_HOSTS_PASSWD) == -1) {
    die(ERR_OUT_OF_MEMORY);
  }
  readWPasswd(&auth->wpasswd);
  auth->name			= xstrdup(SYSTEM_AUTHNAME);/* set server name */
  auth->next			= NULL;
  authTable[SYSTEM_AUTH]	= auth;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	allowUser
 * Description:		check user access
 * Arguments:		thread
 * Return Value:	allow: TRUE, otherwise: FALSE
 */
bool_t
allowUser(thd)
	register Thread *	thd;
{
  AuthTable *		auth;
  register AuthDir *	ap;

  /* server level authentication */
  if (thd->userName == NULL) {
    auth = authTable[SYSTEM_AUTH];
  }
  /* user level authentication */
  else {
    auth = checkAuthTable(thd->userName);
  }

  if (auth == NULL) {	/* read user authentication file */
    auth = readUserAuthFile(thd);
    if (auth == NULL) {
      return FALSE;	/* error; access denied */
    }
  }
  else {
    struct stat		waccess;
    struct stat		wpasswd;

    if (lstat(auth->waccess.path, &waccess) != -1 &&
	lstat(auth->wpasswd.path, &wpasswd) != -1) {
      if (waccess.st_mtime > auth->waccess.mtime) {
	freeAuthDirList(auth->waccess.authDir);
	readWAccess(&auth->waccess);
      }
      if (wpasswd.st_mtime > auth->wpasswd.mtime) {
	freeAuthMgrList(auth->wpasswd.authMgr);
	readWPasswd(&auth->wpasswd);
      }
    }
    else if (errno == ENOENT) {
      return TRUE;
    }
  }

  for (ap = auth->waccess.authDir; ap != NULL; ap = ap->next) {
    if (strstr(thd->requestURI, ap->dir) != NULL) {	/* found */
      /*
       * first, check IP address/Domain name,
       * next, try whether user can access the resource by username
       */
      if (checkAddress(thd, ap->allow) == FALSE &&
	  checkPassword(thd, auth->wpasswd.authMgr) == FALSE) {
	setResponseCode(thd, CODE_UNAUTHORIZED);
	if (ap->allow->method == AUTH_BASIC) {
	  setBasicAuth(thd);
	}
#if	DEBUG
	fprintf(stderr, "[mod_auth]: %s authentication NG\n", thd->userName);
#endif	/* DEBUG */
	return FALSE;
      }
      else {
#if	DEBUG
	fprintf(stderr, "[mod_Auth]: %s authentication OK\n", thd->userName);
#endif	/* DEBUG */
	return TRUE;
      }
    }
  }

#if	DEBUG
  fprintf(stderr, "[mod_AUTH]: %s authentication OK\n", thd->userName);
#endif	/* DEBUG */
  return TRUE;
}


/*
 * Function Name:	readUserAuthFile
 * Description:		read user authentication file and set it to authTable
 * Arguments:		thread
 * Return Value:	none
 */
static AuthTable *
readUserAuthFile(thd)
	Thread *	thd;
{
  unsigned int	h;
  AuthTable *	auth;

#if	DEBUG
  fprintf(stderr, "[mod_auth]: reading authentication file\n");
#endif	/* DEBUG */

  auth = (AuthTable *)xcalloc(1, sizeof(AuthTable));
  if (auth == NULL) {
    wvSyslog(LOG_ERR, "failed to allocate memory to read %s's auth file",
	      thd->userName);
    return NULL;	/* write error log before return */
  }

  h		= hash(thd->userName);
  auth->name	= xstrdup(thd->userName);	/* set username */
#if	DEBUG
  fprintf(stderr, "[mod_auth]: name = %s\n", thd->userName);
#endif	/* DEBUG */

  /* read 'UserDir/.waccess' */
  if (xasprintf(&auth->waccess.path, "%s/%s/%s",
		thd->userHome, httpd->vhost->userDir, USER_ACCESS) == -1) {
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
  }
  readWAccess(&auth->waccess);
#if	DEBUG
  fprintf(stderr, "[mod_auth]: waccess = %s : %s",
	  auth->waccess.path, ctime(&auth->waccess.mtime));
#endif	/* DEBUG */

  /* read 'UserDir/.wpasswd' */
  if (xasprintf(&auth->wpasswd.path, "%s/%s/%s",
		thd->userHome, httpd->vhost->userDir, USER_PASSWD) == -1) {
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
  }
  readWPasswd(&auth->wpasswd);
#if	DEBUG
  fprintf(stderr, "[mod_auth]: wpasswd = %s : %s",
	  auth->wpasswd.path, ctime(&auth->wpasswd.mtime));
#endif	/* DEBUG */
  auth->next	= authTable[h];
  authTable[h]	= auth;

  return auth;
}

/*
 * Function Name:	checkAuthTable
 * Description:		check an username in the auth table
 * Arguments:		username
 * Return Value:	found: pointer of table, otherwise: NULL
 */
static AuthTable *
checkAuthTable(name)
	String	name;
{
  if (name != NULL) {
    register AuthTable *	atp;

    for (atp = authTable[hash(name)]; atp != NULL; atp = atp->next) {
      if (strcmp(atp->name, name) == 0) {
	return atp;
      }
    }
  }
  return NULL;
}


/*
 * Function Name:	setBasicAuth
 * Description:		set WWW-Authenticate field for Basic authentication
 * Arguments:		thread
 * Return Value:	none
 */
static void
setBasicAuth(thd)
	register Thread *	thd;
{
  if (xasprintf(&thd->response.www_authenticate, "Basic realm=\"%s%%%s\"",
		thd->userName == NULL ? "anonymous" : thd->userName,
		httpd->vhost->serverName) == -1) {
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	freeAuthTable
 * Description:		release memory area of Auth's and AuthDir's
 * Arguments:		pointer of AuthTable
 * Return Value:	none
 */
static void
freeAuthTable(ap)
	register AuthTable *	ap;
{
  XFree((String)ap->name);
  XFree((String)ap->wpasswd.path);
  XFree((String)ap->waccess.path);
  freeAuthMgrList(ap->wpasswd.authMgr);
  freeAuthDirList(ap->waccess.authDir);
  ap->waccess.path	= NULL;
  ap->waccess.authDir	= NULL;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	hash
 * Description:		calculates a hash value of string
 * Arguments:		string
 * Return Value:	hash value
 */
static unsigned int
hash(str)
	register String	str;
{
  unsigned int	h;

  for (h = 7; *str != EOL; str++) {
    h = (67 * h + *str) % ALLOW_USERS_TABLE;
  }

  return h;
}

