/*****************************************************************************
 *
 * FILE:	mod_autoindex.c
 * DESCRIPTION:	directory indexing module for wyvern
 * DATE:	Mon, Aug 26 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2002-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_autoindex.c,v 1.1.2.16 2003/03/18 16:06:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_DIRENT_H
#include <dirent.h>
#else
#define	dirent	direct
#if	HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif	/* HAVE_SYS_NDIR_H */
#if	HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif	/* HAVE_SYS_DIR_H */
#if	HAVE_NDIR_H
#include <ndir.h>
#endif	/* HAVE_NDIR_H */
#endif	/* HAVE_DIRENT_H */
#ifdef HAVE_D_NAMLEN
#define NAMLEN(d)	(d->d_namlen)
#else	/* !HAVE_D_NAMLEN */
#define NAMLEN(d)	(strlen(d->d_name))
#endif	/* !HAVE_D_NAMLEN */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	DIR_TABLE_SIZE	(17)	/* directory indexing cache */
#define	MAX_DIR_ENTRY	(1024)

#define	U_KILO		(1024L)
#define	U_MEGA		(1024L*1024L)
#define	U_GIGA		(1024L*1024L*1024L)

#define	HEADER_FILE	".HEADER"
#define	FOOTER_FILE	".FOOTER"

typedef struct _DirIndex {	/* directory cache structure */
  String		path;	/* path of data file */
  String		name;	/* path of cache file */
  time_t		mtime;	/* last modified time */
  struct _DirIndex *	next;
  struct _DirIndex *	prev;
} DirIndex;


#define TOTAL_KEYWORDS	(47)
#define MIN_WORD_LENGTH (2)
#define MAX_WORD_LENGTH (6)
#define MIN_HASH_VALUE	(0)
#define MAX_HASH_VALUE	(31)
/* maximum key range = 32, duplicates = 25 */

typedef enum {		/* file form */
  FF_COMPRESSED,	/* compressed */
  FF_TAR,		/* tar archived */
  FF_ARCHIVE,		/* archived */
  FF_BINARY,		/* binary */
  FF_HTML,		/* HTML */
  FF_TEXT,		/* text */
  FF_PDF,		/* PDF */
  FF_DVI,		/* DVI */
  FF_POSTSCRIPT,	/* PostScript */
  FF_C,			/* C language */
  FF_F77,		/* Fortran language */
  FF_PERL,		/* Perl language */
  FF_TEX,		/* TeX/LaTeX */
  FF_PATCH,		/* patch file */
  FF_SHELL,		/* shell script */
  FF_IMAGE,		/* image */
  FF_MOVIE,		/* movie */
  FF_SOUND		/* sound/audio */
} fform_t;

typedef struct fileForm {
  String	name;
  fform_t	form;
} FileForm;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int		directoryIndexor(Message *);
static int		cacheOpen(Message *, DirIndex *);
static int		cacheUpdate(Message *, DirIndex *);
static void		makeContent(Message *, FILE *, const String *, int);
static void		includeFile(const String, const String, FILE *);
static void		printFileForm(const String, mode_t, FILE *);
static unsigned int	hash(const char *, unsigned int);
static const FileForm *	fformGet(const char *, unsigned int);
static void		quickSort(String *, int, int);
static void		swap(String *, int, int);
static void		cacheInit(void);
static void		cacheFree(void);
static void		cacheInsert(DirIndex *, DirIndex *);
static DirIndex *	cacheRemove(DirIndex *);
static void		autoindexRegisterHooks(void);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static DirIndex	dirCache[DIR_TABLE_SIZE];	/* cache entry */
static DirIndex	dirList;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	directoryIndexor
 * Description:		file listing in the directory
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
directoryIndexor(msg)
	register Message *	msg;
{
  register DirIndex *	dp;

#if	DEBUG
  fputs("<directoryIndexor>\n", stderr);
#endif	/* DEBUG */

  if (msg->vhost->autoIndexing == FALSE) {
    msg->response.status = STATUS_FORBIDDEN;
    return RET_NG;
  }

  /* obtain directory information */
  (void)stat(msg->file.path, &msg->file.status);

  /* search cache (dirList) */
  for (dp = dirList.next; dp != &dirList; dp = dp->next) {
    if (dp->mtime == 0) {
      continue;
    }
    if (wvStrcmp(msg->file.path, dp->path) == 0) {	/* cache hit */
      if (dp->mtime == msg->file.status.st_mtime) {
#if	DEBUG
	fputs("[DEBUG] cache hit!\n", stderr);
#endif	/* DEBUG */
	return cacheOpen(msg, dp);
      }
      else {	/* directory was modified. */
#if	DEBUG
	fputs("[DEBUG] cache update!\n", stderr);
#endif	/* DEBUG */
	return cacheUpdate(msg, dp);
      }
    }
#if	DEBUG
    fprintf(stderr, "[DEBUG] checking %s = %s\n",
	    msg->file.path, dp->path);
#endif	/* DEBUG */
  }

  /* get top entry from list, and move one to end of list */
  dp = cacheRemove(dirList.next);
  cacheInsert(&dirList, dp);

  /* register new entry */
  wvFree(dp->path);
  dp->path = wvStrdup(msg->file.path);

  wvFree(dp->name);
  if (asprintf(&dp->name,  "%s/dir_cache%lx%lx%lx",
		httpd->cacheDir,
		(long)msg->file.status.st_ino,
		(long)msg->file.status.st_size,
		msg->file.status.st_mtime) == -1) {
    dp->mtime = 0;	/* mark of no use */
    wvSyslog(LOG_ERR, "[mod_autoindex] failed to allocate memory\n");
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    return RET_NG;
  }
  dp->mtime = msg->file.status.st_mtime;

#if	DEBUG
  fputs("[DEBUG] no cache ... create new\n", stderr);
#endif	/* DEBUG */
  return cacheUpdate(msg, dp);
}

/*****************************************************************************/

/*
 * Function Name:	cacheOpen
 * Description:		open cache file
 * Arguments:		message, cache
 * Return Value:	successful: 0, otherwise: -1
 */
static int
cacheOpen(msg, cache)
	register Message *	msg;
	DirIndex *		cache;
{
  int		fd;
  struct stat	sbuf;

  /* open directory listed file */
  fd = wvSafeOpen(cache->name, O_RDONLY, &sbuf, FALSE);
  if (fd == -1) {
    wvSyslog(LOG_ERR, "[mod_autoindex] failed to open file: %s\n", cache->name);
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    return RET_NG;
  }

  /* close previous opened file */
  if (msg->file.fd > 0) {
    wvClose(msg->file.fd);
  }

  msg->file.fd			= fd;
  msg->file.status.st_size	= sbuf.st_size;
  msg->response.content_length	= sbuf.st_size;
  msg->response.status		= STATUS_REQUEST_OK;
  msg->file.mime		= HTML_MAGIC_TYPE;

  return RET_OK;
}

/*
 * Function Name:	cacheUpdate
 * Description:		update cache file for directory indexing
 * Arguments:		message, cache
 * Return Value:	successful: 0, otherwise: -1
 */
static int
cacheUpdate(msg, cache)
	register Message *	msg;
	register DirIndex *	cache;
{
  register DIR *	dfd;
  struct dirent *	dp;
  struct stat		sbuf;
  size_t		items;	/* total entries in directory */
  String *		dir;
  int			dnum = 0;
  register FILE *	fout;
  int			i;
  int			ret = 0;

  /*
   * directory walking
   */
  dfd = opendir(msg->file.path);
  if (dfd == NULL) {
    msg->response.status = STATUS_FORBIDDEN;
    return RET_NG;
  }

  /*
   * estimate total entries in directory
   */
#ifdef HAVE_DIRFD
  if (fstat(dirfd(dfd), &sbuf) == -1) {
#else
  if (fstat(dfd->dd_fd, &sbuf) == -1) {
#endif	/* !HAVE_DIRFD */
    closedir(dfd);
    wvSyslog(LOG_ERR, "[mod_autoindex] failed to get file status\n");
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    return RET_NG;
  }
  items = (sbuf.st_size / 24);
  dir	= (String *)calloc(items, sizeof(String));
  if (dir == NULL) {
    closedir(dfd);
    return wvNoMemoryError(msg);
  }

  /*
   * directory scanning ...
   */
  while ((dp = readdir(dfd)) != NULL) {
    register String	p;

    /* skip dot files */
    if (dp->d_name[0] == '.' || dnum > MAX_DIR_ENTRY - 1) {
      continue;
    }

    /* skip CVS/RCS directory */
    if (wvStrcmp(dp->d_name, "CVS") == 0 ||
	wvStrcmp(dp->d_name, "RCS") == 0) {
      continue;
    }
    /* skip *,? and *~ files (for CVS and emacs series) */
    if (dp->d_name[NAMLEN(dp) - 2] == ',' ||
	dp->d_name[NAMLEN(dp) - 1] == '~') {
      continue;
    }
    /* skip #* and ,* files (for MH) */
    if (dp->d_name[0] == '#' || dp->d_name[0] == ',') {
      continue;
    }
    /* skip *.{bak,org} files (for vim series) */
    p = strrchr(dp->d_name, '.');
    if (p != NULL) {
      if (wvStrcasecmp(p, ".bak") == 0 ||
	  wvStrcasecmp(p, ".org") == 0 || wvStrcasecmp(p, ".orig") == 0) {
	continue;
      }
    }

    dir[dnum] = wvStrdup(dp->d_name);
    if (dir[dnum] == NULL) {
      closedir(dfd);
      ret = wvNoMemoryError(msg);;
      goto error;
    }
    dnum++;
  }
  closedir(dfd);

  /*
   * sort directory indexing
   */
  quickSort(dir, 0, dnum - 1);

  /*
   * update cache file
   */
  fout = wvSafeFOpen(cache->name, "w");
  if (fout == NULL) {
    wvSyslog(LOG_ERR, "[mod_autoindex] failed to open file: %s\n", cache->name);
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    ret = RET_NG;
    goto error;
  }
  makeContent(msg, fout, dir, dnum);
  fclose(fout);

  /* bind new file descriptor to current message */
  ret = cacheOpen(msg, cache);

error:
  for (i = 0; i < dnum; i++) {
    wvFree((String)dir[i]);
  }
  wvFree((String *)dir);

  return ret;
}

/*
 * Function Name:	makeContent
 * Description:		make a directory listing HTML file
 * Arguments:		message, output file pointer, directory name,
 *			# of entry in directory
 * Return Value:	none
 */
static void
makeContent(msg, fout, dir, dnum)
	register Message *	msg;
	register FILE *		fout;
	const String *		dir;
	int			dnum;
{
  int	i;

  /* make HTML text */
  fputs(
    "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n"
    "<html lang=\"ja\">\n"
    "<head>\n"
    "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=EUC-JP\">\n"
    "<meta http-equiv=\"Content-Style-Type\" content=\"text/css\">\n"
    "<style type=\"text/css\">\n"
    "<!--\n"
    ".size_b { text-align: right; padding-right: 20px; color: orange }\n"
    ".size_k { text-align: right; padding-right: 20px; color: green }\n"
    ".size_m { text-align: right; padding-right: 20px; color: red }\n"
    ".size_g { text-align: right; padding-right: 20px; color: purple }\n"
    ".text { color: brown }\n"
    "span { color: #ff0033 }\n"
    "table { width: 90%; margin-left: 5%; margin-right: 5%; background-color: #a0b8c8 }\n"
    "tbody { max-height: 70%; min-height: 300px; overflow: auto }\n"
    "tr.none { background-color: transparent }\n"
    "td.timestamp { text-align: center }\n"
    "td.size { text-align: right }\n"
    "tr { background-color: #dcdcdc }\n"
    "td { color: #000033 }\n"
    "-->\n"
    "</style>\n",
    fout);

  fprintf(fout,
	"<title>Directory list of %s</title>\n"
	"</head>\n"
	"<body lang=\"ja\">\n"
	"<h1>Directory list of <span>%s</span></h1>\n",
	msg->request.uri, msg->request.uri);

  /* include HeaderFile if it exists */
  includeFile(msg->file.path, HEADER_FILE, fout);

  fprintf(fout,
	"<div>\n"
	"<table border=\"0\" cellspacing=\"0\" summary=\"%s\">\n"
	"<thead>"
	"<tr class=\"none\">"
	"<th colspan=\"2\">Name</th>"
	"<th>Last modified</th>"
	"<th>Size</th></tr>\n"
	"<tr><td colspan=\"4\"><hr></tr></thead>\n<tbody>\n",
	msg->request.uri, msg->request.uri, msg->request.uri);

  for (i = 0; i < dnum; i++) {
    String	path;
    struct stat	sbuf;
    int		n;
    char	tbuf[32];

    if (asprintf(&path, "%s/%s", msg->file.path, dir[i]) == -1) {
      wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
      break;
    }
    /* failed to get file status */
    n = lstat(path, &sbuf);	free((String)path);
    if (n == -1) {
      fprintf(fout,
	"<tr>\n<td>&nbsp;</td>"
	"<td><a href=\"%s\">%s</a></td>\n"
	"<td colspan=\"2\"></td>\n</tr>\n",
	dir[i], dir[i]);
      continue;
    }

    /* print the icon of the file type */
    fputs("<tr>\n<td>", fout);
    printFileForm(dir[i], (mode_t)sbuf.st_mode, fout);
    fputs("</td>\n", fout);

    /* print file name (directory link is a special) */
    if (S_ISDIR(sbuf.st_mode)) {
      fprintf(fout, "<td><a href=\"%s/\">%s</a>/</td>\n", dir[i], dir[i]);
    }
    else {
      fprintf(fout, "<td><a href=\"%s\">%s</a></td>\n", dir[i], dir[i]);
    }

    /* print last modified */
    memset(tbuf, 0, sizeof(tbuf));
    strftime(tbuf, sizeof(tbuf),
	     "%d-%b-%Y&nbsp;%H:%M", localtime(&sbuf.st_mtime));
    fprintf(fout, "<td class=\"timestamp\">&nbsp;%s&nbsp;</td>\n", tbuf);

    /* print file size and file description */
    if (S_ISDIR(sbuf.st_mode)) {
      fputs("<td class=\"size\">&nbsp;-&nbsp;</td>\n</tr>\n", fout);
    }
    else {
      fputs("<td class=\"", fout);
      if (sbuf.st_size < 1024) {
	fprintf(fout, "size_b\">%ldB", (long)sbuf.st_size);
      }
      else if (sbuf.st_size < 1024 * 1024) {
	fprintf(fout, "size_k\">%ld.%ldK",
		(long)sbuf.st_size / U_KILO,
		((long)sbuf.st_size % U_KILO) * 10 / U_KILO);
      }
      else if (sbuf.st_size < 1024 * 1024 * 1024) {
	fprintf(fout, "size_m\">%ld.%ldM",
		(long)sbuf.st_size / U_MEGA,
		((long)sbuf.st_size % U_MEGA) * 10 / U_MEGA);
      }
      else {
	fprintf(fout, "size_g\">%ld.%ldG",
		(long)sbuf.st_size / U_GIGA,
		((long)sbuf.st_size % U_GIGA) * 10 / U_GIGA);
      }
      fputs("</td>\n</tr>\n", fout);
    }
  }
  fputs("</tbody>\n"
	"</table>\n"
	"</div>\n",
	fout);

  /* include FooterFile if it exists */
  includeFile(msg->file.path, FOOTER_FILE, fout);

  fprintf(fout,
	"<hr>\n"
	"<address>\n"
	"<a href=\"%s\">%s</a> Server at %s Port %d\n"
	"</address>\n"
	"</body>\n"
	"</html>\n",
	SERVER_ADDRESS, SERVER_VERSION,
	msg->vhost->serverName, msg->conn->sock.port);
}

/*
 * Function Name:	includeFile
 * Description:		include HeaderFile/FooterFile file
 * Arguments:		file path, file name, output file pointer
 * Return Value:	none
 */
static void
includeFile(directory, filename, fout)
	const String	directory;
	const String	filename;
	register FILE *	fout;
{
  register FILE *	fin;
  String		path;

  if (asprintf(&path, "%s%s", directory, filename) == -1) {
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
    return;
  }

  fin = wvSafeFOpen(path, "r");	free((String)path);
  if (fin != NULL) {
    char	buf[BUFSIZ];

    memset(buf, 0, sizeof(buf));
    while (fgets(buf, sizeof(buf), fin) != NULL) {
      fputs(buf, fout);
      memset(buf, 0, sizeof(buf));
    }
    fputs("\n<hr>\n", fout);
    fclose(fin);
  }
}

/*****************************************************************************/

/*
 * Function Name:	printFileForm
 * Description:		print the file form
 * Arguments:		file name, file mode, file pointer
 * Return Value:	none
 */
static void
printFileForm(fname, fmode, fout)
	const String	fname;
	mode_t		fmode;
	register FILE *	fout;
{
  fputs("<img src=\"/icons/", fout);

  if (S_ISDIR(fmode)) {
    fputs("dir.gif\" alt=\"[DIR]\" ", fout);
  }
  else if (S_ISREG(fmode)) {
/*
 * Uncomment the following line if you don't like the detailed icons.
 */
/*#define	NO_DETAIL_INDEXING	1*/
#ifndef	NO_DETAIL_INDEXING
    String	p = NULL;

    p = strrchr(fname, '.');
    if (p == NULL) {
      fputs("unknown.gif\" alt=\"[   ]\" ", fout);
    }
    else {
      const FileForm *	ff = fformGet(p, strlen(p));

      if (ff == NULL) {
	fputs("generic.gif\" alt=\"[   ]\" ", fout);
      }
      else {
	switch (ff->form) {
	  case FF_COMPRESSED:	/* compressed */
	    fputs("compressed.gif\" alt=\"[CMP]\" ", fout);
	    break;
	  case FF_TAR:		/* tar archived */
	    fputs("tar.gif\" alt=\"[TAR]\" ", fout);
	    break;
	  case FF_ARCHIVE:	/* archived */
	    fputs("compressed.gif\" alt=\"[ARC]\" ", fout);
	    break;
	  case FF_BINARY:		/* binary */
	    fputs("binary.gif\" alt=\"[BIN]\" ", fout);
	    break;
	  case FF_HTML:		/* HTML */
	    fputs("layout.gif\" alt=\"[HTM]\" ", fout);
	    break;
	  case FF_TEXT:		/* text */
	    fputs("text.gif\" alt=\"[TXT]\" ", fout);
	    break;
	  case FF_PDF:		/* PDF */
	    fputs("pdf.gif\" alt=\"[PDF]\" ", fout);
	    break;
	  case FF_DVI:		/* DVI */
	    fputs("dvi.gif\" alt=\"[DVI]\" ", fout);
	    break;
	  case FF_POSTSCRIPT:	/* PostScript */
	    fputs("ps.gif\" alt=\"[PS ]\" ", fout);
	    break;
	  case FF_C:		/* C language */
	    fputs("c.gif\" alt=\"[C  ]\" ", fout);
	    break;
	  case FF_F77:		/* Fortran language */
	    fputs("f.gif\" alt=\"[F77]\" ", fout);
	    break;
	  case FF_PERL:		/* Perl language */
	    fputs("p.gif\" alt=\"[P  ]\" ", fout);
	    break;
	  case FF_TEX:		/* TeX/LaTeX */
	    fputs("tex.gif\" alt=\"[TEX]\" ", fout);
	    break;
	  case FF_PATCH:		/* patch file */
	    fputs("patch.gif\" alt=\"[PAT]\" ", fout);
	    break;
	  case FF_SHELL:		/* shell script */
	    fputs("script.gif\" alt=\"[SH ]\" ", fout);
	    break;
	  case FF_IMAGE:		/* image */
	    fputs("image1.gif\" alt=\"[IMG]\" ", fout);
	    break;
	  case FF_MOVIE:		/* movie */
	    fputs("movie.gif\" alt=\"[MOV]\" ", fout);
	    break;
	  case FF_SOUND:		/* sound/audio */
	    fputs("sound2.gif\" alt=\"[SND]\" ", fout);
	    break;
	  default:
	    break;
	}
      }
    }
#else
    fputs("generic.gif\" alt=\"[   ]\" ", fout);
#endif	/* NO_DETAIL_INDEXING */
  }
  else if (S_ISLNK(fmode)) {
    fputs("link.gif\" alt=\"[LNK]\" ", fout);
  }
  else {
    fputs("unknown.gif\" alt=\"[UNK]\" ", fout);
  }

  fputs("width=\"20\" height=\"22\">", fout);
}

/*
 * following is automatically generated by gperf
 */
#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
static unsigned int
hash(str, len)
	register const char *	str;
	register unsigned int	len;
{
  static const unsigned char asso_values[] =
    {
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32,  0, 32, 32, 32,
       6, 11, 32, 32, 32,  9, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
       1, 32, 32, 32, 32, 32, 32, 32, 32,  4,
      30, 18,  5, 10,  0, 25, 32, 32, 15,  8,
      23, 32,  3, 32, 28, 31, 26, 32, 21, 32,
      13, 16, 20, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32
    };
  return asso_values[(unsigned char)str[len - 1]] + asso_values[(unsigned char)str[0]];
}

#ifdef __GNUC__
__inline
#endif
static const FileForm *
fformGet(str, len)
	register const char *	str;
	register unsigned int	len;
{
  static const unsigned char lengthtable[] =
    {
       4,  4,  3,  6,  3,  2,  2,  4,  4,  2,  4,  5,  4,  4,
       5,  2,  4,  4,  3,  4,  4,  5,  4,  5,  4,  4,  4,  4,
       5,  5,  3,  6,  4,  4,  3,  4,  3,  4,  4,  4,  5,  4,
       4,  4,  4,  4,  3
    };
  static const FileForm wordlist[] =
    {
      {".lzh",	FF_ARCHIVE},
      {".csh",	FF_SHELL},
      {".sh",	FF_SHELL},
      {".patch",	FF_PATCH},
      {".ph",	FF_PERL},
      {".h",	FF_C},
      {".Z",	FF_COMPRESSED},
      {".zip",	FF_ARCHIVE},
      {".bmp",	FF_IMAGE},
      {".c",	FF_C},
      {".pdf",	FF_PDF},
      {".tiff",	FF_IMAGE},
      {".tif",	FF_IMAGE},
      {".gif",	FF_IMAGE},
      {".diff",	FF_PATCH},
      {".f",	FF_F77},
      {".bz2",	FF_COMPRESSED},
      {".htm",	FF_HTML},
      {".pm",	FF_PERL},
      {".f77",	FF_F77},
      {".png",	FF_IMAGE},
      {".mpeg",	FF_MOVIE},
      {".mpg",	FF_MOVIE},
      {".jpeg",	FF_IMAGE},
      {".jpg",	FF_IMAGE},
      {".mp3",	FF_SOUND},
      {".hqx",	FF_BINARY},
      {".tex",	FF_TEX},
      {".html",	FF_HTML},
      {".perl",	FF_PERL},
      {".pl",	FF_PERL},
      {".shtml",	FF_HTML},
      {".sty",	FF_TEX},
      {".exe",	FF_BINARY},
      {".gz",	FF_COMPRESSED},
      {".tgz",	FF_COMPRESSED},
      {".bz",	FF_COMPRESSED},
      {".mov",	FF_MOVIE},
      {".bin",	FF_BINARY},
      {".dvi",	FF_DVI},
      {".midi",	FF_SOUND},
      {".avi",	FF_MOVIE},
      {".txt",	FF_TEXT},
      {".tar",	FF_TAR},
      {".snd",	FF_SOUND},
      {".mid",	FF_SOUND},
      {".ps",	FF_POSTSCRIPT}
    };

  static const short lookup[] =
    {
      -98,   6,  -1, -96,   9, -94,  16,  -1, -92,  19,
      -90,  25,  -1, -88,  -1, -86,  32,  -1,  33,  -1,
      -84,  37,  -1,  38,  -1, -82,  42,  -1,  43,  -1,
      -80,  46,  -3,  -2,  -8,  -3, -13,  -3, -19,  -4,
      -21,  -2, -27,  -5, -30,  -2, -37,  -6, -40,  -2,
      -47,  -6
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH) {
    register int	key = hash(str, len);

    if (key <= MAX_HASH_VALUE && key >= 0) {
      register int	index = lookup[key];

      if (index >= 0) {
	if (len == lengthtable[index]) {
	  register const char *	s = wordlist[index].name;

	  if (*str == *s && !memcmp(str + 1, s + 1, len - 1)) {
	    return &wordlist[index];
	  }
	}
      }
      else if (index < -TOTAL_KEYWORDS) {
	register int	offset = - 1 - TOTAL_KEYWORDS - index;
	register const unsigned char *	lengthptr = &lengthtable[TOTAL_KEYWORDS + lookup[offset]];
	register const FileForm *	wordptr = &wordlist[TOTAL_KEYWORDS + lookup[offset]];
	register const FileForm *	wordendptr = wordptr + -lookup[offset + 1];

	while (wordptr < wordendptr) {
	  if (len == *lengthptr) {
	    register const char *	s = wordptr->name;

	    if (*str == *s && !memcmp(str + 1, s + 1, len - 1)) {
	      return wordptr;
	    }
	  }
	  lengthptr++;
	  wordptr++;
	}
      }
    }
  }

  return 0;
}

/*****************************************************************************/

/*
 * Function Name:	quickSort
 * Description:		sort string by upper order
 * Arguments:		string array, left value, right value
 * Return Value:	none
 */
static void
quickSort(s, left, right)
	String	s[];
	int	left;
	int	right;
{
  int	i, last;

  if (left >= right) {
    return;
  }

  swap(s, left, (left + right) / 2);
  last = left;

  for (i = left + 1; i <= right; i++) {
    if (strcmp(s[i], s[left]) < 0) {
      swap(s, ++last, i);
    }
  }

  swap(s, left, last);
  quickSort(s, left, last - 1);
  quickSort(s, last + 1, right);
}

/*
 * Function Name:	swap
 * Description:		swap the string
 * Arguments:		string array, swapped position (1), swapped position (2)
 * Return Value:	none
 */
static void
swap(s, i, j)
	String	s[];
	int	i;
	int	j;
{
  String	t;

  t	= s[i];
  s[i]	= s[j];
  s[j]	= t;
}

/*****************************************************************************/

/*
 * Function Name:	cacheInit
 * Description:		initialize dirCache called at only once
 * Arguments:		none
 * Return Value:	none
 */
static void
cacheInit(void)
{
  int	i;

  dirList.next = dirList.prev = &dirList;
  for (i = 0; i < DIR_TABLE_SIZE; i++) {
    dirCache[i].path = NULL;
    dirCache[i].name = NULL;
    dirCache[i].mtime = 0;
    cacheInsert(&dirList, &dirCache[i]);
  }
}

/*
 * Function Name:	cacheFree
 * Description:		unlink directory cache file
 * Arguments:		none
 * Return Value:	none
 */
static void
cacheFree(void)
{
  register int		i;
  register DIR *	dfd;

  for (i = 0; i < DIR_TABLE_SIZE; i++) {
    wvFree(dirCache[i].path);
    unlink(dirCache[i].name);
    wvFree(dirCache[i].name);
  }

  dfd = opendir(httpd->cacheDir);
  if (dfd != NULL) {
    register struct dirent *	dp;

    chdir(httpd->cacheDir);
    while ((dp = readdir(dfd)) != NULL) {
      if (dp->d_name[0] == '.') {
	continue;
      }
      unlink(dp->d_name);
    }
    chdir(httpd->serverRoot);
    closedir(dfd);
  }
}


/*
 * Function Name:	cacheInsert
 * Description:		insert 'new' before 'pos'
 * Arguments:		position, new slot
 * Return Value:	none
 */
static void
cacheInsert(pos, new)
	register DirIndex *	pos;
	register DirIndex *	new;
{
  pos->prev->next = new;
  new->prev	  = pos->prev;
  new->next	  = pos;
  pos->prev	  = new;
}

/*
 * Function Name:	cacheRemove
 * Description:		remove the slot from dirCache
 * Arguments:		removed slot
 * Return Value:	removed slot
 */
static DirIndex *
cacheRemove(x)
	register DirIndex *	x;
{
  x->prev->next = x->next;
  x->next->prev = x->prev;
  return x;
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
autoindexRegisterHooks(void)
{
  wvHookHandler(directoryIndexor, HOOK_MIDDLE, DIR_MAGIC_TYPE);
}

Module	mod_autoindex = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_VULNERABLE,		/* safety */
  "mod_autoindex",		/* name */
  "Directory Listing Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  (void *)cacheInit,			/* init */
  (void *)cacheFree,			/* finish */
  (void *)autoindexRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_autoindex;
}
