/*****************************************************************************
 *
 * FILE:	mod_env.c
 * DESCRIPTION:	makeing environment variables module for wyvern
 * DATE:	Tue, Aug 20 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002-2003 졿Kouichi ABE (WALL), All rights reserved.
 * REFERENCE:	draft-coar-cgi-v11-03.txt
 * LICENSE:
 *   Copyright (C) 2002-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_env.c,v 1.1.2.13 2003/03/18 16:06:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <pwd.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	envInit(void);
static void	envFinish(void);
static int	setScriptEnvironment(Message *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static String *	env;
static int	envno;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	envInit
 * Description:		initialize server environment for CGI/SSI
 * Arguments:		none
 * Return Value:	none
 * Notes:
 *
 * SERVER_SOFTWARE:	the name and version of the server software
 * SERVER_NAME:		the name of the server
 * GATEWAY_INTERFACE:	the revision of the CGI specification
 *
 * SERVER_PROTOCOL:	the protocol version of the HTTP
 * SERVER_PORT:		the port which the request is received on the server
 *
 * REQUEST_METHOD:	the method of the HTTP
 * SCRIPT_NAME:		the identity of the CGI script
 * PATH_INFO:		the path to be interpreted by the CGI script
 * PATH_TRANSLATED:	the location in the server's document
 * QUERY_STRING:	a URL-encoded string
 * REMOTE_HOST:		the FQDN of the client
 * REMOTE_ADDR:		the IP address of the client
 * REMOTE_PORT:		the remote port of the client
 * AUTH_TYPE:		the requests "Authorization"
 * REMOTE_USER:		the remote user-ID
 * REMOTE_IDENT:	the identity information reported about the connection
 *			by a RFC 1413
 * CONTENT_TYPE:	the Internet Media Type
 * CONTENT_LENGTH:	the size of the message-body entity
 *
 *
 * HTTP_ACCEPT:
 * HTTP_ACCEPT_ENCODING:
 * HTTP_ACCEPT_LANGUAGE:
 * HTTP_USER_AGENT:
 * HTTP_REFERER:
 *
 * DOCUMENT_ROOT:	NCSA and APACHE added -- not in CGI spec
 * SERVER_ROOT:		NCSA added
 * SERVER_ADMIN:	APACHE added
 */
static void
envInit(void)
{
  /* allocate memory area of environment variables */
  env = (String *)wvCalloc(MAX_CGI_ENVNO, sizeof(String *));
  if (env == NULL) {
    die(ERR_OUT_OF_MEMORY);
  }
  envno = 0;

  /*
   * set common environment variables, but ignore errors.
   */
  asprintf(&env[envno++], "PATH=%s", DEFAULT_PATH);
  asprintf(&env[envno++], "SERVER_SOFTWARE=%s", SERVER_VERSION);
  asprintf(&env[envno++], "GATEWAY_INTERFACE=%s", CGI_VERSION);
  asprintf(&env[envno++], "SERVER_ROOT=%s", httpd->serverRoot);
#if	DEBUG
  fprintf(stderr, "[mod_env] init: %d\n", envno);
#endif	/* DEBUG */
}

/*
 * Function Name:	envFinish
 * Description:		clear server environment variables
 * Arguments:		none
 * Return Value:	none
 */
static void
envFinish(void)
{
  int	i;

  for (i = 0; i < envno; i++) {
    wvFree(env[i]);
  }
  wvFree(env);
}


/*
 * Function Name:	setScriptEnvironment
 * Description:		set script environments for CGI/SSI
 * Arguments:		message
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
setScriptEnvironment(msg)
	register Message *	msg;
{
  String	val;

#define	ENV_MAGIC_TYPE	"application/x-httpd-"
  if (wvStrncmp(msg->file.mime, ENV_MAGIC_TYPE, 20) != 0 &&
      wvStrcmp(msg->file.mime, SSI_MAGIC_TYPE) != 0) {
    return RET_OK;
  }
#if	DEBUG
  fputs("<setScriptEnvironment>\n", stderr);
#endif	/* DEBUG */

  /* create CGI environments */
  msg->cgi.env = (String *)calloc(MAX_CGI_ENVNO, sizeof(String *));
  if (msg->cgi.env == NULL) {
    return wvNoMemoryError(msg);
  }

  for (msg->cgi.envno = 0; msg->cgi.envno < envno; msg->cgi.envno++) {
    msg->cgi.env[msg->cgi.envno] = wvStrdup(env[msg->cgi.envno]);
  }
  asprintf(&msg->cgi.env[msg->cgi.envno++], "SERVER_PROTOCOL=%s",
	    httpVer[msg->request.http]);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "SERVER_PORT=%d",
	    msg->conn->sock.port);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "SERVER_NAME=%s",
	    msg->vhost->serverName);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "SERVER_ADMIN=%s",
	    msg->vhost->serverAdmin);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "DOCUMENT_ROOT=%s",
	    msg->vhost->docRoot);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "REQUEST_METHOD=%s",
	    methodName[msg->request.method]);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "SCRIPT_NAME=%s", msg->request.uri);
  if (msg->request.path_info != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "PATH_INFO=%s",
	      msg->request.path_info);

    if (msg->request.path_info[1] == '~') {
      size_t		len;
      String		name;
      struct passwd *	pwd;

      len  = strcspn(msg->request.path_info   + TILDE_OFFSET, "/");
      name = wvStrndup(msg->request.path_info + TILDE_OFFSET, len);
      pwd  = getpwnam(name);
      if (pwd == NULL) {	/* invalid user: no home directory  */
	asprintf(&msg->cgi.env[msg->cgi.envno++], "PATH_TRANSLATED=%s/%%7E%s",
		 msg->vhost->docRoot, msg->request.path_info + TILDE_OFFSET);
      }
      else {
	asprintf(&msg->cgi.env[msg->cgi.envno++], "PATH_TRANSLATED=%s/%s%s",
		 pwd->pw_dir, msg->vhost->userDir,
		 msg->request.path_info + TILDE_OFFSET + len);
      }
      free((String)name);
    }
    else {
      asprintf(&msg->cgi.env[msg->cgi.envno++], "PATH_TRANSLATED=%s%s",
		msg->vhost->docRoot, msg->request.path_info);
    }
  }

  asprintf(&msg->cgi.env[msg->cgi.envno++], "QUERY_STRING=%s",
	   msg->request.query_string == NULL ? "" : msg->request.query_string);

  if (msg->conn->client.name != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "REMOTE_HOST=%s",
	      msg->conn->client.name);
  }

  asprintf(&msg->cgi.env[msg->cgi.envno++], "REMOTE_ADDR=%s",
	   msg->conn->client.ip);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "REMOTE_PORT=%s",
	   msg->conn->client.port);
  asprintf(&msg->cgi.env[msg->cgi.envno++], "AUTH_TYPE=Basic");

  if (msg->user.authName != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "REMOTE_USER=%s",
	     msg->user.authName);
  }

  if (msg->user.identName != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "REMOTE_IDENT=%s",
	     msg->user.identName);
  }

  if (msg->request.method == M_POST) {
    val = headerLookup(msg->request.headers, "CONTENT_TYPE");
    if (val != NULL) {
      asprintf(&msg->cgi.env[msg->cgi.envno++], "CONTENT_TYPE=%s", val);
    }
    if (msg->request.content_length != 0) {
      asprintf(&msg->cgi.env[msg->cgi.envno++], "CONTENT_LENGTH=%d",
		msg->request.content_length);
    }
  }

  /* no standard environment variable */
  val = headerLookup(msg->request.headers, "ACCEPT");
  if (val != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "HTTP_ACCEPT=%s", val);
  }

  val = headerLookup(msg->request.headers, "ACCEPT_ENCODING");
  if (val != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "HTTP_ACCEPT_ENCODING=%s", val);
  }

  val = headerLookup(msg->request.headers, "ACCEPT_LANGUAGE");
  if (val != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "HTTP_ACCEPT_LANGUAGE=%s", val);
  }

  val = headerLookup(msg->request.headers, "USER_AGENT");
  if (val != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "HTTP_USER_AGENT=%s", val);
  }

  val = headerLookup(msg->request.headers, "REFERER");
  if (val != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "HTTP_REFERER=%s", val);
  }

  val = headerLookup(msg->request.headers, "COOKIE");
  if (val != NULL) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "HTTP_COOKIE=%s", val);
  }

#if	USE_SSL
  if (msg->conn->sock.type == SOC_SECURE) {
    asprintf(&msg->cgi.env[msg->cgi.envno++], "SSL_VERSION=%s",
	     SSL_get_version(msg->conn->sock.ssl));
    asprintf(&msg->cgi.env[msg->cgi.envno++], "SSL_CIPHER=%s",
	     SSL_get_cipher(msg->conn->sock.ssl));
  }
#endif	/* USE_SSL */

  msg->cgi.env[msg->cgi.envno] = NULL;

#if	NODEBUG
  {
    int	i;
    for (i = 0; i < msg->cgi.envno; i++) {
      fprintf(stderr, "[%02d] %s\n", i, msg->cgi.env[i]);
    }
  }
#endif	/* DEBUG */

  return RET_OK;
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
envRegisterHooks(void)
{
  wvHookFixups(setScriptEnvironment, HOOK_FIRST);
}

Module	mod_env = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_VULNERABLE,		/* safety */
  "mod_env",			/* name */
  "Environment Variables Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  (void *)envInit,		/* init */
  (void *)envFinish,		/* finish */
  (void *)envRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_env;
}
