/*****************************************************************************
 *
 * FILE:	mod_gzip.c
 * DESCRIPTION:	gzip module using the zlib compression library
 * DATE:	Thu, May 17 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2003 졿Kouichi ABE (WALL), All rights reserved.
 * NOTE:	Based on minigzip.c
 * LICENSE:
 *  Copyright (C) 2001-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_gzip.c,v 1.20.2.12 2003/03/18 16:06:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_MMAP
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#endif	/* HAVE_MMAP */
#include "wyvern.h"
#include "functions.h"
#include "zlib.h"	/* collision of free_func used with OpenSSL */


/*****************************************************************************
 *
 *	definition of macros
 *
 *****************************************************************************/
#define	GZIP_MODE	"wb6"	/* See zlib.h */
#define GZIP_BUFSIZE	(BUF_SIZE*4)
#define	MIN_FILE_SIZE	(512)


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	gzipEncoding(Message *);
static int	gzipCompress(int, gzFile);
#if	HAVE_MMAP
static int	gzipCompressWithMmap(int, gzFile);
#endif	/* HAVE_MMAP */
static int	reopen(Message *, const String);
static void	gzipRegisterHooks(void);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	gzipEncoding
 * Description:		compress content file with gzip compress
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
gzipEncoding(msg)
	register Message *	msg;
{
  String	accept_encoding;
  String	content_type;
  String	gzfile;
  int		fd;
  int		in;
  gzFile	out;
  int		ret = RET_OK;

#if	1
  /*
   * Does OpenSSL conflict with zlib package?
   * I found the conflict name as "free_func" in both zlib.h and ssl.h etc.
   */
  if (msg->conn->sock.type == SOC_SECURE) {
    return RET_OK;	/* no gzipped */
  }
#endif
  if (msg->response.status > STATUS_CODE_3XX ||
      msg->response.status == STATUS_PARTIAL_CONTENT) {
    return RET_OK;	/* no gzipped */
  }

  accept_encoding = headerLookup(msg->request.headers, "ACCEPT_ENCODING");
  content_type	  = headerLookup(msg->response.headers, "Content-Type");
  /* no supported by client */
  if (accept_encoding == NULL || strstr(accept_encoding, "gzip") == NULL) {
    return RET_OK;	/* no gzipped */
  }
  /* content size < MIN_FILE_SIZE */
  if (msg->response.content_length < MIN_FILE_SIZE) {
    return RET_OK;	/* no gzipped */
  }
#if	DEBUG
  fputs("<gzipCompress>\n", stderr);
#endif	/* DEBUG */

  /* open gzip'ed output file */
  if (asprintf(&gzfile, "%s/gzfile.XXXXXX", httpd->tmpDir) == -1) {
    wvSyslog(LOG_WARNING,
	     "[mod_gzip] failed to create temporary file ... No gzipped\n");
    return RET_NG;
  }
  fd = mkstemp(gzfile);
  if (fd == -1) {
    ret = RET_NG;
    goto error;
  }

  out = gzdopen(fd, GZIP_MODE);
  if (out == NULL) {
    ret = RET_NG;
    goto error;
  }
  in = msg->file.fd;

#if	HAVE_MMAP
  /*
   * Try first compressing with mmap.
   * If mmap fails, use the normal fread loop.
   */
  if (gzipCompressWithMmap(in, out) == Z_OK || gzipCompress(in, out) == Z_OK) {
    ret = reopen(msg, gzfile);
  }
#else
  if (gzipCompress(in, out) == Z_OK) {
    ret = reopen(msg, gzfile);
  }
#endif	/* HAVE_MMAP */

  unlink(gzfile);
  if (ret == RET_NG) {	/* error */
    (void)lseek(msg->file.fd, (off_t)0L, SEEK_SET);	/* rewind */
  }

error:
  free((String)gzfile);
  return ret;
}

/*
 * Function Name:	gzipCompress
 * Description:		compress content file with gzip compress
 * Arguments:		file pointer, gzip'ed file pointer
 * Return Value:	successful: Z_OK, otherwise: Z_ERRNO
 */
static int
gzipCompress(in, out)
	int	in;
	gzFile	out;
{
  char		buf[GZIP_BUFSIZE];
  size_t	len;
  bool_t	flag;	/* successful: TRUE, otherwise: FALSE */

  flag = TRUE;
  memset(buf, 0, sizeof(buf));
  while ((len = read(in, buf, sizeof(buf))) > 0) {
    if (gzwrite(out, buf, (unsigned)len) != len) {
      flag = FALSE;
      break;
    }
    memset(buf, 0, sizeof(buf));
  }
  if (gzclose(out) != Z_OK) {	/* close the compressed content file */
    flag = FALSE;
  }

  return (flag ? Z_OK : Z_ERRNO);
}

#if	HAVE_MMAP
/*
 * Function Name:	gzipCompressWithMmap
 * Description:		compress content file with gzip compress using mmap
 * Arguments:		file pointer, gzip'ed file pointer
 * Return Value:	successful: Z_OK, otherwise: Z_ERRNO
 */
static int
gzipCompressWithMmap(in, out)
	int	in;
	gzFile	out;
{
  struct stat	sbuf;
  caddr_t	buf;	/* mmap'ed buffer for the entire input file */
  off_t		size;	/* length of the input file */
  size_t	len;
  bool_t	flag;	/* successful: TRUE, otherwise: FALSE */

  /* Determine the size of the file, needed for mmap: */
  if (fstat(in, &sbuf) == -1) {
    return Z_ERRNO;
  }

  size = sbuf.st_size;
  if (size <= 0) {
    return Z_ERRNO;
  }

  /* Now do the actual mmap: */
  buf = mmap((caddr_t) 0, (size_t)size, PROT_READ, MAP_SHARED, in, (off_t)0); 
  if (buf == (caddr_t)(-1)) {
    return Z_ERRNO;
  }
  flag = TRUE;

  /* Compress the whole file at once: */
  len = gzwrite(out, (String)buf, (unsigned)size);
  if (len != (int)size) {
    flag = FALSE;
  }
  munmap(buf, (size_t)size);

  if (gzclose(out) != Z_OK) {
    flag = FALSE;
  }

  return (flag ? Z_OK : Z_ERRNO);
}
#endif /* HAVE_MMAP */


/*
 * Function Name:	reopen
 * Description:		reopen file of content
 * Arguments:		message, gzip'ed file
 * Return Value:	successful: 0, otherwise: -1
 */
static int
reopen(msg, gzfile)
	register Message *	msg;
	const String		gzfile;
{
  int		fd;
  struct stat	sbuf;

  /* open new file */
  fd = wvSafeOpen(gzfile, O_RDONLY, &sbuf, FALSE);
  if (fd == -1) {
    return RET_NG;
  }

  /* change file descriptor from old to new */
  if (responseHeaderEntry(&msg->response.headers,
			  "Content-Encoding", "gzip") == NULL) {
    (void)close(fd);
    return RET_NG;
  }

  /* the file will be removed when it's closed. */
  wvClose(msg->file.fd);
  msg->file.fd			= fd;
  msg->file.status.st_size	= sbuf.st_size;
  msg->response.content_length	= sbuf.st_size;

  return RET_OK;
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
gzipRegisterHooks(void)
{
  wvHookFilter(gzipEncoding, HOOK_LAST, "text/*");
}

Module	mod_gzip = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_STANDARD,		/* safety */
  "mod_gzip",			/* name */
  "GZip Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)gzipRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_gzip;
}
