/*****************************************************************************
 *
 * FILE:	mod_ident.c
 * DESCRIPTION:	Identification Protocol (RFC1413) module for wyvern
 * DATE:	Sat, Sep 16 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2000-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_ident.c,v 1.19.2.7 2003/03/18 16:06:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <sys/types.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <setjmp.h>
#include <signal.h>
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	IDENT_TIMEOUT	(30)	/* wait at least 30 seconds (or longer) */
#define	IDENT_PORT	(113)	/* authentication service in /etc/service */
#define	ANY_PORT	(0)


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	useRemoteIdent(Thread *);
static void	timeout(int);
static void	identRegisterHooks(void);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static jmp_buf	idenv;	/* to store program environments */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	useRemoteIdent
 * DESCRIPTION:		identify remote user
 * ARGUMENTS:		thread
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
useRemoteIdent(thd)
	register Thread *	thd;
{
  unsigned int		rmt_port;
  unsigned int		our_port;
#ifndef	USE_IPV6
  struct sockaddr_in	rmt_query;
  struct sockaddr_in	our_query;
#else
  struct sockaddr_in6	rmt_query;
  struct sockaddr_in6	our_query;
#endif	/* USE_IPV6 */
  int			sd;
  int			ret = RET_NG;

  /* create the socket */
  sd = socket(thd->sock.ainfo.ai_family, thd->sock.ainfo.ai_socktype,
	      thd->sock.ainfo.ai_protocol);
  if (sd == -1) {
    wvSyslog(LOG_ERR, "failed to create an ident socket\n");
    return RET_NG;
  }

  /* set up a timer to avoid blocking the server */
  if (setjmp(idenv) == 0) {
    struct addrinfo	hints;
    struct addrinfo *	res;
    struct addrinfo *	ai;
    char		buf[BUFSIZ];
    char		user[MAX_NAME_LEN];
    int			n;

    signal(SIGALRM, timeout);
    alarm(IDENT_TIMEOUT);

    /* copy local and remote socket address */
#if	USE_IPV6
    our_query		= httpd->socket.addr;
    our_query.sin6_port	= htons(ANY_PORT);
    rmt_query		= thd->client.addr;
    rmt_query.sin6_port	= htons(IDENT_PORT);
#else
    our_query		= httpd->socket.addr;
    our_query.sin_port	= htons(ANY_PORT);
    rmt_query		= thd->client.addr;
    rmt_query.sin_port	= htons(IDENT_PORT);
#endif	/* USE_IPV6 */

    /* bind the socket */
    if (bind(sd, thd->sock.ainfo.ai_addr, thd->sock.ainfo.ai_addrlen) == -1) {
      wvSyslog(LOG_ERR, "failed to bind a name to an ident socket\n");
      goto ident_exit;
    }

    /* create the query message to the ident server (see sec.4 in RFC1413) */
    memset(&hints, 0, sizeof(hints));
    hints.ai_family	= thd->sock.ainfo.ai_family;
    hints.ai_socktype	= thd->sock.ainfo.ai_socktype;
    hints.ai_protocol	= thd->sock.ainfo.ai_protocol;
    hints.ai_flags	= AI_NUMERICHOST;
    n = getaddrinfo(thd->client.ip, thd->client.port, &hints, &res);
    if (n != 0) {
      goto ident_exit;
    }

    memset(buf, 0, sizeof(buf));
#if	USE_IPV6
    n = snprintf(buf, sizeof(buf), "%u,%u" CRLF,
				    thd->client.port,
				    ntohs(httpd->socket.addr.sin6_port));
#else
    n = snprintf(buf, sizeof(buf), "%u,%u" CRLF,
				    thd->client.port,
				    ntohs(httpd->socket.addr.sin_port));
#endif	/* USE_IPV6 */
    if (n >= sizeof(buf)) {
      wvSyslog(LOG_ALERT, "caused buffer overflow in ident.mod\n");
      goto ident_exit;
    }

    /* connect the ident server */
    if (connect(sd, (struct sockaddr *)&rmt_query, sizeof(rmt_query)) == -1) {
      if ((errno != ECONNREFUSED) && (errno != EHOSTUNREACH) &&
	  (errno != ECONNRESET)) {
	wvSyslog(LOG_ERR, "failed to connect a remote identd\n");
      }
      goto ident_exit;
    }

    /* send the query to the ident server */
    if (write(sd, buf, (size_t)n) != n) {
      wvSyslog(LOG_ERR, "failed to write a data in ident.mod\n");
      goto ident_exit;
    }

    /* nothing more to send, so shutdown the write socket */
    if (shutdown(sd, SHUT_WR) < 0) {
      wvSyslog(LOG_ERR, "failed to shutdown in ident.mod\n");
      goto ident_exit;
    }

    /* read response from the ident server */
    memset(buf, 0, sizeof(buf));
    n = read(sd, buf, (size_t)sizeof(buf));
    if (n == -1) {
      wvSyslog(LOG_ERR, "failed to read a data in ident.mod\n");
      goto ident_exit;
    }

    /* parse the response */
    memset(user, 0, sizeof(user));
    n = sscanf(buf, "%5u , %5u : USERID :%*[^:]:%8s",
		    &rmt_port, &our_port, user);
#if	USE_IPV6
    if ((n == 3) &&
	(thd->client.port == rmt_port) &&
	(ntohs(httpd->socket.addr.sin6_port) == our_port)) {
      String	p = strchr(user, CR);
      if (p != NULL) {
	*p = EOL;
      }
      thd->identUser = xstrdup(user);
    }
#else
    if ((n == 3) &&
	(thd->client.port == rmt_port) &&
	(ntohs(httpd->socket.addr.sin_port) == our_port)) {
      String	p = strchr(user, CR);
      if (p != NULL) {
	*p = EOL;
      }
      thd->identUser = xstrdup(user);
    }
#endif	/* USE_IPV6 */

ident_exit:
    alarm(0);	/* reset timer */
    wvClose(sd);
  }

  return ret;
}

/*
 * FUNCTION NAME:	timeout
 * DESCRIPTION:		timeout to request
 * ARGUMENTS:		signal id
 * RETURN VALUE:	none
 */
static void
timeout(sig)
	int	sig;
{
  longjmp(idenv, sig);
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
identRegisterHooks(void)
{
  wvHookCheckUserID(useRemoteIdent, HOOK_MIDDLE);
}

Module	mod_ident = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_STANDARD,		/* safety */
  "mod_ident",			/* name */
  "Identification Protocol Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)identRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_ident;
}
