/*****************************************************************************
 *
 * FILE:	mod_rcgi.c
 * DESCRIPTION:	Rapid CGI (for NPH programs) module for wyvern
 * DATE:	Sat, Oct 19 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002-2003 졿Kouichi ABE (WALL), All rights reserved.
 * REFERENCE:	draft-coar-cgi-v11-03.txt
 * LICENSE:
 *  Copyright (C) 2002-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_rcgi.c,v 1.1.2.8 2003/03/18 16:06:14 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <signal.h>
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of global functions
 *
 *****************************************************************************/
#define	DEF_RCGI_SUFFIX		".rcgi"
#define	DEF_RCGI_PATH		"/rcgi-bin/"
#define	RCGI_MAGIC_TYPE		"application/x-httpd-rcgi"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	rcgiTranslateURI(Message *);
static int	rcgiTypeChecker(Message *);
static int	rcgiRunScript(Message *);
static void	rcgiTrap(int);

static void	sigUsrTrap(int);
static void	waitInit(void);
static void	waitParent(void);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static volatile sig_atomic_t	sigflag;
static sigset_t			newmask;
static sigset_t			oldmask;
static sigset_t			zeromask;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * SERVER_SOFTWARE:	the name and version of the server software
 * SERVER_NAME:		the name of the server
 * GATEWAY_INTERFACE:	the revision of the CGI specification
 *
 * SERVER_PROTOCOL:	the protocol version of the HTTP
 * SERVER_PORT:		the port which the request is received on the server
 *
 * REQUEST_METHOD:	the method of the HTTP
 * SCRIPT_NAME:		the identity of the CGI script
 * PATH_INFO:		the path to be interpreted by the CGI script
 * PATH_TRANSLATED:	the location in the server's document
 * QUERY_STRING:	a URL-encoded string
 * REMOTE_HOST:		the FQDN of the client
 * REMOTE_ADDR:		the IP address of the client
 * REMOTE_PORT:		the remote port of the client
 * AUTH_TYPE:		the requests "Authorization"
 * REMOTE_USER:		the remote user-ID
 * REMOTE_IDENT:	the identity information reported about the connection
 *			by a RFC 1413
 * CONTENT_TYPE:	the Internet Media Type
 * CONTENT_LENGTH:	the size of the message-body entity
 *
 *
 * HTTP_ACCEPT:
 * HTTP_ACCEPT_ENCODING:
 * HTTP_ACCEPT_LANGUAGE:
 * HTTP_USER_AGENT:
 * HTTP_REFERER:
 *
 * DOCUMENT_ROOT:	NCSA and APACHE added -- not in CGI spec
 * SERVER_ROOT:		NCSA added
 * SERVER_ADMIN:	APACHE added
 */

/*****************************************************************************/

/*
 * Function Name:	rcgiTranslateURI
 * Description:		translate an alias of request URI
 * Arguments:		message
 * Return Value:	successful: 0, no translated: 1, otherwise: -1
 */
static int
rcgiTranslateURI(msg)
	register Message *	msg;
{
  if (wvStrncmp(msg->request.uri, DEF_RCGI_PATH, 10) == 0) {
    int	n;
    
    n = asprintf(&msg->file.path, "%s%s", httpd->serverRoot, msg->request.uri);
    if (n == -1) {
      return wvNoMemoryError(msg);
    }
    msg->file.mime = RCGI_MAGIC_TYPE;
#if	DEBUG
    fprintf(stderr, "[rcgiTranslateURI] %s\n", msg->file.path);
#endif	/* DEBUG */
    return RET_OK;
  }
#if	DEBUG
  fprintf(stderr, "[rcgiTranslateURI] %s(%d)\n",
	  msg->request.uri, strlen(DEF_RCGI_PATH));
#endif	/* DEBUG */
  return RET_CONT;
}

/*
 * Function Name:	rcgiTypeChecker
 * Description:		check mimetype of supplied file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
rcgiTypeChecker(msg)
	register Message *	msg;
{
  if (msg->file.mime == NULL) {
    String	suffix;	/* suffix of file */

    suffix = strrchr(msg->file.path, '.');
    if (suffix != NULL) {
      if (wvStrncmp(suffix, DEF_RCGI_SUFFIX, 5) == 0) {
	msg->file.mime = RCGI_MAGIC_TYPE;
#if	DEBUG
	fprintf(stderr, "[rcgiTypeChecker] OK: %s\n", msg->file.path);
#endif	/* DEBUG */
	return RET_OK;
      }
#if	DEBUG
      fprintf(stderr, "[rcgiTypeChecker] NG: %s\n", msg->file.path);
#endif	/* DEBUG */
      return RET_NG;
    }
#if	DEBUG
    fprintf(stderr, "[rcgiTypeChecker] No suffix: %s\n", msg->file.path);
#endif	/* DEBUG */
  }
#if	DEBUG
  fprintf(stderr, "[rcgiTypeChecker] media type: %s\n", msg->file.mime);
#endif	/* DEBUG */

  return RET_CONT;
}

/*
 * Function Name:	rcgiRunScript
 * Description:		execute NPH-CGI script
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
rcgiRunScript(msg)
	register Message *	msg;
{
  char		cwd[MAX_PATH_LEN + 1];	/* CGI working directory */
#if	USE_SUEXEC
  String	args[3] = { SUEXEC_PATH, NULL, NULL };
#else
  String	args[2] = { NULL, NULL };
#endif	/* USE_SUEXEC */
  pid_t		pid;
  String	p;

#if	DEBUG
  fprintf(stderr, "[rcgiRunScript] %s\n", msg->file.path);
#endif	/* DEBUG */
  /* set CGI working directory */
  memset(cwd, 0, sizeof(cwd));
  if (wvStrncpy(cwd, msg->file.path, sizeof(cwd)) == NULL) {
    wvSyslog(LOG_ERR, "[mod_rcgi] failed to change directory\n");
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    return RET_NG;
  }
  p = strrchr(cwd, '/');
  if (p != NULL) {
    *p = EOL;
  }

  waitInit();

  /* execute the CGI script */
  pid = fork();
  switch (pid) {
    case -1:	/* fork error */
      wvSyslog(LOG_ERR, "[mod_rcgi] failed to fork(2)\n");
      msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
      return RET_NG;

    case 0:	/* child process */
      dup2(fileno(msg->vhost->errorLog.fp), STDERR_FILENO);
      dup2(msg->conn->sock.sd, STDOUT_FILENO);
      if (msg->request.method == M_POST) {
	/* bind cfd to stdin */
	dup2(msg->request.post_fd, STDIN_FILENO);
      }

      waitParent();
#if	USE_SUEXEC
      args[1] = msg->file.path;
      if (chdir(cwd) == -1 ||
	  execve(SUEXEC_PATH, args, msg->cgi.env) == -1) {
       	rcgiTrap(errno);
      }
#else
      args[0] = msg->file.path;
      if (chdir(cwd) == -1 ||
	  execve(msg->file.path, args, msg->cgi.env) == -1) {
       	rcgiTrap(errno);
      }
#endif	/* USE_SUEXEC */
      _exit(CGI_ERROR_STATUS);

    default:	/* parent process */
      msg->status  = MSG_IDLING;
      msg->cgi.pid = pid;
      kill(pid, SIGUSR1);	/* send start signal to child */
      msg->conn->keepalive.status = KA_CLOSE;
      break;
  }

  return RET_WAIT;
}

/*
 * Function Name:	rcgiTrap
 * Description:		send error message of CGI
 * Arguments:		error number
 * Return Value:	none
 */
static void
rcgiTrap(e)
	int	e;
{
#if	USE_JP_MESG
#define	ERR_CGI_TITLE	"CGI 顼"
#define	ERR_ACCES_H1	"CGI ץ˼¹ԵĤޤ"
#define	ERR_ACCES_MESG	"ꤵ줿 CGI ץ˼¹ԵĤޤ"
#define	ERR_OTHER_H1	"¾Υ顼"
#define	ERR_OTHER_MESG	"CGI ¹Իͽ̥顼ȯޤ"
#else
#define	ERR_CGI_TITLE	"CGI Error"
#define	ERR_ACCES_H1	"Permission denied"
#define	ERR_ACCES_MESG	"The CGI file has no executable permission."
#define	ERR_OTHER_H1	"Unexcepted error"
#define	ERR_OTHER_MESG	"Server had unexpected error."
#endif	/* USE_JP_MESG */
  fputs("Content-Type: text/html" CRLF CRLF, stdout);
  fprintf(stdout,
	  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
	  "<html>\n"
	  "<head>\n"
	  "<title>%s</title>\n"
	  "</head>\n"
	  "<body bgcolor=\"white\">\n",
	  ERR_CGI_TITLE);

  fputs("<h1 align=\"center\">", stdout);
  switch (e) {
    case EACCES:
      fprintf(stdout, "%s</h1>\n<p align=\"center\">%s",
	      ERR_ACCES_H1, ERR_ACCES_MESG);
      break;
    default:
      fprintf(stdout, "%s</h1>\n<p align=\"center\">%s",
	      ERR_OTHER_H1, ERR_OTHER_MESG);
      break;
  }
  fprintf(stdout,
	  "</p>\n"
	  "<hr>\n"
	  "<address>\n"
	  "Powered by&nbsp;"
	  "<a href=\"%s\">%s</a>\n"
	  "&nbsp;&lt;<a href=\"mailto:%s\">%s</a>&gt;\n"
	  "</address>\n"
	  "</body>\n"
	  "</html>\n",
	  SERVER_ADDRESS, SERVER_VERSION,
	  httpd->vhost->serverAdmin, httpd->vhost->serverAdmin);
  fflush(stdout);
}

static void
sigUsrTrap(sig)
	int	sig;
{
  sigflag = 1;
}

static void
waitInit(void)
{
  sigflag = 0;

  signal(SIGUSR1, sigUsrTrap);

  sigemptyset(&zeromask);
  sigemptyset(&newmask);
  sigaddset(&newmask, SIGUSR1);

  sigprocmask(SIG_BLOCK, &newmask, &oldmask);
}

static void
waitParent(void)
{
  while (sigflag == 0) {
    sigsuspend(&zeromask);
  }
  sigflag = 0;
  sigprocmask(SIG_SETMASK, &oldmask, NULL);
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
rcgiRegisterHooks(void)
{
  wvHookTranslateURI(rcgiTranslateURI, HOOK_MIDDLE);
  wvHookTypeChecker(rcgiTypeChecker, HOOK_MIDDLE);
  wvHookHandler(rcgiRunScript, HOOK_MIDDLE, RCGI_MAGIC_TYPE);
}

Module	mod_rcgi = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_VULNERABLE,		/* safety */
  "mod_rcgi",			/* name */
  "Rapid CGI Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)rcgiRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_rcgi;
}
