/*****************************************************************************
 *
 * FILE:	mod_ssi.c
 * DESCRIPTION:	SSI module for wyvern
 * DATE:	Sat, Sep 16 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2000-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_ssi.c,v 1.41.2.18 2003/03/18 16:06:14 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_VFORK_H
#include <vfork.h>
#endif	/* HAVE_VFORK_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>
#include "wyvern.h"
#include "wpath.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	DEF_SSI_SUFFIX	".shtml"
#ifndef	MAP_FAILED
#define	MAP_FAILED	((caddr_t)-1)
#endif	/* MAP_FAILED */
#define	U_KILO		(1024L)
#define	U_MEGA		(1024L*1024L)
#define	U_GIGA		(1024L*1024L*1024L)
#define	TIMEFMT_LEN	(64)
#define	SSI_BUFSIZE	(256)

typedef enum {
  SSI_CONFIG,
  SSI_INCLUDE,
  SSI_ECHO,
  SSI_FSIZE,
  SSI_FLASTMOD,
  SSI_EXEC
} ssi_directives_t;

typedef enum {	/* ssi token parsing state */
  ST_GND,
  ST_LBR,
  ST_EXC,
  ST_HYP1,
  ST_HYP2,
  ST_CMD
} state_t;

typedef enum {
  SZ_BYTES,
  SZ_ABBREV
} sizefmt_t;

typedef enum {
  ET_TAG,
  ET_VALUE,
  ET_FORMAT
} error_ssi_t;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	ssiTypeChecker(Message *);
static int	ssiFilter(Message *);
static void	ssiExecDirective(Message *, FILE *, const String);

static void	do_config(Message *, const String, const String);
static void	do_include(Message *, FILE *, const String, const String);
static void	do_echo(Message *, FILE *, const String, const String);
static void	do_fsize(Message *, FILE *, const String, const String);
static void	do_flastmod(Message *, FILE *, const String, const String);
static void	do_exec(Message *, FILE *, const String, const String);
static void	ssiRunProgram(Message *, FILE *, const String);
static void	ssiReapChild(int);

static void	printFormattedTime(Message *, FILE *, time_t, bool_t);
static int	getFilePath(Message *, const String, const String, String *);
static int	ssiError(error_ssi_t, const String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static char		timefmt[TIMEFMT_LEN];
static sizefmt_t	sizefmt;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	ssiTypeChecker
 * Description:		check mimetype of supplied file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
ssiTypeChecker(msg)
	register Message *	msg;
{
  if (msg->file.mime == NULL) {
    String	suffix;	/* suffix of file */

#if	DEBUG
    fprintf(stderr, "[ssiTypeChecker] %s\n", msg->file.path);
#endif	/* DEBUG */
    suffix = strrchr(msg->file.path, '.');
    if (suffix != NULL) {	/* no suffix */
      if (wvStrncmp(suffix, DEF_SSI_SUFFIX, 6) == 0) {
	msg->file.mime = SSI_MAGIC_TYPE;
	return RET_OK;
      }
      return RET_NG;
    }
  }

  return RET_CONT;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	ssiFilter
 * DESCRIPTION:		address an SSI content
 * ARGUMENTS:		message
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
ssiFilter(msg)
	register Message *	msg;
{
  String		ssifile;
  int			fd;
  String		file;
  String		ch;
  register FILE *	fout;
  state_t		st;
  String		directive;
  String		dirp;
  String		text;
  size_t		txtc;	/*c length ounter */
  struct stat		sbuf;

#if	DEBUG
  fputs("<ssiFilter>\n", stderr);
#endif	/* DEBUG */
  /* default formats */
  memset(timefmt, 0, sizeof(timefmt));
  				/* ex. Sat Nov 27 01:26:37 JST 1999	*/
  wvStrncpy(timefmt, "%a %b %e %T %Z %Y", sizeof(timefmt));
  sizefmt = SZ_BYTES;

  /* mapping file */
  file = (String)mmap(NULL, (size_t)msg->file.status.st_size,
		      PROT_READ|PROT_WRITE, MAP_PRIVATE, msg->file.fd,
		      (off_t)0);
  if (file == MAP_FAILED) {
    wvSyslog(LOG_ERR, "[mod_ssi] mmap(2) error\n");
    goto noparse;
  }

  /* open temporary file */
  if (asprintf(&ssifile, "%s/ssi.XXXXXX", httpd->tmpDir) == -1) {
    munmap((String)file, (size_t)msg->file.status.st_size);
    wvSyslog(LOG_ERR, "[mod_ssi] failed to allocate memory\n");
    goto noparse;
  }
  fd = mkstemp(ssifile);	unlink(ssifile); free((String)ssifile);
  if (fd == -1) {
    munmap((String)file, (size_t)msg->file.status.st_size);
    wvSyslog(LOG_ERR, "[mod_ssi] failed to create temporary file\n");
    goto noparse;
  }

  /* open output file */
  fout = fdopen(fd, "r+");
  if (fout == NULL) {
    close(fd);
    munmap((String)file, (size_t)msg->file.status.st_size);
    wvSyslog(LOG_ERR, "[mod_ssi] failed to open file descriptors\n");
    goto noparse;
  }

  /* parsing */
  text = directive = file;
  dirp = NULL;
  txtc = 0;
  st   = ST_GND;
  for (ch = file; *ch != EOL; ch++) {
    switch (st) {
      case ST_LBR:
	txtc++;
	if (*ch == '!')	{ st = ST_EXC; }
	else		{ st = ST_GND; }
	break;
      case ST_EXC:
	txtc++;
	if (*ch == '-')	{ st = ST_HYP1; }
	else		{ st = ST_GND; }
	break;
      case ST_HYP1:
	txtc++;
	if (*ch == '-')	{ st = ST_HYP2; }
	else		{ st = ST_GND; }
	break;
      case ST_HYP2:
	txtc++;
	if (*ch == '#')	{ st = ST_CMD; txtc -= 5; directive = &ch[1];}
	else		{ st = ST_GND;
			  fwrite(text, 1, txtc, fout);
			  text = &ch[1];
			  txtc = 0;
			}
	break;
      case ST_CMD:
	dirp = ch;
	if (*ch == '-') {
	  if (ch[1] == '-' && ch[2] == '>') {
	    ch += 3;
	    fwrite(text, 1, txtc, fout);
	    text  = ch;
	    txtc  = 1;
	    *dirp = EOL;
	    ssiExecDirective(msg, fout, directive);
	    directive	= NULL;
	    dirp	= NULL;
	    st		= ST_GND;
	  }
	}
	break;
      case ST_GND:
      default:
	if (*ch == '<') {
	  st = ST_LBR;
	}
	txtc++;
	break;
    }
  }
  fputs(text, fout);
  fflush(fout);
  munmap((String)file, (size_t)msg->file.status.st_size);

  if (fstat(fd, &sbuf) == -1) {
    close(fd);
    wvSyslog(LOG_ERR, "[mod_ssi] failed to get file status\n");
    goto noparse;
  }

#if	DEBUG
  fprintf(stderr, "[ssi] Content-Length: %d\n", (int)sbuf.st_size);
#endif	/* DEBUG */
  close(msg->file.fd);
  msg->file.fd = fd;
  msg->file.status.st_size	= sbuf.st_size;
  msg->response.content_length	= sbuf.st_size;
noparse:
  msg->response.status		= STATUS_REQUEST_OK;
  if (headerLookup(msg->response.headers, "Content-Type") == NULL) {
    msg->file.mime		= HTML_MAGIC_TYPE;
  }
  else {
    (void)responseHeaderEntry(&msg->response.headers,
			      "Content-Type", HTML_MAGIC_TYPE);
  }
  (void)lseek(msg->file.fd, (off_t)0L, SEEK_SET);

  return RET_OK;
}

/*
 * FUNCTION NAME:	ssiExecDirective
 * DESCRIPTION:		execute directive
 * ARGUMENTS:		message, mapped buffer
 * RETURN VALUE:	none
 */
static void
ssiExecDirective(msg, fout, cmd)
	register Message *	msg;
	register FILE *		fout;
	String			cmd;
{
  String	directive;
  String	tag;
  String	val;
  String	p;

#if	DEBUG
  fprintf(stderr, "[ssiExecDirective] %s\n", cmd);
#endif	/* DEBUG */
  /* SSI directive */
  directive  = cmd;
  directive += strspn(cmd, " \t\r\n");

  /* SSI tag */
  tag = (String)strpbrk(directive, " \t\r\n");
  if (tag == NULL) {
    wvSyslog(LOG_ERR, "illegal SSI format: %s\n", directive);
    return;
  }
  *tag++ = EOL;
  tag	+= strspn(tag, " \t\r\n=");

  /* SSI value */
  val = (String)strpbrk(tag, " \t\r\n=");
  if (val == NULL) {
    wvSyslog(LOG_ERR, "SSI %s needs a tag\n", directive);
    return;
  }
  *val++ = EOL;
  val	 = (String)strchr(val, '"');
  if (val++ == NULL) {
    wvSyslog(LOG_ERR, "SSI %s's %s needs a value\n", directive, tag);
    return;
  }
  p = (String)strchr(val, '"');
  if (p == NULL) {
    wvSyslog(LOG_ERR, "SSI %s's value of %s is not closed by \".\n",
	     directive, tag);
    return;
  }
  *p = EOL;

  /* act by the directive */
  if (wvStrcmp(directive, "config") == 0) {
    do_config(msg, tag, val);
  }
  else if (wvStrcmp(directive, "include") == 0) {
    do_include(msg, fout, tag, val);
  }
  else if (wvStrcmp(directive, "echo") == 0) {
    do_echo(msg, fout, tag, val);
  }
  else if (wvStrcmp(directive, "fsize") == 0) {
    do_fsize(msg, fout, tag, val);
  }
  else if (wvStrcmp(directive, "flastmod") == 0) {
    do_flastmod(msg, fout, tag, val);
  }
  else if (wvStrcmp(directive, "exec") == 0) {
    do_exec(msg, fout, tag, val);
  }
  else {
    ssiError(ET_FORMAT, directive);
  }
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	do_config
 * DESCRIPTION:		address SSI config format
 * ARGUMENTS:		message, tag, val
 * RETURN VALUE:	none
 */
static void
do_config(msg, tag, val)
	register Message *	msg;
	const String		tag;
	const String		val;
{
  if (wvStrcmp(tag, "timefmt") == 0) {
    wvStrncpy(timefmt, val, TIMEFMT_LEN);
  }
  else if (wvStrcmp(tag, "sizefmt") == 0) {
    if (wvStrcmp(val, "bytes") == 0) {
      sizefmt = SZ_BYTES;
    }
    else if (wvStrcmp(val, "abbrev") == 0) {
      sizefmt = SZ_ABBREV;
    }
    else {
      ssiError(ET_VALUE, val);
    }
  }
  else {
    ssiError(ET_TAG, tag);
  }
}

/*
 * FUNCTION NAME:	do_include
 * DESCRIPTION:		address SSI include format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_include(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const String		tag;
	const String		val;
{
  register FILE *	fin;
  String		fpath;
  
  if (getFilePath(msg, tag, val, &fpath) == RET_NG) {
    wvSyslog(LOG_ERR, "SSI failed to get file path: %s=%s\n", tag, val);
    return;
  }

  fin = wvSafeFOpen(fpath, "r");	free((String)fpath);
  if (fin != NULL) {
    char	buf[BUF_SIZE];

    while (!feof(fin)) {
      size_t	size = fread(buf, sizeof(char), (size_t)sizeof(buf), fin);
      fwrite(buf, sizeof(char), size, fout);
    }
    fclose(fin);
  }
}

/*
 * FUNCTION NAME:	do_echo
 * DESCRIPTION:		address SSI echo format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_echo(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const String		tag;
	const String		val;
{
  time_t	cur_time;

  if (wvStrcmp(tag, "var") != 0) {
    ssiError(ET_TAG, tag);
    return;
  }

  time(&cur_time);	/* set current time in cur_time */
  if (wvStrcmp(val, "DOCUMENT_NAME") == 0) {
    /* the current filename */
    String	p = (String)strrchr(msg->file.path, '/');
    if (p != NULL) {
      fputs(p + 1, fout);
    }
  }
  else if (wvStrcmp(val, "DOCUMENT_URI") == 0) {
    /* the virtual path to this document (such as /docs/tutorials/foo.shtml) */
    fputs(msg->request.uri, fout);
  }
  else if (wvStrcmp(val, "QUERY_STRING_UNESCAPED") == 0) {
    if (msg->request.query_string != NULL) {
      fputs(msg->request.query_string, fout);
    }
  }
  else if (wvStrcmp(val, "DATE_LOCAL") == 0) {
    /* the current date, local time zone */
    printFormattedTime(msg, fout, cur_time, FALSE);	/* localtime */
  }
  else if (wvStrcmp(val, "DATE_GMT") == 0) {
    /* the current date in Greenwich mean  time */
    printFormattedTime(msg, fout, cur_time, TRUE);	/* Greenwich */
  }
  else if (wvStrcmp(val, "LAST_MODIFIED") == 0) {
    struct stat	sbuf;

    if (lstat(msg->file.path, &sbuf) == 0) {
      /* the last modification date of the current document */
      printFormattedTime(msg, fout, sbuf.st_mtime, FALSE);	/* localtime */
    }
  }
}

/*
 * FUNCTION NAME:	do_fsize
 * DESCRIPTION:		address SSI fsize format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_fsize(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const String		tag;
	const String		val;
{
  struct stat	sbuf;
  String	fpath;
  int		ret;

  if (getFilePath(msg, tag, val, &fpath) == RET_NG) {
    wvSyslog(LOG_ERR, "SSI failed to get file path: %s=%s\n", tag, val);
    return;
  }

  ret = lstat(fpath, &sbuf);	free((String)fpath);
  if (ret == -1) {
    wvSyslog(LOG_ERR, "SSI failed to get file status: %s\n", fpath);
    return;
  }

  switch (sizefmt) {
    case SZ_BYTES:
      fprintf(fout, "%ld", (long)sbuf.st_size);
      break;
    case SZ_ABBREV:
      if (sbuf.st_size < 1024) {
	fprintf(fout, "%ld", (long)sbuf.st_size);
      }
      else if (sbuf.st_size < 1024 * 1024) {
	fprintf(fout, "%ld.%ldK", (long)sbuf.st_size / U_KILO,
				  ((long)sbuf.st_size % U_KILO) * 10 / U_KILO);
      }
      else if (sbuf.st_size < 1024 * 1024 * 1024) {
	fprintf(fout, "%ld.%ldM", (long)sbuf.st_size / U_MEGA,
				  ((long)sbuf.st_size % U_MEGA) * 10 / U_MEGA);
      }
      else {
	fprintf(fout, "%ld.%ldG", (long)sbuf.st_size / U_GIGA,
				  ((long)sbuf.st_size % U_GIGA) * 10 / U_GIGA);
      }
    default:
      break;
  }
}

/*
 * FUNCTION NAME:	do_flastmod
 * DESCRIPTION:		address SSI flastmod format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_flastmod(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const String		tag;
	const String		val;
{
  struct stat	sbuf;
  String	fpath;
  int		ret;

  if (getFilePath(msg, tag, val, &fpath) == RET_NG) {
    wvSyslog(LOG_ERR, "SSI failed to get file path: %s=%s\n", tag, val);
    return;
  }

  ret = lstat(fpath, &sbuf);	free((String)fpath);
  if (ret == -1) {
    wvSyslog(LOG_ERR, "SSI failed to get file status: %s\n", fpath);
    return;
  }

  printFormattedTime(msg, fout, sbuf.st_mtime, FALSE);	/* localtime */
}

/*
 * FUNCTION NAME:	do_exec
 * DESCRIPTION:		address SSI exec format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_exec(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const String		tag;
	const String		val;
{
  if (wvStrcmp(tag, "cmd") == 0) {
    ssiRunProgram(msg, fout, val);
  }
#if	0
  else if (wvStrcmp(tag, "cgi") == 0) {
    /* not implemented... */
  }
#endif
  else {
    ssiError(ET_TAG, tag);
  }
}

/*
 * FUNCTION NAME:	ssiRunProgram
 * DESCRIPTION:		execute a program for SSI's cmd
 * ARGUMENTS:		message, output file pointer, command
 * RETURN VALUE:	none
 */
#define	READ_FD		(0)	/* descriptor to read */
#define	WRITE_FD	(1)	/* descriptor to write */
#define	MAX_ARGS	(5)	/* maximum number of arguments for program */
static void
ssiRunProgram(msg, fout, cmd)
	register Message *	msg;
	register FILE *		fout;
	const String		cmd;
{
  String		str = cmd;
  String		args[MAX_ARGS];
  String *		ap;
  int			c;
  int			pfd[2];
  pid_t			pid;
  char			cwd[MAX_PATH_LEN + 1];	/* command working directory */
  String		p;
  struct sigaction	act;
  struct sigaction	oact;
  /*
   * shell meta characters
   * \134 = backslash
   * \042 = double quotation
   * \012 = new line
   * \015 = carriage return
   */
  const char	sh_meta[] = "&;`'\134\042|*?~<>^()[]{}$\012\015";

  /* Does cmd include shell meta characters? */
  if (strpbrk(cmd, sh_meta) != NULL) {
    wvSyslog(LOG_SECURITY,
	     "unsecure/untrusted command string [%s] for #cmd in SSI\n", cmd);
    return;
  }

  /* parsing arguments */
  for (c = 0, ap = args; (*ap = strsep(&str, " \t")) != NULL; c++) {
    args[c] = *ap;
    if (**ap != EOL) {
      if (++ap >= &args[MAX_ARGS]) {
	wvSyslog(LOG_ERR,
		 "unexpected too many arguments [%s] for #cmd in SSI\n", cmd);
	return;	/* error: arguments overflow */
      }
    }
  }
  args[c] = EOL;

  /* set working directory */
  memset(cwd, 0, sizeof(cwd));
  if (wvStrncpy(cwd, msg->file.path, sizeof(cwd)) == NULL) {
    return;
  }
  p = strrchr(cwd, '/');
  if (p != NULL) {
    *p = EOL;
  }

  /* make pipeline between parent process and child process */
  if (pipe(pfd) == -1) {
    return;
  }

  /* signal */
  sigemptyset(&act.sa_mask);
  act.sa_flags	 = SA_NOCLDSTOP;
  act.sa_handler = (void *)ssiReapChild;
  sigaction(SIGCHLD, &act, &oact);

  /* execute a program */
  pid = vfork();
  switch (pid) {
    case -1:	/* fork error */
      wvSyslog(LOG_ERR, "failed to fork process for [%s] #cmd in SSI\n", cmd);
      break;

    case 0:	/* child process */
      wvClose(pfd[READ_FD]);
      dup2(pfd[WRITE_FD], STDERR_FILENO);
      dup2(pfd[WRITE_FD], STDOUT_FILENO);
      if (chdir(cwd) == -1 ||
	  execve(args[0], args, msg->cgi.env) == -1) {
	wvSyslog(LOG_ERR, "failed to exec [%s] #cmd in SSI\n", args[0]);
      }
      _exit(1);

    default: {	/* parent process */
	register FILE *	fin;

	wvClose(pfd[WRITE_FD]);
	fin = fdopen(pfd[READ_FD], "r");
	if (fin != NULL) {
	  char	buf[BUF_SIZE];

	  while (!feof(fin)) {
	    size_t size = fread(buf, sizeof(char), (size_t)sizeof(buf), fin);
	    fwrite(buf, sizeof(char), size, fout);
	  }
	  fclose(fin);
	}
      }
      break;
  }
  sigaction(SIGCHLD, &oact, NULL);
}

static void
ssiReapChild(sig)
	int	sig;
{
  pid_t		pid;
  int		status;

  while ((pid = waitpid((pid_t)0, &status, WNOHANG)) > 0) {
#if	DEBUG
    fprintf(stderr, "[ssi #exec] pid=%d, status=%d\n",
	    pid, WEXITSTATUS(status));
#endif	/* DEBUG */
  }
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	printFormattedTime
 * DESCRIPTION:		print the formatted time
 * ARGUMENTS:		message, output file pointer,
 *			time, gmt flag (TRUE: GMT, FALSE: Localtime)
 * RETURN VALUE:	none
 */
static void
printFormattedTime(msg, fout, t, gmt)
	register Message *	msg;
	register FILE *		fout;
	time_t			t;
	bool_t			gmt;
{
  struct tm *	tp;
  char		tbuf[128];

  tp = gmt ? gmtime(&t) : localtime(&t);
  memset(tbuf, 0, sizeof(tbuf));
  if (strftime(tbuf, sizeof(tbuf), timefmt, tp) > 0) {
    fputs(tbuf, fout);
  }
}

/*
 * FUNCTION NAME:	getFilePath
 * DESCRIPTION:		get the filepath
 * ARGUMENTS:		message, tag, val, ret-val
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
getFilePath(msg, tag, val, ret)
	register Message *	msg;
	const String		tag;
	const String		val;
	String *		ret;
{
  if (wvStrcmp(val, USER_ACCESS) == 0 || wvStrcmp(val, USER_PASSWD) == 0) {
    wvSyslog(LOG_SECURITY, "invalid file access [%s] in SSI\n", val);
    return RET_NG;
  }

  if (wvStrcmp(tag, "file") == 0) {
    String	tmp;
    int		n = -1;

    if (val[0] == '/' || val[0] == '.') {
      wvSyslog(LOG_SECURITY, "invalid file access [%s] in SSI\n", val);
      return RET_NG;
    }

    tmp = wvStrdup(msg->file.path);
    if (tmp != NULL) {
      String	p;
      
      p = strrchr(tmp, '/');
      if (p != NULL) {
	*p = EOL;
      }
      n = asprintf(ret, "%s/%s", tmp, val);
      free((String)tmp);
    }
    return n == -1 ? wvNoMemoryError(msg) : RET_OK;
  }
  else if (wvStrcmp(tag, "virtual") == 0) {
    int	n;

    if (msg->user.homeDir != NULL) {
      n = asprintf(ret, "%s/%s", msg->user.homeDir, val);
    }
    else {
      n = asprintf(ret, "%s/%s", msg->vhost->docRoot, val);
    }
    return n == -1 ? wvNoMemoryError(msg) : RET_OK;
  }
  else {
    return ssiError(ET_TAG, tag);
  }

  return RET_OK;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	ssiError
 * DESCRIPTION:		no support SSI tag/value/format
 * ARGUMENTS:		type, name
 * RETURN VALUE:	always: -1
 */
static int
ssiError(type, name)
	error_ssi_t	type;
	const String	name;
{
  switch (type) {
    case ET_TAG:
      wvSyslog(LOG_WARNING, "no support SSI tag: %s\n", name);
      break;
    case ET_VALUE:
      wvSyslog(LOG_WARNING, "no support SSI value: %s\n", name);
      break;
    case ET_FORMAT:
      wvSyslog(LOG_WARNING, "no support SSI format: #%s\n", name);
    default:
      break;
  }

  return RET_NG;
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
ssiRegisterHooks(void)
{
  wvHookTypeChecker(ssiTypeChecker, HOOK_MIDDLE);
  wvHookFilter(ssiFilter, HOOK_FIRST, SSI_MAGIC_TYPE);
}

Module	mod_ssi = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_VULNERABLE,		/* safety */
  "mod_ssi",			/* name */
  "SSI Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)ssiRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_ssi;
}
