/*****************************************************************************
 *
 * FILE:	network.c
 * DESCRIPTION:	lookup the client host
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: network.c,v 1.38.2.13 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <sys/types.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#if	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif	/* HAVE_ARPA_INET_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static ListenAddr *	lookup(const String, const int);
static bool_t		listenAddrInconsistency(void);
static String		getFQDN(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	socketEntry
 * Description:		register bind-address, port into ListenAddr list
 * Arguments:		bind address, port, socket type
 * Return Value:	none
 */
void
socketEntry(ipaddr, port, type)
	const String	ipaddr;
	const int	port;
	socket_t	type;
{
  if (port <= 0) {
    return;	/* maybe only SSL connection */
  }

  if (lookup(ipaddr, port) == NULL) {	/* no conflicted */
    ListenAddr *	new;

    new		  = (ListenAddr *)wvMalloc(sizeof(ListenAddr));
    new->ipaddr	  = wvStrdup(ipaddr);
    if (asprintf(&new->port, "%d", port) == -1) {
      die(ERR_OUT_OF_MEMORY);
    }
    new->type	  = type;
    new->next	  = httpd->listen;
    httpd->listen = new;
  }
}

static ListenAddr *
lookup(ipaddr, port)
	const String	ipaddr;
	const int	port;
{
  register ListenAddr *	lap;

  for (lap = httpd->listen; lap != NULL; lap = lap->next) {
    if (port == atoi(lap->port)) {
      if (ipaddr == NULL || lap->ipaddr == NULL) {
	return lap;
      }
    }
  }
  return NULL;
}


/*
 * Function Name:	openChannels
 * Description:		open available all communication channels
 * Arguments:		none
 * Return Value:	none
 */
void
openChannels(void)
{
  bool_t		inconsistency;
  register ListenAddr *	lap;

  inconsistency = listenAddrInconsistency();
  for (lap = httpd->listen; lap != NULL; lap = lap->next) {
    struct addrinfo	hints;
    struct addrinfo *	res;
    struct addrinfo *	ai;

    if (inconsistency == TRUE && lap->ipaddr == NULL) {
      continue;
    }

    memset(&hints, 0, sizeof(hints));
    hints.ai_family	= AF_UNSPEC;	/* IPv4/IPv6 */
    hints.ai_socktype	= SOCK_STREAM;
    hints.ai_protocol	= IPPROTO_TCP;
    hints.ai_flags	= lap->ipaddr == NULL ? AI_PASSIVE : AI_NUMERICHOST;

    if (getaddrinfo(lap->ipaddr, lap->port, &hints, &res) != 0) {
      continue;
    }

    for (ai = res; ai != NULL; ai = ai->ai_next) {
      int	s;
      int	on;
      Socket *	soc;

      /* create a socket */
      s = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
      if (s < 0) {
	continue;
      }

      /* non-blocking socket */
      if (setNonBlocking(s) == -1) {
	close(s);
	continue;
      }

      /* allow reuse of local addresses */
      on = 1;
      if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
		     (void *)&on, sizeof(on)) == -1) {
	close(s);
	continue;
      }
#if	defined(DEBUG) || defined(SO_REUSEPORT)
      if (setsockopt(s, SOL_SOCKET, SO_REUSEPORT,
		     (void *)&on, sizeof(on)) == -1) {
	close(s);
	continue;
      }
#endif	/* DEBUG || SO_REUSEPORT */

      /* bind a socket */
      if (bind(s, ai->ai_addr, ai->ai_addrlen) < 0) {
	close(s);
	continue;
      }

      /* listen for connection on the socket */
      if (listen(s, httpd->maxRequest) < 0) {
	close(s);
	continue;
      }

      soc		= (Socket *)wvMalloc(sizeof(Socket));
#if	USE_SSL
      if (lap->type == SOC_SECURE) {
	soc->ssl_ctx	= sslOpenChannel();
      }
      soc->ssl		= NULL;
#endif	/* USE_SSL */
      soc->sd		= s;
      soc->type		= lap->type;
      soc->port		= atoi(lap->port);
      soc->events	= POLL_IN;
      soc->revents	= 0;
      soc->ainfo	= *ai;
      soc->next		= httpd->sockets;
      httpd->sockets	= soc;

      if (s > httpd->maxsoc) {
	httpd->maxsoc = s;
      }
#if	DEBUG
      fprintf(stderr, "[channels] socket(%d)\n", soc->sd);
#endif	/* DEBUG */
    }

    if (res != NULL) {
      freeaddrinfo(res);
    }
  }

  if (httpd->maxsoc < 0) {
    die("failed to open any channels\n");
  }
}

/*
 * Function Name:	listenAddrInconsistency
 * Description:		check inconsistency address
 * Arguments:		none
 * Return Value:	inconsistency: TRUE, otherwise: FALSE
 */
static bool_t
listenAddrInconsistency(void)
{
  register ListenAddr *	lap;
  int			wildcard = 0;
  int			bindhost = 0;

  for (lap = httpd->listen; lap != NULL; lap = lap->next) {
    if (lap->ipaddr == NULL) {
      wildcard++;
    }
    else {
      bindhost++;
    }
  }

  return (wildcard > 0 && bindhost > 0) ? TRUE : FALSE;
}

/*
 * Function Name:	setNonBlocking
 * Description:		change new socket descriptor to non-blocking mode
 * Arguments:		socket descriptor
 * Return Value:	successful: 0, otherwise: -1
 */
int
setNonBlocking(sd)
	const int	sd;
{
  int	flags;

  flags = fcntl(sd, F_GETFL, 0);
  if (flags == -1) {
    return RET_NG;
  }
  flags |= O_NONBLOCK;
  if (fcntl(sd, F_SETFL, flags) == -1) {
    return RET_NG;
  }
  return RET_OK;
}

/*
 * Function Name:	closeChannels
 * Description:		close all opened communication channels
 * Arguments:		none
 * Return Value:	none
 */
void
closeChannels(void)
{
  register Socket *	sp;

  for (sp = httpd->sockets; sp != NULL; sp = sp->next) {
#if	USE_SSL
    if (sp->type == SOC_SECURE && sp->ssl_ctx) {
      SSL_CTX_free(sp->ssl_ctx);
    }
#endif	/* USE_SSL */
    shutdown(sp->sd, SHUT_RDWR);	/* shutdown full-duplex connection */
    wvClose(sp->sd);	/* close the main socket */
  }
}

/*****************************************************************************/

/*
 * Function Name:	getFQDN
 * Description:		get Full-Qualified Domain Name
 * Arguments:		none
 * Return Value:	successful: FQDN, otherwise: NULL
 */
static String
getFQDN(void)
{
  struct hostent *	hp;
  char			name[MAX_HOSTNAME_LEN + 1];

  memset(name, 0, sizeof(name));
  if (gethostname(name, sizeof(name)) != 0) {
    return NULL;	/* failed to get hostname */
  }

  hp = gethostbyname(name);
  if (hp == NULL) {
    goto done;
  }

  if (strchr(hp->h_name, '.') == NULL) {/* no dot in hostname */
    if (hp->h_aliases != NULL) {	/* I have an alias list */
      int	i;
      for (i = 0; hp->h_aliases[i] != NULL; i++) {
	if (hp->h_aliases[i] != NULL &&
	    strchr(hp->h_aliases[i], '.') != NULL &&
	    (wvStrncasecmp(hp->h_aliases[i], hp->h_name,
			   strlen(hp->h_name)) == 0)) {
	  return wvStrdup(hp->h_aliases[i]);
	}
      }
    }
  }

done:
  return wvStrdup(name);
}

/*
 * Function Name:	getServerName
 * Description:		get server name
 * Arguments:		none
 * Return Value:	server name
 */
String
getServerName(void)
{
  String	name;

  name = getFQDN();
  if (name != NULL) {
    return name;
  }
  else {	/* failed to get hostname, fall back to the loop-back address */
    return wvStrdup("localhost");
  }
}

