/*****************************************************************************
 *
 * FILE:	security.c
 * DESCRIPTION:	security functions
 * DATE:	Tue, Aug  6 2002
 * UPDATE:	Sun, Dec 19 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(c) 2002-2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (c) 2002-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: security.c,v 1.1.2.18 2004/12/19 13:55:27 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
#ifndef	HAVE_LCHMOD
static int	lchmod(const char *, mode_t);
#endif	/* HAVE_LCHMOD */
#ifndef	HAVE_LCHOWN
static int	lchown(const char *, uid_t, gid_t);
#endif	/* HAVE_LCHOWN */
static void	checkDirsCondition(void);
static void	alienateProperty(void);
//static void	dropPrivilegeEternally(void);
static int	checkSafePath(Message *, const String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static uid_t	saved_uid;	/* effective user ID */
static gid_t	saved_gid;	/* effective group ID */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

#ifndef	HAVE_LCHMOD
static int
lchmod(path, mode)
	const char *	path;
	mode_t		mode;
{
  struct stat	sbuf;

  if (lstat(path, &sbuf) == 0 && S_ISLNK(sbuf.st_mode)) {
    errno = ENOSYS;
    return -1;
  }

  return chmod(path, mode);
}
#endif	/* HAVE_LCHMOD */

#ifndef	HAVE_LCHOWN
static int
lchown(file, uid, gid)
	const char *	file;
	uid_t		uid;
	gid_t		gid;
{
  struct stat	sbuf;

  if (lstat(file, &sbuf) == 0 && S_ISLNK(sbuf.st_mode)) {
    errno = ENOSYS;
    return -1;
  }

  return chown(file, uid, gid);
}
#endif	/* HAVE_LCHOWN */

/*
 * Function Name:	enforceSecureLevel
 * Description:		enforce secure level
 * Arguments:		none
 * Return Value:	none
 */
void
enforceSecureLevel(void)
{
  register VirtualHost *	vp;

#if	DEBUG
  fprintf(stderr, "[enforceSecureLevel] SecureLevel=%d\n", httpd->secureLevel);
#endif	/* DEBUG */
  switch (httpd->secureLevel) {
    case SECURE_LEVEL7:	/* disable virtual hosting */
      vhostDelete();
    case SECURE_LEVEL6:	/* disable all modules */
      moduleDelete();
    case SECURE_LEVEL5:	/* forbid accessing UserDir */
      for (vp = httpd->host; vp != NULL; vp = vp->next) {
	wvFree((String)vp->userDir);
	vp->userDir = NULL;
      }
    case SECURE_LEVEL4:	/* disable modules to make dynamic pages */
      moduleUnloadVulnerable();
    case SECURE_LEVEL3:	/* forbid accessing symbolic linked files */
      for (vp = httpd->host; vp != NULL; vp = vp->next) {
	vp->symlinks = FALSE;
      }
    case SECURE_LEVEL2:	/* change owner and permission */
      alienateProperty();
    case SECURE_LEVEL1:	/* forbid running as root:wheel */
      /* dropPrivilegeEternally(); was moved to main() in wyvern.c */
    case SECURE_LEVEL0:
      checkDirsCondition();
    default:
      break;
  }
}

/*
 * Function Name:	checkDirsCondition
 * Description:		confirm existence of CacheDir/TmpDir directories
 * Arguments:		none
 * Return Value:	none
 */
static void
checkDirsCondition(void)
{
  struct stat	sbuf;

  /* check a temporary directory */
  if (lstat(httpd->tmpDir, &sbuf) == -1 && errno == ENOENT) {
    die("No temporary directory: %s\n", httpd->tmpDir);
  }
  if (S_ISLNK(sbuf.st_mode)) {
    die("%s: untrusted symbolic link\n", httpd->tmpDir);
  }

  /* check a cache directory */
  if (lstat(httpd->cacheDir, &sbuf) == -1 && errno == ENOENT) {
    die("No cache directory: %s\n", httpd->cacheDir);
  }
  if (S_ISLNK(sbuf.st_mode)) {
    die("%s: untrusted symbolic link\n", httpd->cacheDir);
  }
}

/*
 * Function Name:	alienateProperty
 * Description:		change owner/group and permission
 * Arguments:		none
 * Return Value:	none
 * Security Level:	level2
 */
static void
alienateProperty(void)
{
  if (getuid() == UID_ROOT) {
    register VirtualHost *	vp;

    /* temporary dir */
    if (lchown(httpd->tmpDir, httpd->user, httpd->group) == -1) {
      die("failed to change owner of TmpDir: %s\n", httpd->tmpDir);
    }
    if (lchmod(httpd->tmpDir, S_IRWXU|S_IRGRP|S_IXGRP) == -1) {
      die("failed to change mode of TmpDir: %s\n", httpd->tmpDir);
    }

    /* cache dir */
    if (lchown(httpd->cacheDir, httpd->user, httpd->group) == -1) {
      die("failed to change owner of CacheDir: %s\n", httpd->cacheDir);
    }
    if (lchmod(httpd->cacheDir, S_IRWXU|S_IRGRP|S_IXGRP) == -1) {
      die("failed to change mod of CacheDir: %s\n", httpd->cacheDir);
    }

    /* log files */
    if (fchown(fileno(httpd->securityLog.fp),
		httpd->user, httpd->group) == -1) {
      die("failed to change owner of securityLog: %s\n",
	  httpd->securityLog.path);
    }
    for (vp = httpd->host; vp != NULL; vp = vp->next) {
      if (fchown(fileno(vp->errorLog.fp), httpd->user, httpd->group) == -1) {
	die("failed to change owner of errorLog: %s\n", vp->errorLog.path);
      }
      if (fchown(fileno(vp->accessLog.fp), httpd->user, httpd->group) == -1) {
	die("failed to change owner of accessLog: %s\n", vp->accessLog.path);
      }
    }
  }
}

/*
 * Function Name:	dropPrivilegeEternally
 * Description:		drop the privilege of wyvern process
 * Arguments:		none
 * Return Value:	none
 * Security Level:	level1
 */
void
dropPrivilegeEternally(void)
{
  if (httpd->secureLevel > SECURE_LEVEL0) {
    /* drop privilege */
    if (httpd->user == UID_ROOT) {
      httpd->user = UID_NOBODY;
    }
    if (httpd->group == GID_WHEEL) {
      httpd->group = GID_NOGROUP;
    }

    /* change owner/group */
    if (getuid() == UID_ROOT) {
      struct passwd *	pwd;

      pwd = getpwuid(httpd->user);
      if (pwd == NULL) {
	die("No user information in password database: uid = %d\n",
	    httpd->user);
      }

      if (initgroups(pwd->pw_name, (int)pwd->pw_gid) == -1) {
	die("failed to initialize group access list\n");
      }

      if (setgid(httpd->group) == -1) {
	die("failed to change gid %d\n", httpd->group);
      }

      if (setuid(httpd->user) == -1) {
	die("failed to change uid %d\n", httpd->user);
      }
    }
    else {
      wvSyslog(LOG_WARNING,
	      "No running as root: failed to change uid %d and gid %d\n",
	      httpd->user, httpd->group);
      httpd->user  = getuid();
      httpd->group = getgid();
    }
  }
}

/*
 * Function Name:	dropPrivilege
 * Description:		drop the privilege of wyvern process
 * Arguments:		none
 * Return Value:	successful: 0, otherwise: -1
 */
int
dropPrivilege(void)
{
  if (getuid() == UID_ROOT) {
    uid_t	exec_uid;
    gid_t	exec_gid;

    /* save effective uid/gid */
    saved_uid = geteuid();
    saved_gid = getegid();

    exec_uid = httpd->user  == UID_ROOT  ? UID_NOBODY  : httpd->user;
    exec_gid = httpd->group == GID_WHEEL ? GID_NOGROUP : httpd->group;

    /* drop privilege */
    if (setegid(exec_gid) == -1) {
      wvSyslog(LOG_ERR, "failed to change effective gid %d\n", exec_gid);
      return RET_NG;
    }
    if (seteuid(exec_uid) == -1) {
      wvSyslog(LOG_ERR, "failed to change effective uid %d\n", exec_uid);
      return RET_NG;
    }
  }
  return RET_OK;
}

/*
 * Function Name:	raisePrivilege
 * Description:		raise the privilege of wyvern process
 * Arguments:		none
 * Return Value:	successful: 0, otherwise: -1
 */
int
raisePrivilege(void)
{
  if (getuid() == UID_ROOT) {
    if (setegid(saved_gid) == -1) {
      wvSyslog(LOG_WARNING, "failed to change effective gid %d\n", saved_gid);
      return RET_NG;
    }
    if (seteuid(saved_uid) == -1) {
      wvSyslog(LOG_WARNING, "failed to change effective uid %d\n", saved_uid);
      return RET_NG;
    }
  }
  return RET_OK;
}

/*****************************************************************************/

/*
 * Function Name:	safePathChecker
 * Description:		track a symbolic link file
 * Arguments:		message, file path
 * Return Value:	Safe path: 0, otherwise: -1
 */
#define	MAX_SYMLINKS	5
int
safePathChecker(msg, path)
	register Message *	msg;
	const String		path;
{
  int		safety	 = RET_NG;
  static int	symlinks = 0;
  struct stat	sbuf;

  if (lstat(path, &sbuf) == 0) {
    if (S_ISLNK(sbuf.st_mode)) {	/* symbolic link */
      char	realPath[MAX_PATH_LEN + 1];

      if (++symlinks > MAX_SYMLINKS) {
	goto error;
      }
      memset(realPath, 0, sizeof(realPath));
      if (readlink(path, realPath, sizeof(realPath)) == -1) {
	goto error;
      }
      return safePathChecker(msg, realPath);
    }
    safety = checkSafePath(msg, path);
  }
  else if (path[0] != '/' && path[0] != '.' &&
	   path[1] != '/' && path[1] != '.') {
#if	DEBUG
    fprintf(stderr, "[safe] path = %s\n", path);
#endif	/* DEBUG */
    safety = RET_OK;
  }
  else if (symlinks == 0 && errno == ENOENT) {
    msg->response.status = STATUS_NOT_FOUND;
  }
  else {
    msg->response.status = STATUS_FORBIDDEN;
  }

  if (symlinks > 0) {
    symlinks = 0;
    wvSyslog(LOG_SECURITY,
	     "request URI '%s' is %s symbolic file '%s'.\n",
	     msg->request.uri,
	     safety == RET_OK ? "a" : "an unsecure", path);
  }
  return safety;

error:
  msg->response.status = STATUS_FORBIDDEN;
  return RET_NG;
}

/*
 * Function Name:	checkSafePath
 * Description:		check the path which is safe
 * Arguments:		message, file path
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
checkSafePath(msg, path)
	register Message *	msg;
	const String		path;
{
  /* in DocumentRoot? */
  if (wvStrncmp(path, msg->vhost->docRoot, strlen(msg->vhost->docRoot)) == 0) {
    return RET_OK;
  }

  /* in alias? */
  if (msg->vhost->aliasTable != NULL) {
    register Alias *	ap;

    for (ap = msg->vhost->aliasTable; ap != NULL; ap = ap->next) {
      if (ap->type == A_REDIRECT) {
       	continue;
      }
      if (wvStrncmp(path, ap->realName, ap->realLen) == 0) {
	return RET_OK;
      }
    }
  }

  /* in ServerRoot? */
  if (wvStrncmp(path, httpd->serverRoot, strlen(httpd->serverRoot)) == 0) {
    return RET_OK;
  }

  /* in UserDir? */
  if (msg->user.indexDir != NULL &&
      wvStrncmp(path, msg->user.indexDir, strlen(msg->user.indexDir)) == 0) {
    return RET_OK;
  }

  msg->response.status = STATUS_FORBIDDEN;

  return RET_NG;
}

/*****************************************************************************/

/*
 * Function Name:	sanitizeURI
 * Description:		sanitize URI
 * Arguments:		uri string
 * Return Value:	sanitized uri string
 */
String
sanitizeURI(uri)
	String	uri;
{
#define	TMP_BUFSIZE	(1244)	/* MAX_URI_LEN * 4.86 */
  char			tmp[TMP_BUFSIZE];
  register String	s = tmp;
  char			c;

  memset(tmp, 0, sizeof(tmp));
  while ((c = *uri) != EOL) {
    /* escape meta characters for cross-site scripting */
    if (c == '<') {
      *s++ = '&'; *s++ = 'l'; *s++ = 't'; *s++ = ';';
    }
    else if (c == '>') {
      *s++ = '&'; *s++ = 'g'; *s++ = 't'; *s++ = ';';
    }
    else if (c == '"') {
      *s++ = '&'; *s++ = 'q'; *s++ = 'u'; *s++ = 'o'; *s++ = 't'; *s++ = ';';
    }
    else if (c == '\'') {
      *s++ = '&'; *s++ = '#'; *s++ = '3'; *s++ = '9'; *s++ = ';';
    }
    else if (c == '(') {
      *s++ = '&'; *s++ = '#'; *s++ = '4'; *s++ = '0'; *s++ = ';';
    }
    else if (c == ')') {
      *s++ = '&'; *s++ = '#'; *s++ = '4'; *s++ = '1'; *s++ = ';';
    }
    else if (c == '&') {
      *s++ = '&'; *s++ = 'a'; *s++ = 'm'; *s++ = 'p'; *s++ = ';';
    }
    /* normal characters */
    else {
      *s++ = c;
    }
    uri++;
  }
  *s = EOL;

  return wvStrdup(tmp);
}

/*****************************************************************************/

/*
 * Function Name:	wvSafeOpen
 * Description:		open file safety
 * Arguments:		filename, open flags, status buffer, symlink check
 * Return Value:	successful: file descriptor, otherwise: -1
 * Notes:		http://www.securityfocus.com/popups/forums/secprog/secure-programming.shtml
 */
int
wvSafeOpen(filename, open_flags, ret_sbuf, allow_symlink)
	const String	filename;
	int		open_flags;
	struct stat *	ret_sbuf;
	bool_t		allow_symlink;
{
  struct stat	lsbuf;
  struct stat	fsbuf;
  int		fd;

  /* clear errno */
  errno = 0;

  /* check filename */
  if (filename == NULL || filename[0] == EOL) {
    errno = ENOENT;
    return -1;
  }

  /* obtain the file's status information */
  if (lstat(filename, &lsbuf) == -1) {
    if (errno == ENOENT && !(open_flags & O_WRONLY)) {
      return -1;
    }
    memset((char *)&lsbuf, 0, sizeof(lsbuf));
  }
  else {
    /* make sure the file is regular file */
    if (!S_ISREG(lsbuf.st_mode)) {
      if (S_ISDIR(lsbuf.st_mode)) {
	errno = EISDIR;
	return -1;
      }
      if (S_ISLNK(lsbuf.st_mode) && !allow_symlink) {
	errno = EMLINK;
	return -1;
      }
      if (!S_ISLNK(lsbuf.st_mode) && allow_symlink) {
	errno = EACCES;
	return -1;
      }
    }

    /* make sure that the file is owned by user (not uid 0) */
    if (lsbuf.st_uid == UID_ROOT) {
      wvSyslog(LOG_SECURITY, "detected file access: %s is owned by root\n",
		filename);
    }
  }

  /* open the file */
again:
  fd = open(filename, open_flags, 0600);
  if (fd == -1) {
    if (errno == EOPNOTSUPP) {	/* no support locking mechanism */
      open_flags &= ~O_SHLOCK;	/* in the underlying filesystem */
      open_flags &= ~O_EXLOCK;	/* unset O_SHLOCK/O_EXLOCK and try again */
      goto again;
    }
    return -1;
  }

  /* now we fstat() the file, to make sure it's the same file still! */
  if (fstat(fd, &fsbuf) == -1) {
    close(fd);
    return -1;
  }

  /*
   * here we make sure the inode and device numbers are the same in the file
   * we actually opened, compared to the file we performed the inital
   * lstat() call on.
   */
  if (!allow_symlink &&	/* not allow symbolic link */
      (lsbuf.st_ino != 0 && lsbuf.st_dev != 0) &&	/* not create newly */
      (lsbuf.st_ino != fsbuf.st_ino || lsbuf.st_dev != fsbuf.st_dev)) {
    wvSyslog(LOG_SECURITY, "open failure: %s is not identical\n", filename);
    close(fd);
    errno = EACCES;
    return -1;
  }
  *ret_sbuf = fsbuf;

  return fd;
}

/*
 * Function Name:	wvSafeFOpen
 * Description:		open file safety
 * Arguments:		filename, open mode
 * Return Value:	successful: file pointer, otherwise: NULL
 */
FILE *
wvSafeFOpen(filename, mode)
	const String	filename;
	const String	mode;
{
  int		fd;
  int		flags;
  struct stat	sbuf;

  switch (*mode) {
    case 'w':
      flags = O_WRONLY|O_CREAT|O_TRUNC|O_EXLOCK;
      break;
    case 'a':
      flags = O_WRONLY|O_APPEND|O_CREAT|O_EXLOCK;
      break;
    default:
      flags = O_RDONLY|O_SHLOCK;
      break;
  }

  fd = wvSafeOpen(filename, flags, &sbuf, FALSE);
  if (fd == -1) {
    return NULL;
  }
  return fdopen(fd, mode);
}
