/*****************************************************************************
 *
 * FILE:	thread.c
 * DESCRIPTION:	manage the thread pool
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: thread.c,v 1.60.2.22 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	gc(void);
static void	init(Thread *);
static void	cleanupMessage(Thread *);
static void	cleanup(Message *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static Thread *	freeThread;	/* free thread in thread pool */
static Thread *	threadTop;	/* the top of thread pool (const) */
static int	usedThreads;	/* the number of used threads */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/


/*
 * Function Name:	threadPoolInit
 * Description:		create thread pool
 * Arguments:		none
 * Return Value:	none
 */
void
threadPoolInit(void)
{
  register Thread *	tp;
#if	DEBUG
  register int		c = 0;
#endif	/* DEBUG */

  /* allocate threads */
  threadTop  = (Thread *)wvCalloc((size_t)httpd->maxThread, sizeof(Thread));
  freeThread = threadTop;

  /* initialize each thread entity */
  for (tp = threadTop; tp < threadTop + httpd->maxThread; tp++) {
    init(tp);
    cleanup(&tp->mqueue);
#if	DEBUG
    tp->no = c++;
#endif	/* DEBUG */
    tp->next = tp + 1;
  }
  (--tp)->next = NULL;

  /* initialize the number of used threads */
  usedThreads = 0;
}

/*
 * Function Name:	threadNew
 * Description:		get a new thread from thread pool
 * Arguments:		socket descriptor
 * Return Value:	new thread
 */
Thread *
threadNew(sd)
	int	sd;
{
  Thread *	new;

#if	DEBUG
  fprintf(stderr, "[threadNew] sd = %d\n", sd);
#endif	/* DEBUG */
  if (freeThread == NULL) {
    if (gc() == 0) {
      return NULL;
    }
  }
  new			= freeThread;
  freeThread		= freeThread->next;
  new->status		= TH_ACTIVE;
  new->conn.sock.sd	= sd;
  /* connection */
  new->mqueue.status	= MSG_NONE;
  new->mqueue.vhost	= httpd->host;	/* first is default host */
  new->mqueue.conn	= &new->conn;
  usedThreads++;	/* count up */

  return new;
}

/*
 * Function Name:	gc
 * Description:		collect no-active threads
 * Arguments:		none
 * Return Value:	the number of collected threads
 */
static int
gc(void)
{
  register Thread *	tp;
  Thread *		sweeped = NULL;
  int			num = 0;

  for (tp = threadTop; tp < threadTop + httpd->maxThread; tp++) {
    if (tp->status == TH_ACTIVE) {
      continue;
    }
    if (tp->status == TH_TERMINATED) {
#if	DEBUG
      fprintf(stderr, "[GCed] thread #%d\n", tp->no);
#endif	/* DEBUG */
      init(tp);
      cleanupMessage(tp);
      num++;
    }
    tp->next = sweeped;
    sweeped  = tp;
  }
  freeThread = sweeped;

  /* the number of used threads */
  usedThreads -= num;

  return num;
}

/*
 * Function Name:	threadDetach
 * Description:		detach a thread
 * Arguments:		a detached thread
 * Return Value:	none
 */
void
threadDetach(thd)
	register Thread *	thd;
{
  thd->status = TH_TERMINATED;
#if	USE_SSL
  if (thd->conn.sock.type == SOC_SECURE) {
    SSL_shutdown(thd->conn.sock.ssl);
    SSL_clear(thd->conn.sock.ssl);
    SSL_free(thd->conn.sock.ssl);
    thd->conn.sock.ssl = NULL;
  }
#endif	/* USE_SSL */
  shutdown(thd->conn.sock.sd, SHUT_RDWR);
  wvClose(thd->conn.sock.sd);
  thd->conn.sock.sd = -1;
}

/*
 * Function Name:	threadActives
 * Description:		get active thread list
 * Arguments:		none
 * Return Value:	active threads
 */
Thread *
threadActives(void)
{
  register Thread *	tp;
  Thread *		actives = NULL;

  for (tp = threadTop; tp < threadTop + httpd->maxThread; tp++) {
    if (tp->status == TH_ACTIVE) {
      tp->next	= actives;
      actives	= tp;
    }
  }

  return actives;
}

/*
 * Function Name:	threadSweep
 * Description:		sweeping detached threads
 * Arguments:		none
 * Return Value:	number of active threads
 */
int
threadSweep(void)
{
  register Thread *	tp;

  /* check timeout */
  for (tp = threadActives(); tp != NULL; tp = tp->next) {
    (void)gettimeofday(&tp->time.current, (struct timezone *)0);
    if (tp->time.current.tv_sec - tp->time.accepted.tv_sec
						> tp->conn.keepalive.timeout) {
#if	DEBUG
      fprintf(stderr, "[threadSweep] cur: %ld, acpt: %ld, tout: %u\n",
	      tp->time.current.tv_sec, tp->time.accepted.tv_sec,
	      tp->conn.keepalive.timeout);
#endif	/* DEBUG */
      threadDetach(tp);
    }
  }
  if (usedThreads > 0) {
    gc();
  }
#if	DEBUG
  fprintf(stderr, "[threadSweep] active thread = %d\n", usedThreads);
#endif	/* DEBUG */
  return usedThreads;
}

/*
 * Function Name:	threadClear
 * Description:		clearing message entity of thread
 * Arguments:		thread
 * Return Value:	none
 */
void
threadClear(thd)
	register Thread *	thd;
{
#if	DEBUG
  fputs("<threadClear>\n", stderr);
#endif	/* DEBUG */
  /* I/O buffer */
  thd->ios.update = FALSE;
  thd->ios.total  = 0;
  thd->ios.next	  = 0;
  memset(thd->ios.buf, 0, sizeof(thd->ios.buf));
  cleanupMessage(thd);
  thd->mqueue.status	= MSG_NONE;
  thd->mqueue.vhost	= httpd->host;
  thd->mqueue.conn	= &thd->conn;
  thd->msg		= &thd->mqueue;
}

/*
 * Function Name:	init
 * Description:		initialize a thread
 * Arguments:		thread
 * Return Value:	none
 */
static void
init(thd)
	register Thread *	thd;
{
  /* status */
  thd->status = TH_FREE;
  /* timestamp */
  thd->time.accepted.tv_sec = thd->time.accepted.tv_usec = 0;
  thd->time.last.tv_sec	    = thd->time.last.tv_usec	 = 0;
  thd->time.current.tv_sec  = thd->time.current.tv_usec	 = 0;
  /* sockets */
#if	USE_SSL
  thd->conn.sock.ssl_ctx = NULL;
  thd->conn.sock.ssl	 = NULL;
#endif	/* USE_SSL */
  thd->conn.sock.sd	 = -1;
  thd->conn.sock.type	 = SOC_UNSECURE;
  thd->conn.sock.port	 = -1;
  thd->conn.sock.events	 = POLL_IN;
  thd->conn.sock.revents = 0;
  memset((char *)&thd->conn.sock.ainfo, 0, sizeof(thd->conn.sock.ainfo));
  /* client */
  wvFree(thd->conn.client.port); thd->conn.client.port = NULL;
  wvFree(thd->conn.client.ip);   thd->conn.client.ip   = NULL;
  wvFree(thd->conn.client.name); thd->conn.client.name = NULL;
  memset((char *)&thd->conn.client.addr, 0, sizeof(struct sockaddr_storage));
  /* keep-alive */
  thd->conn.keepalive.status	= KA_DISABLED;
  thd->conn.keepalive.timeout	= DEF_KA_TIMEOUT;
  thd->conn.keepalive.max	= DEF_KA_MAX;
  /* I/O buffer */
  thd->ios.update	= FALSE;
  thd->ios.total	= 0;
  thd->ios.next		= 0;
  memset(thd->ios.buf, 0, sizeof(thd->ios.buf));
  /* client request */
  thd->msg		= &thd->mqueue;

  thd->next		= NULL;
}

/*
 * Function Name:	cleanupMessage
 * Description:		cleanup message queue
 * Arguments:		thread
 * Return Value:	none
 */
static void
cleanupMessage(thd)
	register Thread *	thd;
{
  register Message *	mp;
  register Message *	mp_next;

  mp = thd->mqueue.next;
  cleanup(&thd->mqueue);
  for (; mp != NULL; mp = mp_next) {
    mp_next = mp->next;
    cleanup(mp);
    free((Message *)mp);
    mp = NULL;
  }
}

/*
 * Function Name:	cleanup
 * Description:		cleanup message entity in the thread
 * Arguments:		message
 * Return Value:	none
 */
static void
cleanup(msg)
	register Message *	msg;
{
  register Headers *	hp;
  register Headers *	hp_next;

  /* status */
  msg->status = MSG_NONE;
  /* virtual host */
  msg->vhost = NULL;
  /* connection */
  msg->conn = NULL;
  /* timestamp */
  msg->accesstime = 0;
  /* request */
  msg->request.method = M_NONE;
  memset(msg->request.uri, 0, sizeof(msg->request.uri));
  msg->request.http = HTTP_10;
  for (hp = msg->request.headers; hp != NULL; hp = hp_next) {
    hp_next = hp->next;
    wvFree(hp->name);	 hp->name  = NULL;
    wvFree(hp->value);	 hp->value = NULL;
    free((Headers *)hp); hp = NULL;
  }
  msg->request.headers = NULL;
  wvFree(msg->request.path_info);    msg->request.path_info	= NULL;
  wvFree(msg->request.query_string); msg->request.query_string	= NULL;
  if (msg->request.post_fd > 0) {
    wvClose(msg->request.post_fd);
  }
  msg->request.post_fd = -1;
  wvFree(msg->request.post_tmp); msg->request.post_tmp = NULL;
  msg->request.host		= NULL;
  msg->request.content_length	= 0;
  msg->request.x_forwarded_for	= NULL;
  /* response */
  msg->response.status = STATUS_NONE;
  for (hp = msg->response.headers; hp != NULL; hp = hp_next) {
    hp_next = hp->next;
    wvFree(hp->name);	 hp->name  = NULL;
    wvFree(hp->value);	 hp->value = NULL;
    free((Headers *)hp); hp = NULL;
  }
  msg->response.headers	= NULL;
  msg->response.content_length	= 0;
  msg->response.error_mesg	= NULL;
  /* content */
  if (msg->file.fd > 0) {
    wvClose(msg->file.fd);
  }
  msg->file.fd		= -1;
  msg->file.data	= NULL;
  msg->file.curp	= NULL;
  msg->file.rest	= 0;
  wvFree(msg->file.path); msg->file.path = NULL;
  msg->file.mime	= NULL;
  memset((char *)&msg->file.status, 0, sizeof(struct stat));
  msg->file.range.first	= 0;
  msg->file.range.last	= 0;
  /* user */
  wvFree(msg->user.indexDir);  msg->user.indexDir  = NULL;
  msg->user.homeDir = NULL;
  wvFree(msg->user.name);      msg->user.name	   = NULL;
  wvFree(msg->user.authName);  msg->user.authName  = NULL;
  wvFree(msg->user.identName); msg->user.identName = NULL;
  /* cgi */
  msg->cgi.pid = -1;
  if (msg->cgi.envno > 0) {
    register int	i;

    for (i = 0; i < msg->cgi.envno; i++) {
      wvFree(msg->cgi.env[i]);	msg->cgi.env[i] = NULL;
    }
    wvFree((String *)msg->cgi.env); msg->cgi.env    = NULL;
    wvFree(msg->cgi.status);	    msg->cgi.status = NULL;
  }
  msg->cgi.envno = 0;
  msg->cgi.nph	 = FALSE;

  msg->next = NULL;
}

