/*****************************************************************************
 *
 * FILE:	user.c
 * DESCRIPTION:	management user information
 * DATE:	Sat, Sep 18 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: user.c,v 1.26.2.11 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
typedef struct _UserInfo {
  String		name;	/* owner of UserDir */
  String		home;	/* home directory */
  struct _UserInfo *	prev;
  struct _UserInfo *	next;
} UserInfo;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static String		getHomeDir(const String);
static void		freeUserEntry(UserInfo *);
static void		insertUserCache(UserInfo *, UserInfo *);
static UserInfo *	removeUserCache(UserInfo *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
  				/* management user information */
static UserInfo	userTable[USER_TABLE_SIZE];
static UserInfo	userList;	/* header of user_table	*/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	translateUserDir
 * Description:		translate an user directory of requests URI
 * Arguments:		message
 * Return Value:	successful: 0, no translated: 1, otherwise: -1
 */
int
translateUserDir(msg)
	register Message *	msg;
{
  size_t	len;
  int		n;

#if	DEBUG
  fputs("<translateUserDir>\n", stderr);
#endif	/* DEBUG */
  if (msg->vhost->userDir == NULL || msg->request.uri[1] != '~') {
    return RET_CONT;
  }

  len		 = strcspn(msg->request.uri   + TILDE_OFFSET, "/");
  msg->user.name = wvStrndup(msg->request.uri + TILDE_OFFSET, len);
  if (msg->user.name == NULL) {
    goto error;
  }
  msg->user.homeDir = getHomeDir(msg->user.name);
  if (msg->user.homeDir == NULL) {
    msg->response.status = STATUS_NOT_FOUND;
    return RET_NG;
  }

  /* actual path of request file */
  n = asprintf(&msg->user.indexDir, "%s/%s",
	       msg->user.homeDir, msg->vhost->userDir);
  if (n == -1) {
    goto error;
  }
  n = asprintf(&msg->file.path, "%s%s",
	       msg->user.indexDir, msg->request.uri + TILDE_OFFSET + len);
  if (n == -1) {
    goto error;
  }
#if	DEBUG
  fprintf(stderr, "[translateUserDir] %s\n", msg->file.path);
#endif	/* DEBUG */
  return RET_OK;

error:
  return wvNoMemoryError(msg);
}


/*
 * Function Name:	getHomeDir
 * Description:		return the user's home directory
 *			(user record is registered into cache)
 * Arguments:		user name
 * Return Value:	user's home directory
 */
static String
getHomeDir(username)
	const String	username;
{
  register UserInfo *	uip;
  struct passwd	*	pwd;	/* pointer to password entry 	*/

  /* get user information */
  pwd = getpwnam(username);
  if (pwd == NULL) {
    return NULL;
  }

  /* check userList */
  for (uip = userList.next; uip != &userList; uip = uip->next) {
    if (uip->name == NULL || uip->home == NULL) {
      continue;
    }
    if (wvStrcmp(username, uip->name) == 0) {	/* found */
      removeUserCache(uip);
      insertUserCache(&userList, uip);
#if	DEBUG
      fprintf(stderr, "[userDir] found %s(%s)\n", uip->name, uip->home);
#endif	/* DEBUG */
      return uip->home;
    }
  }

  /* get top entry from list, and move one to end of list */
  uip = removeUserCache(userList.next);
  insertUserCache(&userList, uip);

  /* free old memory area */
  if (uip->name != NULL) {
    freeUserEntry(uip);
  }

  /* register new entry */
  wvFree((String)uip->name);
  uip->name = wvStrdup(pwd->pw_name);	/* ignore NULL return */
  wvFree((String)uip->home);
  uip->home = wvStrdup(pwd->pw_dir);	/* ignore NULL return */

#if	DEBUG
  fprintf(stderr, "[userDir] new %s(%s)\n", uip->name, uip->home);
#endif	/* DEBUG */
  return uip->home;
}

/*****************************************************************************/

/*
 * Function Name:	userCacheInit
 * Description:		initialize userTable called at only once
 * Arguments:		none
 * Return Value:	none
 */
void
userCacheInit(void)
{
  register int	i;

  userList.next = userList.prev = &userList;
  for (i = 0; i < USER_TABLE_SIZE; i++) {
    userTable[i].name = NULL;
    userTable[i].home = NULL;
    insertUserCache(&userList, &userTable[i]);
  }
}

/*
 * Function Name:	freeUserEntry
 * Description:		free memory area of userTable
 * Arguments:		user entry
 * Return Value:	none
 */
static void
freeUserEntry(p)
	register UserInfo *	p;
{
#if	DEBUG
  fprintf(stderr, "[userDir] freed %s(%s)\n", p->name, p->home);
#endif	/* DEBUG */
  wvFree(p->name);
  wvFree(p->home);
}

/*
 * Function Name:	insertUserCache
 * Description:		insert 'new' before 'pos'
 * Arguments:		position, new slot
 * Return Value:	none
 */
static void
insertUserCache(pos, new)
	register UserInfo *	pos;
	register UserInfo *	new;
{
  pos->prev->next = new;
  new->prev	  = pos->prev;
  new->next	  = pos;
  pos->prev	  = new;
}

/*
 * Function Name:	removeUserCache
 * Description:		remove the slot from userList
 * Arguments:		removed slot
 * Return Value:	removed slot
 */
static UserInfo *
removeUserCache(x)
	register UserInfo *	x;
{
  x->prev->next = x->next;
  x->next->prev = x->prev;
  return x;
}

