/*****************************************************************************
 *
 * FILE:	util.c
 * DESCRIPTION:	miscellaneous/utility functions
 * DATE:	Sat, Sep 16 2000
 * UPDATE:	Sun, Dec 19 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(c) 2000-2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (c) 2000-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: util.c,v 1.19.2.12 2004/12/19 13:55:27 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#if	TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if	TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif	/* TM_IN_SYS_TIME */
#endif	/* TIME_WITH_SYS_TIME */
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	wvSkipLine
 * Description:		skip white space, '# and CR-LF
 * Arguments:		string
 * Return Value:	white space etc: 0, otherwise: -1
 */
int
wvSkipLine(s)
	const String	s;
{
  switch (s[0]) {
    case '#':
    case SPC:
    case TAB:
    case CR:
    case LF:
      return 0;
    default:
      return -1;
  }
}

/*
 * Function Name:	wvChomp
 * Description:		remove  CRLF, CR or LF from string end
 * Arguments:		string
 * Return Value:	string
 * Note:		destroy original string
 */
String
wvChomp(s)
	String	s;
{
  register String	p;
  size_t		len;

  if (s == NULL) {
    return NULL;
  }

  len = strlen(s);
  p   = &s[len - 1];
  if (*p == LF) {
    *p-- = EOL;
  }
  if (*p == CR) {
    *p-- = EOL;
  }

  return s;
}

/*
 * Function Name:	wvClose
 * Description:		close a file/socket descriptor
 * Arguments:		descriptor
 * Return Value:	same as close(2)
 */
int
wvClose(d)
	int	d;
{
  int	ret;

  while (((ret = close(d)) < 0) && (errno == EINTR)) {
    ;
  }
  return ret;
}


/*
 * Function Name:	wvNoMemoryError
 * Description:		set usual error message
 * Arguments:		message
 * Return Value:	always -1
 */
int
wvNoMemoryError(msg)
	register Message *	msg;
{
  msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
  wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
  return -1;
}

/****************************************************************************/

/*
 * Function Name:	wvStrcmp
 * Description:		compare two strings
 * Arguments:		string, string
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvStrcmp(s1, s2)
	register const String	s1;
	register const String	s2;
{
  register size_t	n1;
  register size_t	n2;

  if (s1 == NULL || s2 == NULL) {
    return -1;
  }

  n1 = strlen(s1);
  n2 = strlen(s2);
  if (n1 != n2) {
    return -1;
  }

  return wvStrncmp(s1, s2, n1);
}

/*
 * Function Name:	wvStrncmp
 * Description:		compare two strings
 * Arguments:		string, string, length
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvStrncmp(s1, s2, n)
	register String	s1;
	register String	s2;
	register size_t	n;
{
  register size_t	n1;
  register size_t	n2;

  if (s1 == NULL || s2 == NULL) {
    return -1;
  }

  if (n < 0) {
    return -1;
  }

  if (n != 0) {
    do {
      if (*s1 != *s2++) {
      	return -1;
      }
      if (*s1++ == EOL) {
	break;
      }
    } while (--n != 0);
  }

  return 0;
}

/*
 * Function Name:	wvStrcasecmp
 * Description:		compare two strings
 * Arguments:		string, string
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvStrcasecmp(s1, s2)
	register const String	s1;
	register const String	s2;
{
  register size_t	n1;
  register size_t	n2;

  if (s1 == NULL || s2 == NULL) {
    return -1;
  }

  n1 = strlen(s1);
  n2 = strlen(s2);
  if (n1 != n2) {
    return -1;
  }

  return wvStrncasecmp(s1, s2, n1);
}

/*
 * Function Name:	wvStrncasecmp
 * Description:		compare two strings
 * Arguments:		string, string, length
 * Return Value:	successful: 0, otherwise: -1
 */
int
wvStrncasecmp(s1, s2, n)
	register const String	s1;
	register const String	s2;
	register size_t		n;
{
  if (s1 == NULL || s2 == NULL) {
    return -1;
  }

  if (n < 0) {
    return -1;
  }

  if (n != 0) {
    register const unsigned char *	us1 = (const unsigned char *)s1;
    register const unsigned char *	us2 = (const unsigned char *)s2;
    do {
      if (tolower(*us1) != tolower(*us2++)) {
      	return -1;
      }
      if (*us1++ == EOL) {
	break;
      }
    } while (--n != 0);
  }

  return 0;
}

/*****************************************************************************/

/*
 * Function Name:	wvStrncpy
 * Description:		boundary checked strncpy()
 * Arguments:		dst, src, size
 * Return Value:	same as strncpy()
 */
String
wvStrncpy(dst, src, size)
	String	dst;
	String	src;
	size_t	size;	/* size of dst buffer */
{
  register String	dp = dst;	/* working pointer */
  register String	sp = src;
  register size_t	n  = size;

  if (n != 0 && --n != 0) {
    do {
      if ((*dp++ = *sp++) == EOL) {
	break;
      }
    } while (--n != 0);
  }

  if (n == 0) {
    if (size != 0) {
      *dp = EOL;
    }
  }

  return dst;
}


/*
 * Function Name:	wvStrdup
 * Description:		strdup() with error checking
 * Arguments:		string
 * Return Value:	successful: same as strdup(), otherwise: NULL
 */
String
wvStrdup(s)
	const String	s;
{
  size_t	len;
  String	copy;

  if (s == NULL) {
    return NULL;
  }
  len  = strlen(s);
  copy = wvCalloc(1, len + 1);
  memcpy(copy, s, len);

  return copy;
}

/*
 * Function Name:	wvStrndup
 * Description:		duplicate a string like as strdup()
 * Arguments:		string, length
 * Return Value:	successful: new string, otherwise: NULL
 */
String
wvStrndup(s, len)
	String	s;
	size_t	len;
{
  size_t	size = len + 1;
  String	dst;
  String	ret;
  int		i;

  ret = dst = malloc(size);
  if (ret == NULL) {
    return NULL;
  }
  for (i = 0; i < len; i++) {
    *dst++ = *s++;
  }
  *dst = EOL;

  return ret;
}


/*
 * Function Name:	wvMalloc
 * Description:		malloc() with error checking
 * Arguments:		memory size
 * Return Value:	successful: same as malloc(), otherwise: NULL
 */
void *
wvMalloc(size)
	size_t	size;
{
  void *	p;

  p = malloc(size);
  if (p == NULL) {
    die(ERR_OUT_OF_MEMORY);
  }
  return p;
}

/*
 * Function Name:	wvCalloc
 * Description:		calloc() with error checking
 * Arguments:		object number, object size
 * Return Value:	successful: same as calloc(), otherwise: NULL
 */
void *
wvCalloc(num, size)
	size_t	num;
	size_t	size;
{
  void *	p;

  size *= num;
  p = wvMalloc(size);
  memset(p, 0, size);
  return p;
}

#if	0
/*
 * Function Name:	wvFree
 * Description:		free() with error checking
 * Arguments:		memory area
 * Return Value:	successful: same as free(), otherwise: NULL
 */
void
wvFree(p)
	void *	p;
{
  if (p != NULL) {
    free(p);
  }
}
#endif

/*****************************************************************************/

#ifndef	HAVE_USLEEP
/*
 * Function Name:	usleep
 * Description:		suspend process execution for an interval measurement
 *			in microseconds
 * Arguments:		microseconds
 * Return Value:	Successful: 0, otherwise: -1
 */
int
usleep(useconds)
	unsigned int	useconds;
{
  struct timeval	sleep_time;

  sleep_time.tv_usec = (useconds % USECONDS_BASE);
  sleep_time.tv_sec  = useconds / USECONDS_BASE;

  return select(0, NULL, NULL, NULL, &sleep_time);
}
#endif	/* HAVE_USLEEP */

#ifndef	HAVE_MKSTEMP
/*
 * Function Name:	mkstemp
 * Description:		create temporary file and return its file descriptor
 * Arguments:		template path
 * Return Value:	Successful: file descriptor, otherwise: -1
 */
int
mkstemp(path)
	char *	path;
{
  char *	p = mktemp(path);

  if (p == NULL) {
    return -1;
  }

  return open(path, O_CREAT|O_EXCL|O_RDWR, 0600);
}
#endif	/* HAVE_MKSTEMP */
