/*****************************************************************************
 *
 * FILE:	vhost.c
 * DESCRIPTION:	virtual host
 * DATE:	Sun, Apr 22 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2001-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: vhost.c,v 1.17.2.16 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <sys/types.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#if	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif	/* HAVE_ARPA_INET_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static VirtualHost *	lookup(const String);
static String		hostname(Message *);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	vhostNew
 * Description:		make a new virtual host in vhostList
 * Arguments:		virtual hostname
 * Return Value:	new virtual host pointer
 */
VirtualHost *
vhostNew(name)
	const String	name;
{
  VirtualHost *	new;

  new			= (VirtualHost *)wvMalloc(sizeof(VirtualHost));
  new->serverName	= wvStrdup(name);/* ServerName	*/
  new->serverAdmin	= NULL;		/* ServerAdmin	*/
  new->errorLog.path	= NULL;		/* ErrorLog	*/
  new->errorLog.fp	= NULL;		/* ErrorLog	*/
  new->accessLog.path	= NULL;		/* AccessLog	*/
  new->accessLog.fp	= NULL;		/* AccessLog	*/
  new->timeout		= DEF_TIMEOUT;	/* Timeout	*/
  new->docRoot		= NULL;		/* DocumentRoot	*/
  new->userDir		= NULL;		/* UserDir	*/
  new->dirIndexes	= NULL;		/* DirectoryIndex */
  new->autoIndexing	= FALSE;	/* AutoIndexing */
  new->symlinks		= FALSE;	/* FollowSymLinks */
  new->defaultType	= DEF_DEFTYPE;	/* DefaultType	*/
  new->errorDocTable	= NULL;		/* ErrorDocument */
  new->aliasTable	= NULL;		/* Redirect/Alias/ScriptAlias */
#if	0	//USE_SSL
  new->ssl_ctx			= NULL;
  new->sslCertificateFile	= NULL;	/* SSLCertificateFile */
  new->sslCertificateKeyFile	= NULL;	/* SSLCertificateKeyFile */
#endif	/* USE_SSL */
  new->next		= httpd->vhostTable;
  httpd->vhostTable	= new;

  return new;
}

/*
 * Function Name:	vhostDelete
 * Description:		delete virtual hosts in vhostList
 * Arguments:		none
 * Return Value:	none
 */
void
vhostDelete(void)
{
  register VirtualHost *	vp;
  register VirtualHost *	vp_next;

  for (vp = httpd->vhostTable; vp != NULL; vp = vp_next) {
    register StrList *	slp;
    register StrList *	slp_next;
    register ErrorDoc *	edp;
    register ErrorDoc *	edp_next;
    register Alias *	ap;
    register Alias *	ap_next;

    vp_next = vp->next;
    for (slp = vp->dirIndexes; slp != NULL; slp = slp_next) {
      slp_next = slp->next;
      wvFree((String)slp->text);
      wvFree((StrList *)slp);
    }
    for (edp = vp->errorDocTable; edp != NULL; edp = edp_next) {
      edp_next = edp->next;
      wvFree((String)edp->text);
      wvFree((ErrorDoc *)edp);
    }
    for (ap = vp->aliasTable; ap != NULL; ap = ap_next) {
      ap_next = ap->next;
      wvFree((String)ap->fakeName);
      wvFree((String)ap->realName);
      wvFree((Alias *)ap);
    }

    closeLog(vp);
    wvFree((String)vp->errorLog.path);
    wvFree((String)vp->accessLog.path);
    wvFree((String)vp->serverName);
    wvFree((String)vp->serverAdmin);
    wvFree((String)vp->docRoot);
    wvFree((String)vp->userDir);
#if	0
    /* no dynamic allocated */
    wvFree((String)vp->defaultType);
#endif
    wvFree((VirtualHost *)vp);
  }
  httpd->vhostTable = NULL;
}

/*
 * Function Name:	vhostSwitch
 * Description:		search virtual host and set current virtual host
 * Arguments:		message
 * Return Value:	Successful: 0, otherwise: -1
 */
int
vhostSwitch(msg)
	register Message *	msg;
{
  String		host;
  register ListenAddr *	lap;
  int			ret = RET_NG;

  host = hostname(msg);
  if (host == NULL) {
    msg->vhost = httpd->host;	/* default */
    return RET_OK;
  }

  msg->vhost = lookup(host);
  if (msg->vhost != NULL) {
    ret = RET_OK;
    goto done;
  }

  msg->vhost = httpd->host;	/* default */
  /* listen addresses are special. */
  for (lap = httpd->listen; lap != NULL; lap = lap->next) {
    if (lap->ipaddr == NULL ||	/* wildcard address */
	wvStrcmp(lap->ipaddr, host) == 0) {
      ret = RET_OK;
      goto done;
    }
  }

  /* localhost/127.0.0.1 are special. */
  if (wvStrcmp(host, "localhost") == 0 || wvStrcmp(host, "127.0.0.1") == 0 ||
      wvStrcmp(host, "::") == 0) {
    ret = RET_OK;
    goto done;
  }

  /* Why Bad Request? See RFC2616 Section 5.2! */
  msg->response.status = STATUS_BAD_REQUEST;
done:
  httpd->vhost = msg->vhost;
  wvFree((String)host);
  return ret;
}

/*
 * Function Name:	hostname
 * Description:		determine hostname for virtual host
 * Arguments:		message
 * Return Value:	Successful: hostname, otherwise: NULL
 */
static String
hostname(msg)
	register Message *	msg;
{
  String	host = msg->request.host;

  /* get hostname from Host field such as "Host: hostname[:port]" */
  if (host != NULL) {
    String		tmp;
    register String	p;
    int			n;

    tmp = wvStrdup(host);
    if (tmp == NULL) {
      goto error;
    }
    p = strchr(tmp, ':');
    if (p != NULL) {
      *p = EOL;
    }

    n = asprintf(&host, "%s", tmp);	free((String)tmp);
    if (n == -1) {
      goto error;
    }
  }
  /* get hostname from Request-URI */
  else if (msg->request.uri[0] != '/') {	/* absoluteURI */
    String		tmp;
    register String	sp;
    register String	ep;
    int			n;

    tmp = wvStrdup(msg->request.uri);
    if (tmp == NULL) {
      goto error;
    }
    sp  = strstr(tmp, "://");
    if (sp != NULL) {
      sp += 3;
      ep  = strchr(sp, '/');
      if (ep != NULL) {
	*ep = EOL;
      }
      ep = strchr(sp, ':');
      if (ep != NULL) {
	*ep = EOL;
      }
    }

    n = asprintf(&host, "%s", sp);	free((String)tmp);
    if (n == -1) {
      goto error;
    }
  }
  return host;

error:
  wvSyslog(LOG_ERR, "failed to resolve the virtual hostname\n");
  msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
  return NULL;
}

/*
 * Function Name:	lookup
 * Description:		lookup the virtual hosts in the list
 * Arguments:		hostname
 * Return Value:	successful: VirtualHost pointer, otherwise: NULL
 */
static VirtualHost *
lookup(name)
	const String	name;
{
  register VirtualHost *	vp;

  for (vp = httpd->host; vp != NULL; vp = vp->next) {
    if (wvStrcasecmp(vp->serverName, name) == 0) {
      return vp;
    }
  }

  return NULL;
}

