/*****************************************************************************
 *
 * FILE:	wyvern.c
 * DESCRIPTION:	wyvern main source file
 * DATE:	Sun, Aug 29 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: wyvern.c,v 1.65.2.23 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <signal.h>

#define	MAIN_MODULE
#include "wyvern.h"
#undef	MAIN_MODULE
#include "functions.h"


/*****************************************************************************
 *
 *	Program & Copyright
 *
 *****************************************************************************/
const char	Program[]   = "wyvern";
const char	Copyright[]
	= "Copyright (C) 1999-2003 Kouichi ABE (WALL), All rights reserved.";


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	httpdEntityInit(const String);
static void	internalHooks(void);
static int	defaultHandler(Message *);
static void	version(void);
static void	usage(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/


/*
 * Function Name:	quit
 * Description:		stop the service and exit
 * Arguments:		none
 * Return Value:	none
 */
void
quit(void)
{
  sigset_t	set;

  /* all signals are blocked */
  sigfillset(&set);
  sigprocmask(SIG_BLOCK, &set, NULL);

  /* close service channel */
  closeChannels();

  /* remove all modules */
  moduleDelete();

  /* close all log files */
  closeLogFiles();

  if (httpd->pid > 0) {		/* remove PID file to unlock */
    (void)unlink(httpd->pidFile);
  }

  exit(0);	/* release allocated memories automatically */
}

/*****************************************************************************/

/*
 * Function Name:	httpdEntityInit
 * Description:		initialize server entity
 * Arguments:		path of the configuration file
 * Return Value:	none
 */
static void
httpdEntityInit(configFile)
	const String	configFile;
{
  httpd = (Httpd *)malloc(sizeof(Httpd));
  if (httpd == NULL) {
    fprintf(stderr, "*** No enough memory for Wyvern ***\n");
    exit(71);	/* EX_OSERR */
  }

  httpd->listen		= NULL;			/* Listen */
  httpd->maxThread	= DEF_MAX_THREAD;	/* MaxThread */
  httpd->maxRequest	= DEF_MAX_REQUEST;	/* MaxRequest */
  httpd->user		= UID_WWW;		/* User */
  httpd->group		= UID_WWW;		/* Group */
  httpd->serverRoot	= DEF_SERVER_ROOT;	/* ServerRoot */
  httpd->secureLevel	= DEF_SECURE_LEVEL;	/* SecureLevel */
  httpd->securityLog.path	= DEF_SECURITY_LOG;	/* SecurityLog */
  httpd->securityLog.fp	= NULL;
  httpd->pidFile	= DEF_PID_FILE;		/* PidFile */
  httpd->pid		= -1;			/* Pid */
  httpd->vhostTable	= NULL;			/* virtual hosts manager */
  httpd->host		= vhostNew(DEF_VHOSTNAME);/* original server host */
  /* initialize original host entity */
  httpd->host->errorLog.path	= DEF_ERROR_LOG;
  httpd->host->errorLog.fp	= stderr;
  httpd->host->accessLog.path	= DEF_ACCESS_LOG;
  httpd->host->accessLog.fp	= NULL;
  httpd->host->docRoot		= DEF_DOC_ROOT;
  httpd->host->dirIndexes	= NULL;
  httpd->host->defaultType	= DEF_DEFTYPE;
  httpd->host->next		= NULL;
  httpd->vhostTable		= NULL;	/* reset virtual host table */
  /**/
  httpd->vhost		= httpd->host;		/* current virtual host */
  /**/
  httpd->shutdown	= FALSE;		/* server reboot flag */
  /**/
  httpd->sockets	= NULL;
  httpd->maxsoc		= -1;
  /**/
  httpd->trap.sigexit	= FALSE;
  httpd->trap.sighup	= FALSE;
  /**/
  httpd->startuptime	= time(NULL);
  /**/
  httpd->tmpDir		= DEF_TMP_DIR;	/* temporary working directory */
  httpd->cacheDir	= DEF_CACHE_DIR;	/* CacheDir */

  /* preserve config (wyvern.conf) path */
  if (configFile == NULL) {
    httpd->configFile = DEF_CONFIG;	/* constant variable */
  }
  else {
    if (configFile[0] != '/') {
      die("wyvern.conf requests an absolute path\n");
    }
    httpd->configFile = wvStrdup(configFile);	/* constant variable */
  }
}

static void
internalHooks(void)
{
  wvHookHeaderParser(parseHeader, HOOK_INITIAL);
  wvHookTranslateURI(parseURI, HOOK_INITIAL);		/* parsing */
  wvHookTranslateURI(translateAlias, HOOK_FINAL);	/* Alias */
  wvHookTranslateURI(translateUserDir, HOOK_FINAL);	/* UserDir */
  wvHookTranslateURI(translateDocRoot, HOOK_FINAL);	/* DocumentRoot */
  wvHookTypeChecker(defaultMimeTypeChecker, HOOK_FINAL);
  wvHookHandler(conditionalRequestChecker, HOOK_INITIAL, ANY_MAGIC_TYPE);
  wvHookHandler(defaultHandler, HOOK_FINAL, ANY_MAGIC_TYPE);
  wvHookLogger(defaultLogWriter, HOOK_FINAL);
}

static int
defaultHandler(msg)
	register Message *	msg;
{
#if	DEBUG
  fprintf(stderr, "[defaultHandler] mime = %s\n", msg->file.mime);
#endif	/* DEBUG */

  if (wvStrcmp(msg->file.mime, CGI_MAGIC_TYPE) == 0 ||
      wvStrcmp(msg->file.mime, SSI_MAGIC_TYPE) == 0 ||
      wvStrcmp(msg->file.mime, ECGI_MAGIC_TYPE) == 0 ||
      wvStrcmp(msg->file.mime, SUEXEC_MAGIC_TYPE) == 0 ||
      wvStrcmp(msg->file.mime, DIR_MAGIC_TYPE) == 0) {
    msg->response.status = STATUS_FORBIDDEN;
  }
  /* ordinary request */
  else if (msg->response.status == STATUS_NONE) {
    msg->response.status = STATUS_REQUEST_OK;
  }

  return RET_OK;
}

/*****************************************************************************/

/*
 * Function Name:	version
 * Description:		show the version of Wyvern
 * Arguments:		none
 * Return Value:	none
 */
static void
version(void)
{
  fprintf(stderr, "%s\n%s\n", SERVER_SOFTWARE, Copyright);
  fprintf(stderr, "[compiled at %s on %s by %s]\n",
		  __TIME__, __DATE__, COMPILED_USER);
#if	DEBUG
  fputs("+DEBUG ", stderr);
#endif	/* DEBUG */
#if	USE_JP_MESG
  fputs("+JMSG ", stderr);
#endif	/* USE_JP_MESG */
#if	USE_SUEXEC
  fputs("+SUEXEC ", stderr);
#endif	/* USE_SUEXEC */
#ifndef	BUGGY_KEEP_ALIVE
  fputs("-KEEP ", stderr);
#endif	/* BUGGY_KEEP_ALIVE */
#if	USE_SSL
  fputs("+SSL ", stderr);
#endif	/* USE_SSL */
  fputs("\n\n", stderr);
  exit(64);
}

/*
 * Function Name:	usage
 * Description:		show the simple usage for Wyvern
 * Arguments:		none
 * Return Value:	none
 */
static void
usage(void)
{
  fprintf(stderr,
	  "%s\n%s\n%s\n"
	  "usage: %s [-r config_file]\n\n",
	  SERVER_SOFTWARE, Copyright, SERVER_ADDRESS,
	  Program);
  exit(64);
}

/*****************************************************************************/

/*
 *	Wyvern Main
 */
int
main(argc, argv)
	int	argc;
	char	**argv;
{
  String	config	= NULL;		/* use default config file if NULL */

  switch (argc) {
    case 3:
      if (wvStrncmp(argv[1], "-r", 2) == 0) {
	config = argv[2];
      }
    case 1:
      break;
    default:
      if (wvStrncmp(argv[1], "-v", 2) == 0) {
	version();
      }
      usage();
  }

  /*
   * bootstrap
   */
  httpdEntityInit(config);	/* create server entity */
  internalHooks();		/* register internal hooks */
  readConfigFile();		/* read configuration file */
  detach();			/* detach the control terminal */
  openLogFiles();		/* initialize the log files */
  enforceSecureLevel();
  moduleInitializer();		/* initialize all modules */
  openChannels();		/* open service channel (socket) */
  storePID();			/* preserve server process ID	*/
  dropPrivilegeEternally();	/* drop privilege eternally */
  threadPoolInit();		/* initialize the thread queue */
  userCacheInit();		/* initialize the user cache list */
  signalTrap();			/* set signal handlers */

  httpdMain();			/* main request loop */

  quit();

  return 0;
}

