/*****************************************************************************
 *
 * FILE:	plain2dbm.c
 * DESCRIPTION:	convert a plain text-based password to dbm-based one
 * DATE:	Mon, Nov 13 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2000-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: plain2dbm.c,v 1.3 2002/01/08 16:30:00 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#if	USE_GDBM
#include <gdbm.h>
#else	/* NO GDBM */
#if	USE_NDBM
#include <fcntl.h>
#include <ndbm.h>
#endif	/* USE_NDBM */
#endif	/* USE_GDBM */
#include "common.h"
#include "wpath.h"


/*****************************************************************************
 *
 *	declaration of macros and structures
 *
 *****************************************************************************/
#if	0
#define	USE_VERIFY	1
#endif
#define	USERNAME_SIZE	(16+1)
#define	PASSWORD_SIZE	(48+1)


const char	Program[]   = "plain2dbm";
const char	Copyright[] = "Copyright (C) 2000-2002 Kouichi ABE (WALL)";


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	plain2dbm(String, String, String);
#if	USE_VERIFY
static void	verify(String);
#endif	/* USE_VERIFY */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	plain2dbm
 * DESCRIPTION:		convert to dbm-based password
 * ARGUMENTS:		user name, password, path of '.wpasswd'
 * RETURN VALUE:	none
 */
static void
plain2dbm(name, passwd, file)
	String	name;
	String	passwd;
	String	file;
{
#if	USE_GDBM
  String	path = NULL;
  size_t	size;
  GDBM_FILE	dp;
  datum		key;
  datum		val;
  int		ret;

  /* set database path */
  size = strlen(file) + strlen(GDBM_EXT) + 1;
  path = (String)calloc(size, sizeof(char));
  if (path == NULL) {
    perror("calloc");
    exit(1);
  }
  snprintf(path, size, "%s%s", file, GDBM_EXT);

  dp = gdbm_open(path, 512, GDBM_WRCREAT, 0600, NULL);
  if (dp == NULL) {
    perror("gdbm_open");
    exit(1);
  }
  free((String)path);

  key.dptr	= name;
  key.dsize	= strlen(key.dptr);
  val.dptr	= passwd;
  val.dsize	= strlen(val.dptr);

  ret = gdbm_store(dp, key, val, GDBM_REPLACE);
  if (ret != 0) {
    perror("gdbm_store");
    exit(1);
  }

  gdbm_close(dp);
#else	/* NO GDBM */
#if	USE_NDBM
  DBM *	dp;
  datum	key;
  datum	val;
  int	ret;

  dp = dbm_open(file, O_RDWR|O_CREAT|O_EXLOCK, 0600);
  if (dp == NULL) {
    perror("dbm_open");
    exit(1);
  }
  key.dptr	= name;
  key.dsize	= strlen(key.dptr);
  val.dptr	= passwd;
  val.dsize	= strlen(val.dptr);

  ret = dbm_store(dp, key, val, DBM_REPLACE);
  if (ret != 0) {
    perror("dbm_store");
    exit(1);
  }

  dbm_close(dp);
#endif	/* USE_NDBM */
#endif	/* USE_GDBM */
}

#if	USE_VERIFY
static void
verify(file)
	String	file;
{
#if	USE_GDBM
  String	path = NULL;
  size_t	size;
  GDBM_FILE	dp;
  datum		key;
  datum		val;
  String	name = NULL;
  String	pass = NULL;

  /* set database path */
  size = strlen(file) + strlen(GDBM_EXT) + 1;
  path = (String)calloc(size, sizeof(char));
  if (path == NULL) {
    perror("calloc");
    exit(1);
  }
  snprintf(path, size, "%s%s", file, GDBM_EXT);

  dp = gdbm_open(path, 512, GDBM_READER, 0400, NULL);
  if (dp == NULL) {
    perror("gdbm_open");
    exit(1);
  }
  free((String)path);

  for (key = gdbm_firstkey(dp); key.dptr; key = gdbm_nextkey(dp, key)) {
    name = (String)calloc((size_t)(key.dsize + 1), sizeof(char));
    if (name == NULL) {
      perror("calloc");
      gdbm_close(dp);
      exit(1);
    }
    val = gdbm_fetch(dp, key);
    pass = (String)calloc((size_t)(val.dsize + 1), sizeof(char));
    if (pass == NULL) {
      perror("calloc");
      gdbm_close(dp);
      exit(1);
    }
    strncpy(name, key.dptr, (size_t)key.dsize);
    strncpy(pass, val.dptr, (size_t)val.dsize);
    fprintf(stderr, "verify: %s = %s\n", name, pass);

    free((String)name);	name = NULL;
    free((String)pass);	pass = NULL;
  }

  gdbm_close(dp);
#else	/* NO GDBM */
#if	USE_NDBM
  DBM *		dp;
  datum		key;
  datum		val;
  String	name = NULL;
  String	pass = NULL;

  dp = dbm_open(file, O_RDONLY, 0400);
  if (dp == NULL) {
    perror("dbm_open");
    exit(1);
  }

  for (key = dbm_firstkey(dp); key.dptr; key = dbm_nextkey(dp)) {
    name = (String)calloc((size_t)(key.dsize + 1), sizeof(char));
    if (name == NULL) {
      perror("calloc");
      dbm_close(dp);
      exit(1);
    }
    val = dbm_fetch(dp, key);
    pass = (String)calloc((size_t)(val.dsize + 1), sizeof(char));
    if (pass == NULL) {
      perror("calloc");
      dbm_close(dp);
      exit(1);
    }
    strncpy(name, key.dptr, (size_t)key.dsize);
    strncpy(pass, val.dptr, (size_t)val.dsize);
    fprintf(stderr, "verify: %s = %s\n", name, pass);

    free((String)name);	name = NULL;
    free((String)pass);	pass = NULL;
  }

  dbm_close(dp);
#endif	/* USE_NDBM */
#endif	/* USE_GDBM */
}
#endif	/* USE_VERIFY */


/*
 *	MAIN FUNCTION
 */
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
#if	defined(USE_GDBM) || defined(USE_NDBM)
  FILE *	fin;
  char		buf[STD_BUFSIZE];
  char		name[USERNAME_SIZE];
  char		pass[PASSWORD_SIZE];
  int		n;

  if (argc == 1) {
    fprintf(stderr, "usage: %s <wpasswd file>\n", Program);
    exit(0);
  }

  fin = fopen(*++argv, "r");
  if (fin == NULL) {
    fprintf(stderr, "failed to open %s\n", *argv);
    exit(1);
  }

  memset(buf,  0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fin) != NULL) {
    memset(name, 0, sizeof(name));
    memset(pass, 0, sizeof(pass));
    n = sscanf(buf, "%16[^:]%48s", (char *)&name, (char *)&pass);
    if (n == 2) {
      plain2dbm(name, pass+1, *argv);
    }
    else {
      fprintf(stderr, "failed to parse text: %s.\n", buf);
    }
    memset(buf,  0, sizeof(buf));
  }

  fclose(fin);
#if	USE_VERIFY
  verify(*argv);
#endif	/* USE_VERIFY */
#else	/* DON'T USE ANY DBM */
  fputs("WARNING: wyvern does NOT support any dbm.\n", stderr);
#endif	/* DBM */
  return 0;
}

